/* sqldb.c
 *
 * Copyright (C) 1999,2000 Vivien Malerba
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 */

#include <config.h>
#include "sqldb.h"
#include <string.h>
#include "gasql_conf.h"

/*******************************************************************************/
/*                                                                             */
/* Main object: SqlDb                                                          */
/*                                                                             */
/*******************************************************************************/

static void      sql_db_class_init      (SqlDbClass *class);
static void      sql_db_init            (SqlDb *db);
static void      sql_db_destroy         (GtkObject *object);
static void      sql_db_cb              (GtkObject *obj, gpointer data);
static void      s_fault                (SqlDb *db);
static void      s_table_created_dropped(SqlDb *db, SqlMemTable *t);
static void      s_seq_created_dropped  (SqlDb *db, SqlMemSeq *t);
static void      s_field_created_dropped  (SqlDb *db, SqlMemTable *t,
					   SqlMemField *field);
static void      s_ff_link_created_dropped (SqlDb *db, SqlMemFlink *f);
static void      s_fs_link_created_dropped (SqlDb *db, SqlMemSeq *seq,
					    SqlMemField *f);


/* updates the DB structure */
static void      sql_db_load_tables     (SqlDb *db, SqlAccess *srv);
static void      sql_db_load_views      (SqlDb *db, SqlAccess *srv);
static void      sql_db_load_clean_unused_tables_views(SqlDb *db, 
						       SqlAccess *srv);

static void      sql_db_load_sequences  (SqlDb *db, SqlAccess *srv);

/* loads or reloads the fields of a table */
static void sql_mem_table_load_fields(SqlMemTable *t,
				      SqlAccess *srv, SqlDb *db);
static void sql_mem_field_load_contents(SqlAccess *srv, SqlMemField *field,
					Gda_Recordset* recset);
static void sql_db_catch_field_create_cb(GtkObject *obj, 
					SqlMemField *new_field, gpointer data);
static void sql_db_catch_field_drop_cb(GtkObject *obj,
				       SqlMemField *field, gpointer data);

/* signals */
enum {
  UPDATED,
  STRUCT_SAVED,
  FAULT,
  TABLE_CREATED,
  TABLE_CREATED_FILLED,
  TABLE_DROPPED,
  SEQ_CREATED,
  SEQ_DROPPED,
  FIELD_CREATED,
  FIELD_DROPPED,
  FF_LINK_CREATED,
  FF_LINK_DROPPED,
  FS_LINK_CREATED,
  FS_LINK_DROPPED,
  PROGRESS,
  LASTDB_SIGNAL
};

static gint sql_db_signals[LASTDB_SIGNAL] = { 0, 0, 0, 0, 0, 0, 0, 0, 
					      0, 0, 0, 0, 0, 0, 0 }; 

GtkType sql_db_get_type(void) 
{
  static GtkType f_type = 0;
  
  if (!f_type)
    {
      GtkTypeInfo f_info =
      {
        "Sql_Db",
        sizeof (SqlDb),	
        sizeof (SqlDbClass),
        (GtkClassInitFunc) sql_db_class_init,
        (GtkObjectInitFunc) sql_db_init,
        (GtkArgSetFunc) NULL,
        (GtkArgGetFunc) NULL
      };
      
      f_type = gtk_type_unique (gtk_object_get_type (), &f_info);
    }
  
  return f_type;   
}

static void sql_db_class_init(SqlDbClass *class)
{
  GtkObjectClass *object_class;
  
  object_class = (GtkObjectClass*) class;   
  sql_db_signals[FAULT] =
    gtk_signal_new ("fault",
                    GTK_RUN_FIRST,
                    object_class->type,
                    GTK_SIGNAL_OFFSET (SqlDbClass, fault),
                    gtk_signal_default_marshaller, GTK_TYPE_NONE, 0);
  sql_db_signals[STRUCT_SAVED] =
    gtk_signal_new ("struct_saved",
                    GTK_RUN_FIRST,
                    object_class->type,
                    GTK_SIGNAL_OFFSET (SqlDbClass, struct_saved),
                    gtk_signal_default_marshaller, GTK_TYPE_NONE, 0);
  sql_db_signals[UPDATED] =
    gtk_signal_new ("updated",
                    GTK_RUN_FIRST,
                    object_class->type,
                    GTK_SIGNAL_OFFSET (SqlDbClass, updated),
                    gtk_signal_default_marshaller, GTK_TYPE_NONE, 0);
 
  sql_db_signals[TABLE_CREATED] =
    gtk_signal_new ("table_created",
                    GTK_RUN_FIRST,
                    object_class->type,
                    GTK_SIGNAL_OFFSET (SqlDbClass, table_created),
                    gtk_marshal_NONE__POINTER, GTK_TYPE_NONE, 1, 
		    GTK_TYPE_POINTER);

  sql_db_signals[TABLE_CREATED_FILLED] =
    gtk_signal_new ("table_created_filled",
                    GTK_RUN_FIRST,
                    object_class->type,
                    GTK_SIGNAL_OFFSET (SqlDbClass, table_created_f),
                    gtk_marshal_NONE__POINTER, GTK_TYPE_NONE, 1, 
		    GTK_TYPE_POINTER);

  sql_db_signals[TABLE_DROPPED] =
    gtk_signal_new ("table_dropped",
                    GTK_RUN_FIRST,
                    object_class->type,
                    GTK_SIGNAL_OFFSET (SqlDbClass, table_dropped),
                    gtk_marshal_NONE__POINTER, GTK_TYPE_NONE, 1, 
		    GTK_TYPE_POINTER);

  sql_db_signals[SEQ_CREATED] =
    gtk_signal_new ("seq_created",
                    GTK_RUN_FIRST,
                    object_class->type,
                    GTK_SIGNAL_OFFSET (SqlDbClass, seq_created),
                    gtk_marshal_NONE__POINTER, GTK_TYPE_NONE, 1, 
		    GTK_TYPE_POINTER);

  sql_db_signals[SEQ_DROPPED] =
    gtk_signal_new ("seq_dropped",
                    GTK_RUN_FIRST,
                    object_class->type,
                    GTK_SIGNAL_OFFSET (SqlDbClass, seq_dropped),
                    gtk_marshal_NONE__POINTER, GTK_TYPE_NONE, 1, 
		    GTK_TYPE_POINTER);

  sql_db_signals[FIELD_CREATED] =
    gtk_signal_new ("field_created",
                    GTK_RUN_FIRST,
                    object_class->type,
                    GTK_SIGNAL_OFFSET (SqlDbClass, field_created),
                    gtk_marshal_NONE__POINTER_POINTER, GTK_TYPE_NONE, 2, 
		    GTK_TYPE_POINTER, GTK_TYPE_POINTER);

  sql_db_signals[FIELD_DROPPED] =
    gtk_signal_new ("field_dropped",
                   GTK_RUN_FIRST,
                    object_class->type,
                    GTK_SIGNAL_OFFSET (SqlDbClass, field_dropped),
                    gtk_marshal_NONE__POINTER_POINTER, GTK_TYPE_NONE, 2, 
		    GTK_TYPE_POINTER, GTK_TYPE_POINTER);

  sql_db_signals[FF_LINK_CREATED] =
    gtk_signal_new ("ff_link_created",
                    GTK_RUN_FIRST,
                    object_class->type,
                    GTK_SIGNAL_OFFSET (SqlDbClass, ff_link_created),
                    gtk_marshal_NONE__POINTER, GTK_TYPE_NONE, 1, 
		    GTK_TYPE_POINTER);

  sql_db_signals[FF_LINK_DROPPED] =
    gtk_signal_new ("ff_link_dropped",
                    GTK_RUN_FIRST,
                    object_class->type,
                    GTK_SIGNAL_OFFSET (SqlDbClass, ff_link_dropped),
                    gtk_marshal_NONE__POINTER, GTK_TYPE_NONE, 1, 
		    GTK_TYPE_POINTER);

  sql_db_signals[FS_LINK_CREATED] =
    gtk_signal_new ("fs_link_created",
                    GTK_RUN_FIRST,
                    object_class->type,
                    GTK_SIGNAL_OFFSET (SqlDbClass, fs_link_created),
                    gtk_marshal_NONE__POINTER_POINTER, GTK_TYPE_NONE, 2, 
		    GTK_TYPE_POINTER, GTK_TYPE_POINTER);

  sql_db_signals[FS_LINK_DROPPED] =
    gtk_signal_new ("fs_link_dropped",
                    GTK_RUN_FIRST,
                    object_class->type,
                    GTK_SIGNAL_OFFSET (SqlDbClass, fs_link_dropped),
                    gtk_marshal_NONE__POINTER_POINTER, GTK_TYPE_NONE, 2, 
		    GTK_TYPE_POINTER, GTK_TYPE_POINTER);

  sql_db_signals[PROGRESS] =
    gtk_signal_new ("progress",
                    GTK_RUN_FIRST,
                    object_class->type,
                    GTK_SIGNAL_OFFSET (SqlDbClass, progress),
		    gtk_marshal_NONE__POINTER_UINT_UINT, GTK_TYPE_NONE, 3, 
		    GTK_TYPE_POINTER, GTK_TYPE_UINT, GTK_TYPE_UINT);

  gtk_object_class_add_signals (object_class, sql_db_signals, LASTDB_SIGNAL);
  class->updated = NULL;
  class->struct_saved = NULL; 
  class->fault = s_fault;
  class->table_created = s_table_created_dropped;
  class->table_created_f = NULL;/*s_table_created_dropped;*/
  class->table_dropped = s_table_created_dropped;
  class->seq_created = s_seq_created_dropped;
  class->seq_dropped = s_seq_created_dropped;
  class->field_created = s_field_created_dropped;
  class->field_dropped = s_field_created_dropped;
  class->ff_link_created = s_ff_link_created_dropped;
  class->ff_link_dropped = s_ff_link_created_dropped;
  class->fs_link_created = s_fs_link_created_dropped;
  class->fs_link_dropped = s_fs_link_created_dropped;
  class->progress = NULL;
  object_class->destroy = sql_db_destroy;
}

static void sql_db_init(SqlDb *db)
{
  db->name = NULL;
  db->is_accessible = FALSE;
  db->is_fault = FALSE;
  db->tables = NULL;
  db->sequences = NULL;
  db->field_links = NULL;
  db->users = NULL;
}


GtkObject* sql_db_new(SqlAccess *srv)
{
  GtkObject *obj;
  SqlDb *db;

  obj = gtk_type_new(sql_db_get_type());
  db = SQL_DB(obj);

  db->srv = srv;
  gtk_signal_connect(GTK_OBJECT(srv), "conn_closed",
		     GTK_SIGNAL_FUNC(sql_db_empty_all_cb), db);

  return obj;
}

static void sql_db_destroy(GtkObject *object)
{
  SqlDb *db;
  GtkObject *parent_class = NULL;
  GSList *list;
  
  parent_class = gtk_type_class (gtk_object_get_type());
  g_return_if_fail (object != NULL);
  g_return_if_fail (IS_SQL_DB (object));

  db = SQL_DB(object); 

  /* destroying Tables list */
  if (db->tables) {
    list = db->tables;
    while(list) {
      gtk_object_destroy(GTK_OBJECT(list->data));
      list = g_slist_next(list);
    }
    g_slist_free(db->tables);
    db->tables = NULL;
  }
  /* destroying Sequences list */
  if (db->sequences) {
    list = db->sequences;
    while(list) {
      gtk_object_destroy(GTK_OBJECT(list->data));
      list = g_slist_next(list);
    }
    g_slist_free(db->sequences);
    db->sequences = NULL;
  }

  /* Field links list */
  if (db->field_links) {
    list = db->field_links;
    while(list) {
      gtk_object_destroy(GTK_OBJECT(list->data));
      list = g_slist_next(list);
    }
    g_slist_free(db->field_links);
    db->field_links = NULL;
  }

  /* the name */
  if (db->name) {
    g_free(db->name);
    db->name = NULL;
  }

  /* the users list */
  while (db->users)
    {
      g_free(db->users->data);
      db->users = g_slist_remove_link(db->users, db->users);
    }

  /* for the parent class */
  if (GTK_OBJECT_CLASS (parent_class)->destroy)
    (* GTK_OBJECT_CLASS (parent_class)->destroy) (object); 
}

void sql_db_empty_all_cb(GtkObject *obj, SqlDb *db)
{
  GSList *list;

  db->is_accessible = FALSE;

  /* first everything should be saved */
  /* MODIFY */

  list = db->tables;
  while (list) {
    gtk_object_destroy(GTK_OBJECT(list->data));
    list = g_slist_remove_link(list, list);
  }
  db->tables = NULL;

  list = db->sequences;
  while (list) {
    gtk_object_destroy(GTK_OBJECT(list->data));
    list = g_slist_remove_link(list, list);
  }
  db->sequences = NULL;

  list = db->field_links;
  while (list) {
    gtk_object_destroy(GTK_OBJECT(list->data));
    list = g_slist_remove_link(list, list);
  }
  db->field_links = NULL;
}


/* default function for the "fault" signal */
static void s_fault(SqlDb *db)
{
  db->is_fault = TRUE;
  db->is_accessible = FALSE;
}

static void s_ff_link_created_dropped (SqlDb *db, SqlMemFlink *f)
{
#ifdef debug_signal
  g_print(">> 'UPDATED' from sqldb->s_ff_link_created_dropped\n");
#endif
  gtk_signal_emit(GTK_OBJECT(db), sql_db_signals[UPDATED]);
#ifdef debug_signal
  g_print("<< 'UPDATED' from sqldb->s_ff_link_created_dropped\n");
#endif
}

static void s_fs_link_created_dropped (SqlDb *db, SqlMemSeq *seq,
				       SqlMemField *f)
{
#ifdef debug_signal
  g_print(">> 'UPDATED' from sqldb->s_fs_link_created_dropped\n");
#endif
  gtk_signal_emit(GTK_OBJECT(db), sql_db_signals[UPDATED]);
#ifdef debug_signal
  g_print("<< 'UPDATED' from sqldb->s_fs_link_created_dropped\n");
#endif
}

static void s_table_created_dropped(SqlDb *db, SqlMemTable *t)
{
#ifdef debug_signal
  g_print(">> 'UPDATED' from sqldb->s_table_created_dropped\n");
#endif
  gtk_signal_emit(GTK_OBJECT(db), sql_db_signals[UPDATED]);
#ifdef debug_signal
  g_print("<< 'UPDATED' from sqldb->s_table_created_dropped\n");
#endif
}

static void s_seq_created_dropped(SqlDb *db, SqlMemSeq *t)
{
#ifdef debug_signal
  g_print(">> 'UPDATED' from sqldb->s_seq_created_dropped\n");
#endif
  gtk_signal_emit(GTK_OBJECT(db), sql_db_signals[UPDATED]);
#ifdef debug_signal
  g_print("<< 'UPDATED' from sqldb->s_seq_created_dropped\n");
#endif
}

static void s_field_created_dropped(SqlDb *db, SqlMemTable *t,
				    SqlMemField *field)
{
#ifdef debug_signal
  g_print(">> 'UPDATED' from sqldb->s_field_created_dropped\n");
#endif
  gtk_signal_emit(GTK_OBJECT(db), sql_db_signals[UPDATED]);
#ifdef debug_signal
  g_print("<< 'UPDATED' from sqldb->s_field_created_dropped\n");
#endif
}

/****************************************/
/*                                      */
/* Callbacks                            */
/*                                      */
/****************************************/

/* this callback is called whenever a table or sequence 
   emits the signal "load_fault" */
static void sql_db_cb(GtkObject *obj, gpointer data)
{
#ifdef debug_signal
  g_print(">> 'FAULT' from sqldb->sql_db_cb\n");
#endif
  gtk_signal_emit(GTK_OBJECT(data), sql_db_signals[FAULT]);
#ifdef debug_signal
  g_print("<< 'FAULT' from sqldb->sql_db_cb\n");
#endif
  /* note: db->is_fault will automatically be set to TRUE */
}

/* This CB is called whenever a table emits the "field_created" signal, 
   so the DB object can forward that signal and emit the "field_created" signal
   for its listeners. Data points to the DB object */
static void sql_db_catch_field_create_cb(GtkObject *obj,
					 SqlMemField *new_field, gpointer data)
{
#ifdef debug_signal
  g_print(">> 'FIELD_CREATED' from sqldb->sql_db_catch_field_create_cb\n");
#endif
  gtk_signal_emit(GTK_OBJECT(data), sql_db_signals[FIELD_CREATED], 
		  obj, new_field);
#ifdef debug_signal
  g_print("<< 'FIELD_CREATED' from sqldb->sql_db_catch_field_create_cb\n");
#endif
}

/* This CB is called whenever a table emits the "field_dropped" signal, 
   so the DB object can forward that signal and emit the "field_dropped" signal
   for its listeners. Data points to the DB object */
static void sql_db_catch_field_drop_cb(GtkObject *obj,  
				       SqlMemField *field, gpointer data)
{
  /* emit signal */
#ifdef debug_signal
  g_print(">> 'FIELD_DROPPED' from sqldb->sql_db_catch_field_drop_cb\n");
#endif
  gtk_signal_emit(GTK_OBJECT(data), sql_db_signals[FIELD_DROPPED], 
		  obj, field);
#ifdef debug_signal
  g_print("<< 'FIELD_DROPPED' from sqldb->sql_db_catch_field_drop_cb\n");
#endif
}

/****************************************/
/*                                      */
/* Tables and fields manipulation       */
/*                                      */
/****************************************/
GSList *find_user_name(SqlDb *db, gchar *uname)
{
  GSList *list;
  GSList *found=NULL;

  list = db->users;
  while (!found && list)
    {
      if (!strcmp(uname, (gchar *)(list->data)))
	{
	  found=list;
	}
      list = g_slist_next(list);
    }
  if (!found)
    {
      db->users = g_slist_append(db->users, g_strdup(uname));
      found = g_slist_last(db->users);
    }

  return found;
}

SqlMemTable *sql_db_find_table_by_name (SqlDb *db, gchar *name)
{
  GSList *list = db->tables;

  while (list && (strcmp(name, ((SqlMemTable *)(list->data))->name))) {
    list = g_slist_next(list);
  }
  if (list) 
    return (SqlMemTable *) (list->data);
  else
    return NULL;
}

SqlMemSeq *sql_db_find_sequence_by_name(SqlDb *db, gchar *name)
{
  GSList *list = db->sequences;

  while (list && (strcmp(name, ((SqlMemSeq *)(list->data))->name))) {
    list = g_slist_next(list);
  }
  if (list) 
    return (SqlMemSeq *) (list->data);
  else
    return NULL;
}

SqlMemSeq* sql_db_find_sequence_to_field(SqlDb *db, SqlMemField *field)
{
  GSList *list;
  SqlMemSeq *seq=NULL;

  list = db->sequences;
  while (list && !seq) {
    if (g_slist_find(SQL_MEM_SEQ(list->data)->field_links, field))
      seq = SQL_MEM_SEQ(list->data);
    else
      list = g_slist_next(list);
  }
  return seq;
}

SqlMemField* sql_db_find_field_from_names(SqlDb *db, gchar *table, 
					    char *field)
{
  SqlMemTable *tab;
  SqlMemField *f=NULL;
  GSList *list;

  tab = sql_db_find_table_by_name(db, table);
  if (tab) {
    list = tab->fields;
    while (list && !f) {
      if (! strcmp(SQL_MEM_FIELD(list->data)->name, field))
	f = SQL_MEM_FIELD(list->data);
      list = g_slist_next(list);
    }
  }
  return f;
}

SqlMemTable *sql_db_find_table_from_field(SqlDb *db, SqlMemField *field)
{
  SqlMemTable *table = NULL;
  GSList *tlist = db->tables;
  GSList *flist;

  while (tlist && !table) {
    flist = ((SqlMemTable *)(tlist->data))->fields;
    while (flist && !table) {
      if (flist->data == field)
	table = (SqlMemTable *)(tlist->data);
      flist = g_slist_next(flist);
    }
    tlist = g_slist_next(tlist);
  }

  return table;
}

guint sql_db_get_field_order(SqlDb *db, SqlMemTable *table, SqlMemField *field)
{
  SqlMemTable *tab;
  GSList *list;
  guint order, i;

  if (!table) 
    tab = sql_db_find_table_from_field(db, field);
  else
    tab = table;

  if (!tab) /* field does not belong to any table */
    return 0;

  list = tab->fields;
  order = 0;
  i = 0;
  while (list && !order) {
    i++;
    if (list->data == field) 
      order = i;
    list = g_slist_next(list);
  }
  return order;
}

void sql_db_refresh(SqlDb *db, SqlAccess *srv)
{
  /* the update procedure */
  if (sql_access_is_open(srv)) {
    sql_db_load_tables(db, srv);
    sql_db_load_views(db, srv);
    sql_db_load_clean_unused_tables_views(db, srv);
    sql_db_load_sequences(db, srv);
  }
  else
    g_warning("SqlDb::sql_db_refresh() -> connection closed!\n");

#ifdef debug_signal
    g_print(">> 'UPDATED' from sql_db_refresh\n");
#endif
    gtk_signal_emit(GTK_OBJECT(db), sql_db_signals[UPDATED]);
#ifdef debug_signal
    g_print("<< 'UPDATED' from sql_db_refresh\n");
#endif
}

static void sql_db_load_tables(SqlDb *db, SqlAccess *srv)
{
  SqlMemTable *table;
  GSList *list, *list2;
  gboolean created;
  Gda_Recordset* rst;
  Gda_Field* field;
  gchar *str;
  guint now, total;

  db->is_accessible = FALSE;

  if (db->name) {
    g_free(db->name);
    db->name = NULL;
  }
  db->name = g_strdup(srv->gda_datasource->str);

  /* getting the tables list */
  rst = gda_connection_open_schema(GDA_CONNECTION(srv),
				   GDA_Connection_GDCN_SCHEMA_TABLES,
				   GDA_Connection_EXTRA_INFO, "t",
				   GDA_Connection_no_CONSTRAINT);

  if (rst)
    {
      /* FIXME: use the supports() function to see if the total number of
	 tuples can be got that way */
      total = rst->affected_rows;

      now = 0;
      gda_recordset_move(rst, 1, 0);
      while (!gda_recordset_eof(rst))
	{
	  now++;
	  field = gda_recordset_field_idx(rst, 0);
	  str = gda_stringify_value(NULL, 0, field); /* table name */
	  table = sql_db_find_table_by_name(db, str);
	  if (! table) /* the table doesn't already exist, create it */
	    { 
	      table = SQL_MEM_TABLE(sql_mem_table_new());
	      table->name = str;
	      str = NULL;
	      
	      gtk_signal_connect(GTK_OBJECT(table), "load_fault",
				 GTK_SIGNAL_FUNC(sql_db_cb), db);
	      gtk_signal_connect(GTK_OBJECT(table), "field_created",
				 GTK_SIGNAL_FUNC(sql_db_catch_field_create_cb), db);
	      gtk_signal_connect(GTK_OBJECT(table), "field_dropped",
				 GTK_SIGNAL_FUNC(sql_db_catch_field_drop_cb), db);
	      db->tables = g_slist_append(db->tables, table);
	      created = TRUE;
	    }
	  else
	    {
	      created = FALSE;
	    }
	  if (str)
	    g_free(str);

	  /* (re)computes the comments */
	  field = gda_recordset_field_idx(rst, 2); /* comments */
	  if (gda_field_longvarchar(field))
	    {
	      str = gda_stringify_value(NULL, 0, field);
	      sql_mem_table_set_comments(table, NULL, str, FALSE);
	      g_free(str);
	    }
	  /* (re)computes the owner */
	  field = gda_recordset_field_idx(rst, 1); /* owner */
	  if (gda_field_longvarchar(field))
	    {
	      GSList *list;
	      str = gda_stringify_value(NULL, 0, field);
	      list = find_user_name(db, str);
	      table->owner = (gchar *)(list->data);
	      g_free(str);
	    }

	  if (created) /* to emit signals now because then owner and comments filled */
	    {
#ifdef debug_signal
	      g_print(">> 'TABLE_CREATED' from sql_db_load_tables\n");
#endif
	      gtk_signal_emit(GTK_OBJECT(db), sql_db_signals[TABLE_CREATED], table);
#ifdef debug_signal
	      g_print("<< 'TABLE_CREATED' from sql_db_load_tables\n");
#endif
	    }

	  sql_mem_table_load_fields(table, srv, db); 
	  if (created) 
	    {
#ifdef debug_signal
	      g_print(">> 'TABLE_CREATED_FILLED' from sql_db_load_tables\n");
#endif
	      gtk_signal_emit(GTK_OBJECT(db), 
			      sql_db_signals[TABLE_CREATED_FILLED], 
			      table);
#ifdef debug_signal
	      g_print("<< 'TABLE_CREATED_FILLED' from sql_db_load_tables\n");
#endif
	    }
	  /* mark the table as updated */
	  ((SqlMemItem *) table)->updated = TRUE;

	  gtk_signal_emit_by_name(GTK_OBJECT(db), "progress", 
				  _("Updating the list of tables..."), 
				  now, total);
	  gda_recordset_move_next(rst);
	}
      gda_recordset_free(rst);

      if (!(db->is_fault)) 
	{
	  /***************************************************************/
	  /* 2nd step: re-compute the parents table list for every table */
	  /***************************************************************/
	  list = db->tables;
	  while (list) 
	    {
	      if ((SQL_MEM_TABLE(list->data)->is_view == FALSE) &&
		  (SQL_MEM_ITEM(list->data)->updated == TRUE))
		{
		  g_slist_free(SQL_MEM_TABLE(list->data)->parents);
		  SQL_MEM_TABLE(list->data)->parents = NULL;
		  /* get the parents' list */
		  rst = gda_connection_open_schema(GDA_CONNECTION(srv),
						   GDA_Connection_GDCN_SCHEMA_TAB_PARENTS,
						   GDA_Connection_OBJECT_NAME, SQL_MEM_TABLE(list->data)->name ,
						   GDA_Connection_no_CONSTRAINT);
		  
		  if (rst)
		    {
		      list2 = NULL;
		      gda_recordset_move(rst, 1, 0);
		      while (!gda_recordset_eof(rst))
			{
			  field = gda_recordset_field_idx(rst, 0);
			  str = gda_stringify_value(NULL, 0, field); 
			  table = sql_db_find_table_by_name(db, str);
			  g_free(str);
			  if (table)
			    list2 = g_slist_append(list2, table);
			  gda_recordset_move_next(rst);
			}
		      SQL_MEM_TABLE(list->data)->parents = list2;
		      gda_recordset_free(rst);
		    }
		}
	      list = g_slist_next(list);
	    }
	  db->is_accessible = TRUE;
	}
    }
  else 
    {
      /* An error occured, the db emits the fault signal */
#ifdef debug_signal
      g_print(">> 'FAULT' from sql_db_load_tables\n");
#endif
      gtk_signal_emit(GTK_OBJECT(db), sql_db_signals[FAULT]);
#ifdef debug_signal
      g_print("<< 'FAULT' from sql_db_load_tables\n");
#endif
    }
  /* if an error occurs, the DB remains non accessible */
  gtk_signal_emit_by_name(GTK_OBJECT(db), "progress", NULL, 0, 0);
}

static void sql_db_load_views(SqlDb *db, SqlAccess *srv)
{
  SqlMemTable *table;
  gboolean created;
  Gda_Recordset* rst;
  Gda_Field* field;
  gchar *str;
  guint now, total;

  db->is_accessible = FALSE;

  if (db->name) {
    g_free(db->name);
    db->name = NULL;
  }
  db->name = g_strdup(srv->gda_datasource->str);

  /* getting the tables list */
  rst = gda_connection_open_schema(GDA_CONNECTION(srv),
				   GDA_Connection_GDCN_SCHEMA_VIEWS,
				   GDA_Connection_EXTRA_INFO, "t",
				   GDA_Connection_no_CONSTRAINT);

  if (rst)
    {
      /* FIXME: use the supports() function to see if the total number of
	 tuples can be got that way */
      total = rst->affected_rows;

      now = 0;
      gda_recordset_move(rst, 1, 0);
      while (!gda_recordset_eof(rst))
	{
	  now++;
	  field = gda_recordset_field_idx(rst, 0);
	  str = gda_stringify_value(NULL, 0, field); /* table name */
	  table = sql_db_find_table_by_name(db, str);
	  if (! table) /* the table doesn't already exist, create it */
	    { 
	      table = SQL_MEM_TABLE(sql_mem_table_new());
	      table->name = str;
	      str = NULL;

	      gtk_signal_connect(GTK_OBJECT(table), "load_fault",
				 GTK_SIGNAL_FUNC(sql_db_cb), db);
	      gtk_signal_connect(GTK_OBJECT(table), "field_created",
				 GTK_SIGNAL_FUNC(sql_db_catch_field_create_cb), db);
	      gtk_signal_connect(GTK_OBJECT(table), "field_dropped",
				 GTK_SIGNAL_FUNC(sql_db_catch_field_drop_cb), db);
	      db->tables = g_slist_append(db->tables, table);
	      created = TRUE;
	    }
	  else
	    {
	      created = FALSE;
	    }
	  if (str)
	    g_free(str);

	  /* set it here because otherwise if an object was a table and 
	     becomes a view, this won't be set correctly */
	  table->is_view = TRUE;

	  /* (re)computes the comments */
	  field = gda_recordset_field_idx(rst, 2); /* comments */
	  if (gda_field_longvarchar(field))
	    {
	      str = gda_stringify_value(NULL, 0, field);
	      sql_mem_table_set_comments(table, NULL, str, FALSE);
	      g_free(str);
	    }
	  /* (re)computes the owner */
	  field = gda_recordset_field_idx(rst, 1); /* owner */
	  if (gda_field_longvarchar(field))
	    {
	      GSList *list;
	      str = gda_stringify_value(NULL, 0, field);
	      list = find_user_name(db, str);
	      table->owner = (gchar *)(list->data);
	      g_free(str);
	    }

	  if (created) /* to emit signals now because then owner and comments filled */
	    {
#ifdef debug_signal
	      g_print(">> 'TABLE_CREATED' from sql_db_load_tables\n");
#endif
	      gtk_signal_emit(GTK_OBJECT(db), sql_db_signals[TABLE_CREATED], table);
#ifdef debug_signal
	      g_print("<< 'TABLE_CREATED' from sql_db_load_tables\n");
#endif
	    }
	  sql_mem_table_load_fields(table, srv, db); 
	  if (created) 
	    {
#ifdef debug_signal
	      g_print(">> 'TABLE_CREATED_FILLED' from sql_db_load_tables\n");
#endif
	      gtk_signal_emit(GTK_OBJECT(db), 
			      sql_db_signals[TABLE_CREATED_FILLED], 
			      table);
#ifdef debug_signal
	      g_print("<< 'TABLE_CREATED_FILLED' from sql_db_load_tables\n");
#endif
	    }
	  /* mark the table as updated */
	  ((SqlMemItem *) table)->updated = TRUE;

	  gtk_signal_emit_by_name(GTK_OBJECT(db), "progress", 
				  _("Updating the list of views..."), 
				  now, total);
	  gda_recordset_move_next(rst);
	}
      gda_recordset_free(rst);
    }
  gtk_signal_emit_by_name(GTK_OBJECT(db), "progress", NULL, 0, 0);
}

static void sql_db_load_clean_unused_tables_views(SqlDb *db, SqlAccess *srv)
{
  GSList *list, *list2, *list_hold;

  list = db->tables;
  while (list) {
    if (! SQL_MEM_ITEM(list->data)->updated) {
      GtkObject *obj;
      /* we want to emit the signal "table_dropped" with a pointer to 
	 a table which still exists, but is not anymore in the list 
	 of tables in the DB. */
      obj = GTK_OBJECT(list->data);
      /* we first remove all the links to any field of that table */
      list2 = SQL_MEM_TABLE(obj)->fields;
      while(list2) 
	{
	  sql_db_delete_seq_field_link(db, NULL, 
				       SQL_MEM_FIELD(list2->data));
	  sql_db_delete_field_field_link(db, NULL, 
					 SQL_MEM_FIELD(list2->data));
	  sql_db_delete_field_field_link(db, 
					 SQL_MEM_FIELD(list2->data),
					 NULL);
	  list2 = g_slist_next(list2);
	}
      list_hold = g_slist_next(list);
      db->tables = g_slist_remove_link(db->tables, list);
      list = list_hold;
#ifdef debug_signal
      g_print(">> 'TABLE_DROPPED' from sql_db_load_tables\n");
#endif
      gtk_signal_emit(GTK_OBJECT(db), sql_db_signals[TABLE_DROPPED], 
		      obj);
#ifdef debug_signal
      g_print("<< 'TABLE_DROPPED' from sql_db_load_tables\n");
#endif
      gtk_object_destroy(obj);
    }
    else {
      SQL_MEM_ITEM(list->data)->updated = FALSE;
      list = g_slist_next(list);
    }
  }
}


/* TO MODIFY: to update what is already into memory */
static void sql_db_load_sequences(SqlDb *db, SqlAccess *srv)
{
  SqlMemSeq *seq;
  GSList *list, *list_hold;
  Gda_Recordset* rs;
  Gda_Field* field;
  gchar *str;
  guint now, total;
  gboolean created;

  db->is_accessible = FALSE;

  /* 1: getting the sequence list */
  rs = gda_connection_open_schema(GDA_CONNECTION(srv),
				  GDA_Connection_GDCN_SCHEMA_SEQUENCES,
				  GDA_Connection_EXTRA_INFO, "t",
				  GDA_Connection_no_CONSTRAINT);
  if (rs)
    {
      /* FIXME: use the supports() function to see if the total number of
	 tuples can be got that way */
      total = rs->affected_rows;

      now = 0;
      gda_recordset_move(rs, 1, 0);
      while (!gda_recordset_eof(rs))
	{
	  now++;
	  field = gda_recordset_field_idx(rs, 0);
	  str = gda_stringify_value(NULL, 0, field); 
	  created = FALSE;
	  seq = sql_db_find_sequence_by_name(db, str);	  
	  if (!seq)
	    {
	      seq = SQL_MEM_SEQ(sql_mem_seq_new());
	      seq->name = str;
	      str = NULL;
	      gtk_signal_connect(GTK_OBJECT(seq), "load_fault",
				 GTK_SIGNAL_FUNC(sql_db_cb), db);
	      db->sequences = g_slist_append(db->sequences, seq);
	      created = TRUE;
	    }
	  if (str)
	    g_free(str);

	  /* (re)computes the comments */
	  field = gda_recordset_field_idx(rs, 2); /* comments */
	  if (gda_field_longvarchar(field))
	    {
	      str = gda_stringify_value(NULL, 0, field); /* str allocated */
	      if (seq->comments)
		g_free(seq->comments);
	      seq->comments = str;
	    }

	  /* (re)computes the owner */
	  field = gda_recordset_field_idx(rs, 1); /* owner */
	  if (gda_field_longvarchar(field))
	    {
	      GSList *list;
	      str = gda_stringify_value(NULL, 0, field);
	      list = find_user_name(db, str);
	      seq->owner = (gchar *)(list->data);
	      g_free(str);
	    }

	  if (created) /* then emit a signal */
	    {
#ifdef debug_signal
	      g_print(">> 'SEQ_CREATED' from sql_db_load_sequences (%s)\n", 
		      seq->name);
#endif
	      gtk_signal_emit(GTK_OBJECT(db), 
			      sql_db_signals[SEQ_CREATED], seq);
#ifdef debug_signal
	      g_print("<< 'SEQ_CREATED' from sql_db_load_sequences\n");
#endif
	    }

	  /* mark the sequence as updated */
	  SQL_MEM_ITEM(seq)->updated = TRUE;
	  gtk_signal_emit_by_name(GTK_OBJECT(db), "progress", 
				  _("Updating the list of sequences..."), 
				  now, total);
	  gda_recordset_move_next(rs);
	}
      gda_recordset_free(rs);

      /* 2nd step: destroy sequences that do not exist anymore */
      list = db->sequences;
      while (list) 
	{
	  if (! SQL_MEM_ITEM(list->data)->updated) 
	    {
	      GtkObject *obj;
	      
	      /* we want to emit the "seq_dropped" signal with a SqlMemSeq 
		 object
		 which still exists but is not anymore in the SqlDb 
		 structure */
	      obj = GTK_OBJECT(list->data);
	      /* we want to remove all the links from that sequence first */
	      sql_db_delete_seq_field_link(db, SQL_MEM_SEQ(obj), NULL);
	      
	      list_hold = g_slist_next(list);
	      db->sequences = g_slist_remove_link(db->sequences, list);
	      list = list_hold;
	      /* let's first remove any link from that sequence */
	      sql_db_delete_seq_field_link(db, SQL_MEM_SEQ(obj), NULL); 
#ifdef debug_signal
	      g_print(">> 'SEQ_DROPPED' from sql_db_load_sequences (%s)\n",
		      SQL_MEM_SEQ(obj)->name);
#endif
	      gtk_signal_emit(GTK_OBJECT(db), sql_db_signals[SEQ_DROPPED],
			      obj);
#ifdef debug_signal
	      g_print("<< 'SEQ_DROPPED' from sql_db_load_sequences\n");
#endif
	      gtk_object_destroy(obj);
	    }
	  else 
	    {
	      SQL_MEM_ITEM(list->data)->updated = FALSE;
	      list = g_slist_next(list);
	    }
	}
      db->is_accessible = TRUE;
    }
  else 
    {
      /* An error occured, the db emits the fault signal */
#ifdef debug_signal
      g_print(">> 'FAULT' from sql_db_load_sequences\n");
#endif
      gtk_signal_emit(GTK_OBJECT(db), sql_db_signals[FAULT]);
#ifdef debug_signal
      g_print("<< 'FAULT' from sql_db_load_sequences\n");
#endif
    }  
  /* if an error occurs, the DB remains non accessible */  
  gtk_signal_emit_by_name(GTK_OBJECT(db), "progress", NULL, 0, 0);
}


#ifdef debug
/****************************************/
/*                                      */
/* Dumping the tables and structure     */
/*                                      */
/****************************************/

void sql_db_dump_tables (SqlDb *db)
{
  GSList *list, *plist;

  list = db->tables;
  g_print("*** Dumping Database Structure ***\n");
  while(list) {
    g_print("  TABLE %s", SQL_MEM_TABLE(list->data)->name);
    if (SQL_MEM_TABLE(list->data)->parents) {
      g_print(", parents= ");
      plist = SQL_MEM_TABLE(list->data)->parents;
      while (plist) {
	g_print("%s ", SQL_MEM_TABLE(plist->data)->name);
	plist = g_slist_next(plist);
      }
      g_print(".\n");
    }
    else 
      g_print(", no known parents.\n");
    sql_mem_table_dump_fields(SQL_MEM_TABLE(list->data));
    list = g_slist_next(list);
  }
}

void sql_db_dump_links(SqlDb *db)
{
  GSList *list, *list2;
  gchar *str;

  g_print("*** Dumping Sequences Relations ***\n");
  list = db->sequences;
  while (list) {
    list2 = SQL_MEM_SEQ(list->data)->field_links;
      while (list2) {
	g_print("\t%-25s --> %s.%s\n",
		SQL_MEM_SEQ(list->data)->name,
		sql_db_find_table_from_field(db, SQL_MEM_FIELD(list2->data))->name,
		SQL_MEM_FIELD(list2->data)->name);
	list2 = g_slist_next(list2);
      }
    list = g_slist_next(list);
  }

  g_print("*** Dumping Field to Field Relations ***\n");
  list = db->field_links;
  while (list) {
    str = g_strdup_printf("%s.%s",
	sql_db_find_table_from_field(db, SQL_MEM_FIELD(SQL_MEM_FLINK(list->data)->from))->name,
	SQL_MEM_FIELD(SQL_MEM_FLINK(list->data)->from)->name);
    g_print("\t%-25s --> %s.%s\n",
	    str,
	    sql_db_find_table_from_field(db, SQL_MEM_FIELD(SQL_MEM_FLINK(list->data)->to))->name,
	    SQL_MEM_FIELD(SQL_MEM_FLINK(list->data)->to)->name);
    g_free(str);
    list = g_slist_next(list);
  }
}

void sql_db_dump_struct_as_graph(SqlDb *db, FILE *st)
{
  GSList *list;

  if (db->is_accessible) {
    fprintf(st, "graph: {\n");
    fprintf(st, "title: \"SqlDb tables dumping\"\n\n");
    
    /* sequences */
    list = db->sequences;
    while(list) {
      sql_mem_seq_dump_as_graph(SQL_MEM_SEQ(list->data), db, st);
      list = g_slist_next(list);
    }
    
    /* Field links */
    list = db->field_links;
    while (list) {
      sql_mem_flink_dump_as_graph(SQL_MEM_FLINK(list->data), db, st);
      list = g_slist_next(list);
    }
    
    /* tables */
    list = db->tables;
    while(list) {
      sql_mem_table_dump_as_graph(SQL_MEM_TABLE(list->data), st);
      list = g_slist_next(list);
    }

    fprintf(st, "}\n");
  }
}
#endif


/* if the link does not already exists, insert the link between the
   sequence and the field */
void sql_db_insert_seq_field_link(SqlDb *db, SqlMemSeq *seq, 
				  SqlMemField *field)
{
  GSList *list;
  gboolean found = FALSE;

  if ((sql_db_find_sequence_by_name(db, seq->name) == seq) &&
      (sql_db_find_table_from_field(db, field) != NULL)) {
    list = seq->field_links;
    while (list && !found) {
      if (list->data == field)
	found = TRUE;
      list = g_slist_next(list);
    }
    if (!found) { /* then insert it */
      seq->field_links = g_slist_append(seq->field_links, field);
      /* emit a signal */
#ifdef debug_signal
      g_print(">> 'FS_LINK_CREATED' from sql_db_insert_seq_field_link\n");
#endif
      gtk_signal_emit(GTK_OBJECT(db), sql_db_signals[FS_LINK_CREATED],
		      seq, field);
#ifdef debug_signal
      g_print("<< 'FS_LINK_CREATED' from sql_db_insert_seq_field_link\n");
#endif
    }
  }
  else 
    g_warning("SqlDb::sql_db_insert_seq_field_link\tyou are trying to access "
	      "the fields of the wrong SqlDb structure!\n");
}

/* removes from memory the specified link or all links if seq==NULL or
 field==NULL */
void sql_db_delete_seq_field_link(SqlDb *db, SqlMemSeq *seq, 
				  SqlMemField *field)
{
  GSList *list;
  SqlMemField *f;

  if (field)
    if (((seq && (sql_db_find_sequence_by_name(db, seq->name) == seq) &&
	  (sql_db_find_table_from_field(db, field) != NULL))) ||
	(!seq && (sql_db_find_table_from_field(db, field) != NULL))) {
      if (seq) {
	seq->field_links = g_slist_remove(seq->field_links, field);
#ifdef debug_signal
	g_print(">> 'FS_LINK_DROPPED' from sql_db_delete_seq_field_link "
		"1 (%s->%s)\n", seq->name, field->name);
#endif
	gtk_signal_emit(GTK_OBJECT(db), sql_db_signals[FS_LINK_DROPPED],
			seq, field);
#ifdef debug_signal
	g_print("<< 'FS_LINK_DROPPED' from sql_db_delete_seq_field_link\n");
#endif
      }
      else {
	list = db->sequences;
	while (list) {
	  if (g_slist_find(SQL_MEM_SEQ(list->data)->field_links, field)) {
	    SQL_MEM_SEQ(list->data)->field_links = 
	      g_slist_remove(SQL_MEM_SEQ(list->data)->field_links, field);
#ifdef debug_signal
	    g_print(">> 'FS_LINK_DROPPED' from sql_db_delete_seq_field_link "
		    "2 (%s->%s)\n", SQL_MEM_SEQ(list->data)->name, 
		    field->name);
#endif
	    gtk_signal_emit(GTK_OBJECT(db), sql_db_signals[FS_LINK_DROPPED],
			    SQL_MEM_SEQ(list->data), field);
#ifdef debug_signal
	    g_print("<< 'FS_LINK_DROPPED' "
		    "from sql_db_delete_seq_field_link\n");
#endif
	  }
	  list = g_slist_next(list);
	}
      }
    }
    else 
      g_warning("SqlDb::sql_db_delete_seq_field_link\nyou are "
		"trying to access "
		"the fields of the wrong SqlDb structure!\n");
  else /* field = NULL */
    if (seq) {
      while (seq->field_links) {
	f = SQL_MEM_FIELD(seq->field_links->data);
	seq->field_links = g_slist_remove_link(seq->field_links, 
					       seq->field_links);
#ifdef debug_signal
	g_print(">> 'FS_LINK_DROPPED' from sql_db_delete_seq_field_link "
		"(%s->%s)\n", seq->name, f->name);
#endif
	gtk_signal_emit(GTK_OBJECT(db), sql_db_signals[FS_LINK_DROPPED],
			seq, f);
#ifdef debug_signal
	g_print("<< 'FS_LINK_DROPPED' from sql_db_delete_seq_field_link\n");
#endif
      }
      seq->field_links = NULL;
    }
}

void sql_db_insert_field_field_link(SqlDb *db, SqlMemField *from, 
				    SqlMemField *to)
{
  GSList *list = db->field_links;
  gboolean found = FALSE;
  SqlMemFlink *item;

  while (list && !found) {
    if ((SQL_MEM_FLINK(list->data)->from == from) &&
	(SQL_MEM_FLINK(list->data)->to == to))
      found = TRUE;
    list = g_slist_next(list);
  }
  if (!found) {
    item = SQL_MEM_FLINK(sql_mem_flink_new());
    item->from = from;
    item->to = to;
    db->field_links = g_slist_append(db->field_links, item);
    /* emit a signal */
#ifdef debug_signal
    g_print(">> 'FF_LINK_CREATED' from sql_db_insert_field_field_link\n");
#endif
    gtk_signal_emit(GTK_OBJECT(db), sql_db_signals[FF_LINK_CREATED],
		    item);
#ifdef debug_signal
    g_print("<< 'FF_LINK_CREATED' from sql_db_insert_field_field_link\n");
#endif
  }
}

void sql_db_delete_field_field_link(SqlDb *db, SqlMemField *from, 
				    SqlMemField *to)
{
  GSList *list, *hlist;
  gboolean found = FALSE;
  SqlMemFlink *item;
  guint mode=0;

  if (!to) mode += 1;
  if (!from) mode += 2;

  list = db->field_links;
  while (list && !found) {
    item = NULL;
    switch (mode) {
    case 0:
      if ((SQL_MEM_FLINK(list->data)->from == from) &&
	  (SQL_MEM_FLINK(list->data)->to == to)) {
	found = TRUE;
	item = SQL_MEM_FLINK(list->data);
      }
      break;
    case 1:
      if (SQL_MEM_FLINK(list->data)->from == from) 
	item = SQL_MEM_FLINK(list->data);
      break;
    case 2:
      if (SQL_MEM_FLINK(list->data)->to == to) 
	item = SQL_MEM_FLINK(list->data);
      break;
    }
    if (item) {
      hlist = g_slist_next(list);
      db->field_links = g_slist_remove(db->field_links, item);
#ifdef debug_signal
      g_print(">> 'FF_LINK_DROPPED' from sql_db_delete_field_field_link\n");
#endif
      gtk_signal_emit(GTK_OBJECT(db), sql_db_signals[FF_LINK_DROPPED],
		      item);
#ifdef debug_signal
      g_print("<< 'FF_LINK_DROPPED' from sql_db_delete_field_field_link\n");
#endif
      gtk_object_destroy(GTK_OBJECT(item));
      list = hlist;
    }
    else
      list = g_slist_next(list);
  }
  if (found) {   /* emit a signal */
#ifdef debug_signal
    g_print(">> 'UPDATED' from sql_db_delete_field_field_link\n");
#endif
    gtk_signal_emit(GTK_OBJECT(db), sql_db_signals[UPDATED]);
#ifdef debug_signal
    g_print("<< 'UPDATED' from sql_db_delete_field_field_link\n");
#endif
  }
}

SqlMemFlink* sql_db_find_flink_from_fields(SqlDb *db, SqlMemField *fromf,
					   SqlMemField *tof)
{
  GSList *list;
  SqlMemFlink *retval, *fl;

  retval = NULL;
  list = db->field_links;
  while (list && !retval) {
    fl = SQL_MEM_FLINK(list->data);
    if ((fl->from == fromf) && (fl->to == tof))
      retval = fl;
    list = g_slist_next(list);
  }

  return retval;
}

/*************************************************/
/*                                               */
/* XML storage of the SqlDb object               */
/*                                               */
/*************************************************/
/* to store SqlDb object */
void sql_db_build_xml_tree(SqlDb *db, xmlDocPtr doc)
{
  xmlNodePtr toptree, tree, subtree, subsubtree;
  GSList *list, *list2;
  SqlMemTable *table, *table2;
  SqlMemSeq *seq;
  SqlMemField *field;
  SqlMemFlink *flink;
  gint order;
  gchar *str;
  SqlDataType* datatype=NULL;

  /* main node */
  toptree = xmlNewChild(doc->root, NULL, "SQLDB", db->name);

  /* tables */
  tree = xmlNewChild(toptree, NULL, "TABLES", NULL);
  list = db->tables;
  while(list) {
    table = SQL_MEM_TABLE(list->data);
    subtree = xmlNewChild(tree, NULL, "table", table->name);
    /* comments */
    if (table->comments)
      xmlSetProp(subtree, "comments", table->comments);
    if (table->is_user_comments)
      xmlSetProp(subtree, "user_comments", "t");
    /* owner */
    if (table->owner)
      xmlSetProp(subtree, "owner", table->owner);
    /* table/view */
    if (table->is_view)
      xmlSetProp(subtree, "is_view", "t");

    list2 = table->parents;
    order = 1;
    while (list2) {
      subsubtree = xmlNewChild(subtree, NULL, "parent", 
			       SQL_MEM_TABLE(list2->data)->name);  
      str = g_strdup_printf("%d", order);
      xmlSetProp(subsubtree, "order", str);
      g_free(str);
      order++;
      list2 = g_slist_next(list2);
    }

    list2 = table->fields;
    while (list2) {
      field = SQL_MEM_FIELD(list2->data);
      subsubtree = xmlNewChild(subtree, NULL, "field", field->name);
      datatype = field->type;
      xmlSetProp(subsubtree, "type", datatype->sqlname);
      /* if ((datatype->numparams >= 1) && (field->length > 0)) { */
/* 	str = g_strdup_printf("%d", field->length); */
/* 	xmlSetProp(subsubtree, "length", str); */
/* 	g_free(str); */
/*       } */
      str = g_strdup_printf("%d", field->length);
      xmlSetProp(subsubtree, "length", str);
      g_free(str);

      if (field->null_allowed)
	xmlSetProp(subsubtree, "notnull", "f");
      else
	xmlSetProp(subsubtree, "notnull", "t");
      if (field->is_key)
	xmlSetProp(subsubtree, "iskey", "t");
      else
	xmlSetProp(subsubtree, "iskey", "f");
      if (field->default_val)
	xmlSetProp(subsubtree, "default", field->default_val);

      list2 = g_slist_next(list2);
    }
    list = g_slist_next(list);
  }


  /* Sequences */
  tree = xmlNewChild(toptree, NULL, "SEQUENCES", NULL);
  list = db->sequences;
  while(list) {
    seq = SQL_MEM_SEQ(list->data);
    subtree = xmlNewChild(tree, NULL, "sequence", seq->name);

    /* comments  */
    if (seq->comments)
      xmlSetProp(subtree, "comments", seq->comments);    

    /* owner */
    if (seq->owner)
      xmlSetProp(subtree, "owner", seq->owner);

    list2 = seq->field_links;
    while (list2) {
      field = SQL_MEM_FIELD(list2->data);
      table = sql_db_find_table_from_field(db, field);
      if (table) {
	subsubtree = xmlNewChild(subtree, NULL, "link", NULL);
	xmlSetProp(subsubtree, "to_table", table->name);
	xmlSetProp(subsubtree, "to_field", field->name);
      }
      list2 = g_slist_next(list2);
    }
    list = g_slist_next(list);
  }    
    
  /* Field links */
  tree = xmlNewChild(toptree, NULL, "FIELDLINKS", NULL);
  list = db->field_links;
  while(list) {
    flink = SQL_MEM_FLINK(list->data);
    
    table = sql_db_find_table_from_field(db, flink->from);
    table2 = sql_db_find_table_from_field(db, flink->to);
    if (table && table2) {
      subtree = xmlNewChild(tree, NULL, "link", NULL);
      xmlSetProp(subtree, "from_table", table->name);
      xmlSetProp(subtree, "from_field", flink->from->name);
      xmlSetProp(subtree, "to_table", table2->name);
      xmlSetProp(subtree, "to_field", flink->to->name);
    }
    list = g_slist_next(list);
  }    
}

/* to load sqldb object 
   the SqlDb object MUST NOT have any table, sequence or field_links. */
gboolean sql_db_build_db_from_xml_tree(SqlDb *db, xmlNodePtr node)
{
  xmlNodePtr tree, subtree;
  GSList *ref_list, *list, *fields_ref_list;
  gchar *str;
  gboolean ok = TRUE;
  
  if (db->tables || db->sequences || db->field_links) {
    gnome_error_dialog(_("INTERNAL ERROR:\nTrying to load an SqlDb object "
			 "which already have some data inside. Clean it "
			 "before."));
    return FALSE;
  }
  /* sets the DB name */
  if (db->name)
    g_free(db->name);
  db->name = gasql_xml_clean_string(xmlNodeGetContent(node));

  tree = node->childs;
  fields_ref_list = NULL;
  while (tree) 
    {
      /* Tables loading */
      if (!strcmp(tree->name, "TABLES")) 
	{
	  subtree = tree->childs;
	  ref_list = NULL;
	  while (subtree) 
	    {
	      if (strcmp(subtree->name, "table")) 
		{
		  str = g_strdup_printf(_("Problem with the structure of the "
					  "XML file:\nin tag <TABLES>, the "
					  "tag <%s> is not allowed.\nError "
					  "dismissed"), subtree->name);
		  gnome_warning_dialog(str);
		  g_free(str);
		}
	      else 
		ref_list = sql_mem_table_load_from_xml_node(db, subtree, 
							    ref_list);
	      subtree = subtree->next;
	    }
	  ref_list = gasql_xml_update_ref1(ref_list, db, 
		       GASQL_XML_UPDATE_REF1(sql_db_find_table_by_name));
	  if (ref_list) 
	    {
	      GString *strn;

	      strn = g_string_new(_("The following tables are referenced "
				    "by other "
				    "tables and cannot be found:\n"));
	      list = ref_list;
	      while (list) 
		{
		  g_string_sprintfa(strn, "'%s'\n", 
				    ((gasqlXmlNodeRef *)(list->data))->txt);
		  list = g_slist_next(list);
		}
	      strn = g_string_append(strn, _("It is recommended to repair the "
					     "XML file you were trying "
					     "to load."));
	      gnome_error_dialog(strn->str);
	      g_string_free(strn, TRUE);
	      ok = FALSE;
	      tree = NULL; /* leaving the loading... */
	      /* empty the remaining list */
	      while (ref_list) 
		ref_list = gasql_xml_remove_ref_list(ref_list, ref_list);
	    }
	}
      
      /* Sequences Loading */
      if (tree && !strcmp(tree->name, "SEQUENCES")) 
	{
	  subtree = tree->childs;
	  while (subtree) 
	    {
	      if (strcmp(subtree->name, "sequence")) 
		{
		  str = g_strdup_printf(_("Problem with the structure of the "
					  "XML file:\nin tag <SEQUENCES>, the "
					  "tag <%s> is not allowed.\nError "
					  "dismissed"), subtree->name);
		  gnome_warning_dialog(str);
		  g_free(str);
		}
	      else
		fields_ref_list = sql_mem_seq_load_from_xml_node(db, subtree, 
							      fields_ref_list);
	      subtree = subtree->next;
	    }
	}

      /* Field links Loading */
      if (tree && !strcmp(tree->name, "FIELDLINKS")) 
	{
	  subtree = tree->childs;
	  while (subtree) 
	    {
	      if (strcmp(subtree->name, "link")) 
		{
		  str = g_strdup_printf(_("Problem with the structure of the "
					  "XML file:\nin tag <FIELDLINKS>, "
					  "the "
					  "tag <%s> is not allowed.\nError "
					  "dismissed"), subtree->name);
		  gnome_warning_dialog(str);
		  g_free(str);
		}
	      else 
		fields_ref_list = 
		  sql_mem_flink_load_from_xml_node(db, subtree, 
						   fields_ref_list);
	      subtree = subtree->next;
	    }
	}
      
      if (tree)
	tree = tree->next;
    }
  
  if (ok) {
    fields_ref_list = gasql_xml_update_ref2(fields_ref_list, db, 
		    GASQL_XML_UPDATE_REF2(sql_db_find_field_from_names));
    if (fields_ref_list) {
      GString *strn;
      
      strn = g_string_new(_("The following fields are referenced "
			    "by links and cannot "
			    "be found:\n"));
      list = fields_ref_list;
      while (list) {
	g_string_sprintfa(strn, "'%s.%s'\n", 
			  ((gasqlXmlNodeRef *)(list->data))->txt,
			  ((gasqlXmlNodeRef *)(list->data))->txt2);
	list = g_slist_next(list);
      }
      strn = g_string_append(strn, _("It is recommended to repair the "
				     "XML file you were trying to load."));
      gnome_error_dialog(strn->str);
      g_string_free(strn, TRUE);
      ok = FALSE;
      tree = NULL; /* leaving the loading... */
    }
  }
  else {
    /* remove everything in the SqlDb object */
    list = db->tables;
    while (list) {
      gtk_object_destroy(GTK_OBJECT(list->data));
      list = g_slist_remove_link(list, list);
    }
    list = db->sequences;
    while (list) {
      gtk_object_destroy(GTK_OBJECT(list->data));
      list = g_slist_remove_link(list, list);
    }
    list = db->field_links;
    while (list) {
      gtk_object_destroy(GTK_OBJECT(list->data));
      list = g_slist_remove_link(list, list);
    }
  }

  /* sending signals if everything is correct */
  if (ok) {
    /* tables */
    list = db->tables;
    while (list) {
#ifdef debug_signal
      g_print(">> 'TABLE_CREATED_FILLED' from sql_db_build_db_from_xml_tree\n");
#endif
      gtk_signal_emit(GTK_OBJECT(db), sql_db_signals[TABLE_CREATED_FILLED],
		      list->data);
#ifdef debug_signal
      g_print("<< 'TABLE_CREATED_FILLED' from sql_db_build_db_from_xml_tree\n");
#endif
      list = g_slist_next(list);
    }
    list = db->sequences;
    while (list) {
#ifdef debug_signal
      g_print(">> 'SEQ_CREATED' from sql_db_build_db_from_xml_tree\n");
#endif
      gtk_signal_emit(GTK_OBJECT(db), sql_db_signals[SEQ_CREATED],
		      list->data);
#ifdef debug_signal
      g_print("<< 'SEQ_CREATED' from sql_db_build_db_from_xml_tree\n");
#endif
      ref_list = SQL_MEM_SEQ(list->data)->field_links;
      while (ref_list) {	
#ifdef debug_signal
	g_print(">> 'FS_LINK_CREATED' from sql_db_build_db_from_xml_tree\n");
#endif
	gtk_signal_emit(GTK_OBJECT(db), sql_db_signals[FS_LINK_CREATED],
			list->data, ref_list->data);
#ifdef debug_signal
	g_print("<< 'FS_LINK_CREATED' from sql_db_build_db_from_xml_tree\n");
#endif
	ref_list = g_slist_next(ref_list);
      }
      list = g_slist_next(list);
    }
    list = db->field_links;
    while (list) {
#ifdef debug_signal
      g_print(">> 'FF_LINK_CREATED' from sql_db_build_db_from_xml_tree\n");
#endif
      gtk_signal_emit(GTK_OBJECT(db), sql_db_signals[FF_LINK_CREATED],
		      list->data);
#ifdef debug_signal
      g_print("<< 'FF_LINK_CREATED' from sql_db_build_db_from_xml_tree\n");
#endif
      list = g_slist_next(list);
    }
  }
  
  /* empty the remaining fields_ref_list */
  while (fields_ref_list) 
    fields_ref_list = gasql_xml_remove_ref_list(fields_ref_list, 
						fields_ref_list);
  return ok;
}

/****************************************************************************/
/*                                                                          */
/* Object common to memory representation of the DB: SqlMemItems            */
/* items such as tables, fields, sequences, links inherit this              */
/* object                                                                   */
/*                                                                          */
/****************************************************************************/

static void      sql_mem_item_class_init      (SqlMemItemClass *class);
static void      sql_mem_item_init            (SqlMemItem *mi);
static void      sql_mem_item_destroy         (GtkObject *object);

/* signals */
enum {
  LOAD_FAULT,
  LAST_SIGNAL
};

static gint sql_mem_item_signals[LAST_SIGNAL] = {0}; 


GtkType sql_mem_item_get_type(void) 
{
  static GtkType f_type = 0;
  
  if (!f_type)
    {
      GtkTypeInfo f_info =
      {
        "Sql_Mem_Item",
        sizeof (SqlMemItem),	
        sizeof (SqlMemItemClass),
        (GtkClassInitFunc) sql_mem_item_class_init,
        (GtkObjectInitFunc) sql_mem_item_init,
        (GtkArgSetFunc) NULL,
        (GtkArgGetFunc) NULL
      };
      
      f_type = gtk_type_unique (gtk_object_get_type (), &f_info);
    }
  
  return f_type;   
}

static void sql_mem_item_class_init(SqlMemItemClass *class)
{
  GtkObjectClass *object_class;
  
  object_class = (GtkObjectClass*) class;   

  sql_mem_item_signals[LOAD_FAULT] =
    gtk_signal_new ("load_fault",
                    GTK_RUN_FIRST,
                    object_class->type,
                    GTK_SIGNAL_OFFSET (SqlMemItemClass, load_fault),
                    gtk_signal_default_marshaller, GTK_TYPE_NONE, 0);
  gtk_object_class_add_signals (object_class, sql_mem_item_signals, 
				LAST_SIGNAL);
  class->load_fault = NULL;

  object_class->destroy = sql_mem_item_destroy;
}

static void sql_mem_item_init(SqlMemItem *mi)
{
  mi->updated = FALSE;
}


GtkObject* sql_mem_item_new(void)
{
  GtkObject *obj;

  obj = gtk_type_new(sql_mem_item_get_type());
  return obj;
}

static void sql_mem_item_destroy(GtkObject *object)
{
  GtkObject *parent_class = NULL;

  parent_class = gtk_type_class (gtk_object_get_type());
  g_return_if_fail (object != NULL);
  g_return_if_fail (IS_SQL_MEM_ITEM (object));

  /* does not destroy anything here, just a link to the parent object */

  /* for the parent class */
  if (GTK_OBJECT_CLASS (parent_class)->destroy)
    (* GTK_OBJECT_CLASS (parent_class)->destroy) (object); 
}


/****************************************************************************/
/*                                                                          */
/* Object representing a table                                              */
/*                                                                          */
/****************************************************************************/

static void      sql_mem_table_class_init      (SqlMemTableClass *class);
static void      sql_mem_table_init            (SqlMemTable *table);
static void      sql_mem_table_destroy         (GtkObject *object);
static void      sql_mem_table_cb              (GtkObject *obj, gpointer data);

enum {
  TAB_FIELD_CREATED,
  TAB_FIELD_DROPPED,
  TAB_COMMENTS_CHANGED,
  LASTTA_SIGNAL
};

static gint sql_mem_table_signals[LASTTA_SIGNAL] = { 0, 0, 0 }; 


GtkType sql_mem_table_get_type(void) 
{
  static GtkType f_type = 0;
  
  if (!f_type)
    {
      GtkTypeInfo f_info =
      {
        "Sql_Mem_Table",
        sizeof (SqlMemTable),	
        sizeof (SqlMemTableClass),
        (GtkClassInitFunc) sql_mem_table_class_init,
        (GtkObjectInitFunc) sql_mem_table_init,
        (GtkArgSetFunc) NULL,
        (GtkArgGetFunc) NULL
      };
      
      f_type = gtk_type_unique (sql_mem_item_get_type (), &f_info);
    }
  
  return f_type;   
}

static void sql_mem_table_class_init(SqlMemTableClass *class)
{
  GtkObjectClass *object_class;
  
  object_class = (GtkObjectClass*) class;   

  sql_mem_table_signals[TAB_FIELD_CREATED] =
    gtk_signal_new ("field_created",
                    GTK_RUN_FIRST,
                    object_class->type,
                    GTK_SIGNAL_OFFSET (SqlMemTableClass, field_created),
                    gtk_marshal_NONE__POINTER, GTK_TYPE_NONE, 1, 
		    GTK_TYPE_POINTER);

  sql_mem_table_signals[TAB_FIELD_DROPPED] =
    gtk_signal_new ("field_dropped",
                    GTK_RUN_FIRST,
                    object_class->type,
                    GTK_SIGNAL_OFFSET (SqlMemTableClass, field_dropped),
                    gtk_marshal_NONE__POINTER, GTK_TYPE_NONE, 1, 
		    GTK_TYPE_POINTER);
  sql_mem_table_signals[TAB_COMMENTS_CHANGED] =
    gtk_signal_new ("comments_changed",
                    GTK_RUN_FIRST,
                    object_class->type,
                    GTK_SIGNAL_OFFSET (SqlMemTableClass, comments_changed),
                    gtk_signal_default_marshaller, GTK_TYPE_NONE, 0);

  gtk_object_class_add_signals (object_class, sql_mem_table_signals, 
				LASTTA_SIGNAL);

  object_class->destroy = sql_mem_table_destroy;
  class->field_created = NULL;
  class->field_dropped = NULL;
  class->comments_changed = NULL;
}

static void sql_mem_table_init(SqlMemTable *table)
{
  table->name = NULL;
  table->parents = NULL;
  table->fields = NULL;
  table->comments = NULL;
  table->is_user_comments = FALSE;
  table->is_view = FALSE;
}


GtkObject* sql_mem_table_new(void)
{
  GtkObject *obj;

  obj = gtk_type_new(sql_mem_table_get_type());
  return obj;
}

static void sql_mem_table_destroy(GtkObject *object)
{
  SqlMemTable *table;
  SqlMemItem *parent_class = NULL;
  GSList *list;

  parent_class = gtk_type_class (sql_mem_item_get_type());
  g_return_if_fail (object != NULL);
  g_return_if_fail (IS_SQL_MEM_TABLE (object));

  table = SQL_MEM_TABLE(object); 
  /* we do free the parent GSList 
     (only the lists, not the objects refered to) */
  g_slist_free(table->parents);
  table->parents = NULL;
  /* we do free the fields GSList and destroy every field */
  list = table->fields;
  while (list) {
    gtk_object_destroy(GTK_OBJECT(list->data));
    list = g_slist_next(list);
  }
  g_slist_free(table->fields);
  table->fields = NULL;
  /* we free the name */
  if (table->name)
    g_free(table->name);
  table->name = NULL;

  /* free comments */
  if (table->comments)
    g_free(table->comments);
  table->comments = NULL;
  
  /* for the parent class */
  if (GTK_OBJECT_CLASS (parent_class)->destroy)
    (* GTK_OBJECT_CLASS (parent_class)->destroy) (object); 
}

/* this callback is called whenever a field emits the signal "load_fault" */
static void sql_mem_table_cb(GtkObject *obj, gpointer data)
{
#ifdef debug_signal
  g_print(">> 'LOAD_FAULT' from sqldb->sql_mem_table_cb\n");
#endif
  gtk_signal_emit(GTK_OBJECT(data), sql_mem_item_signals[LOAD_FAULT]);
#ifdef debug_signal
  g_print("<< 'LOAD_FAULT' from sqldb->sql_mem_table_cb\n");
#endif
}

void sql_mem_table_set_comments(SqlMemTable *table, 
				gASQL_Main_Config *conf,
				gchar *comments,
				gboolean is_user_comments)
{
  if (is_user_comments ||
      (!table->is_user_comments && !is_user_comments))
    {
      if (table->comments)
	g_free(table->comments);
      table->comments = g_strdup(comments);
      table->is_user_comments = is_user_comments;
      if (conf)
	conf->save_up_to_date = FALSE;
      gtk_signal_emit(GTK_OBJECT(table),  
		      sql_mem_table_signals[TAB_COMMENTS_CHANGED]);
    }

  if (*(table->comments) == '\0')
    {
      g_free(table->comments);
      table->comments = NULL;
      table->is_user_comments = FALSE;
    }
}

/****************************************/
/*                                      */
/* Fields manipulation                  */
/*                                      */
/****************************************/

SqlMemField* sql_mem_table_find_field_by_name(SqlMemTable *t, gchar *name)
{
  GSList *list = t->fields;

  while (list && (strcmp(name, SQL_MEM_FIELD(list->data)->name))) {
    list = g_slist_next(list);
  }
  if (list) 
    return SQL_MEM_FIELD(list->data);
  else
    return NULL;  
}

static void sql_mem_table_load_fields (SqlMemTable *t, SqlAccess *srv, 
				       SqlDb *db)
{
  gint i;
  SqlMemField *field;
  GSList *list, *list_hold;
  Gda_Recordset* rs;
  Gda_Field* gfield;
  gchar *str;

  rs = gda_connection_open_schema(GDA_CONNECTION(srv),
  				  GDA_Connection_GDCN_SCHEMA_COLS,
				  GDA_Connection_OBJECT_NAME, t->name,
				  GDA_Connection_no_CONSTRAINT);
  if (rs)
    {
      gda_recordset_move(rs, 1, 0);
      i = 0;
      while (!gda_recordset_eof(rs))
	{
	  gfield = gda_recordset_field_idx(rs, 0);
	  str = gda_stringify_value(NULL, 0, gfield);
	  field = sql_mem_table_find_field_by_name(t, str);
	  g_free(str);
	  if (field)
	    sql_mem_field_load_contents(srv, field, rs);
	  else
	    {
	      field = SQL_MEM_FIELD(sql_mem_field_new());
	      gtk_signal_connect(GTK_OBJECT(field), "load_fault",
				 GTK_SIGNAL_FUNC(sql_mem_table_cb), t);
	      t->fields = g_slist_insert(t->fields, field, i);
	      /* updating the field's contents */
	      sql_mem_field_load_contents(srv, field, rs);
#ifdef debug_signal
	      g_print(">> 'TAB_FIELD_CREATED' from sql_mem_table_load_fields\n");
#endif
	      gtk_signal_emit(GTK_OBJECT(t), 
			      sql_mem_table_signals[TAB_FIELD_CREATED],
			      field);
#ifdef debug_signal
	      g_print("<< 'TAB_FIELD_CREATED' from sql_mem_table_load_fields\n");
#endif
	    }

	  /* mark the field as updated */
	  SQL_MEM_ITEM(field)->updated = TRUE;

	  gda_recordset_move_next(rs);
	  i++;
	}
      gda_recordset_free(rs); 

      /* 2nd step: destroy fields that do not exist anymore */
      list = t->fields;
      while (list) {
	if (! (SQL_MEM_ITEM(list->data)->updated)) {
	  GtkObject *obj;
	  /* we want to emit the "field_dropped" signal with the field
	     structure still existing, but not anymore in the table's list
	     of fields. It is destroyed after the signal has been handled 
	     correctly */
	  obj = GTK_OBJECT(list->data);

	  /* we want to remove any link to that field here, will be done
	     by the SqlDb object at reception of the "field_dropped" signal 
	     from SqlMemTable */
	  
	  /* removing any link to or from that field, 
	     and any sequence link to that
	     field */
	  sql_db_delete_seq_field_link(db, NULL, SQL_MEM_FIELD(obj));
	  sql_db_delete_field_field_link(db, NULL, SQL_MEM_FIELD(obj));
	  sql_db_delete_field_field_link(db, SQL_MEM_FIELD(obj), NULL);  

	  list_hold = g_slist_next(list);
	  t->fields = g_slist_remove_link(t->fields, list);
	  list = list_hold;
#ifdef debug_signal
	  g_print(">> 'TAB_FIELD_DROPPED' from sql_mem_table_load_fields\n");
#endif
	  gtk_signal_emit(GTK_OBJECT(t), 
			  sql_mem_table_signals[TAB_FIELD_DROPPED],
			  obj);
#ifdef debug_signal
	  g_print("<< 'TAB_FIELD_DROPPED' from sql_mem_table_load_fields\n");
#endif
	  gtk_object_destroy(obj);
	}
	else {
	  SQL_MEM_ITEM(list->data)->updated = FALSE;
	  list = g_slist_next(list);
	}
      }
    }
  
}


/****************************************/
/*                                      */
/* Dumping the fields and structure     */
/*                                      */
/****************************************/

#ifdef debug
void sql_mem_table_dump_fields (SqlMemTable *t)
{
  GSList *list;
  
  list = t->fields;
  g_print("\tField name                  Type Length  Other\n");
  while(list) {
    g_print("\t+ ");
    sql_mem_field_dump(SQL_MEM_FIELD(list->data));
    list = g_slist_next(list);
  }
}
#endif

void sql_mem_table_dump_as_graph(SqlMemTable *t, FILE *st)
{
  GSList *list = t->fields;
  fprintf(st, "node: { title:\"table:%s\"\n", t->name);
  fprintf(st, "        label: \"");
  
  /* this table's fields */
  fprintf(st, "\\fb\\fu%s\\fn", t->name);
  while (list) {
    fprintf(st, "\\n%s", SQL_MEM_FIELD(list->data)->name);
    list = g_slist_next(list);
  }
  fprintf(st, "\"\n      }\n");

  /* links from the parent tables */
  list = t->parents;
  while (list) {
    fprintf(st, "edge: { sourcename: \"table:%s\" targetname: \"table:%s\""
	    " linestyle: dashed }\n", 
	    SQL_MEM_TABLE(list->data)->name, t->name);
    list = g_slist_next(list);
  }
}

/* creates the SqlMemTable. 
   the table MUST NOT exist before. */
GSList *sql_mem_table_load_from_xml_node(SqlDb *db,
					 xmlNodePtr tabletree, 
					 GSList *ref_list)
{
  gchar *txt;
  xmlNodePtr node;
  GSList *ret_list = ref_list;
  SqlMemField *field;
  SqlDataType *type;
  SqlMemTable *table, *t2;
  gboolean used;

  txt = gasql_xml_clean_string(xmlNodeGetContent(tabletree));

  table = SQL_MEM_TABLE(sql_mem_table_new());
  db->tables = g_slist_append(db->tables, table);    
  table->name = txt;

  txt = xmlGetProp(tabletree, "user_comments");
  used = (txt && (*txt == 't')) ? TRUE : FALSE;
  txt = xmlGetProp(tabletree, "comments");
  if (txt)
    sql_mem_table_set_comments(table, NULL, txt, used);

  txt = xmlGetProp(tabletree, "owner");
  if (txt)
    table->owner = (gchar *)((find_user_name(db, txt))->data);
  txt = xmlGetProp(tabletree, "is_view");
  table->is_view = (txt && (*txt == 't')) ? TRUE : FALSE;

  gtk_signal_connect(GTK_OBJECT(table), "load_fault",
		     GTK_SIGNAL_FUNC(sql_db_cb), db);
  gtk_signal_connect(GTK_OBJECT(table), "field_created",
		     GTK_SIGNAL_FUNC(sql_db_catch_field_create_cb), db);
  gtk_signal_connect(GTK_OBJECT(table), "field_dropped",
		     GTK_SIGNAL_FUNC(sql_db_catch_field_drop_cb), db);
#ifdef debug_signal
  g_print(">> 'TABLE_CREATED' from sql_mem_table_load_from_xml_node\n");
#endif
  gtk_signal_emit(GTK_OBJECT(db), sql_db_signals[TABLE_CREATED], table);
#ifdef debug_signal
  g_print("<< 'TABLE_CREATED' from sql_mem_table_load_from_xml_node\n");
#endif

  node = tabletree->childs;
  while(node) 
    {
      used = FALSE;
      if (!strcmp(node->name, "parent")) 
	{
	  txt = gasql_xml_clean_string(xmlNodeGetContent(node));
	  if (txt) 
	    {
	      t2 = sql_db_find_table_by_name(db, txt);
	      if (!t2) 
		{
		  table->parents = g_slist_append(table->parents, NULL);
		  ret_list = gasql_xml_append_ref_list(ret_list, txt, NULL,
				     &(g_slist_last(table->parents)->data));
		}
	      else 
		if (!g_slist_find(table->parents, t2))
		  table->parents = g_slist_append(table->parents, t2);
	      /*g_free(txt); NO because txt put in a gasql_xml_ref_list */
	    }
	  used = TRUE;
	}

      if (!used && !strcmp(node->name, "field")) 
	{
	  txt = gasql_xml_clean_string(xmlNodeGetContent(node));
	  if (txt) 
	    {
	      if (!(field = sql_mem_table_find_field_by_name(table, txt))) 
		{
		  field = SQL_MEM_FIELD(sql_mem_field_new());
		  gtk_signal_connect(GTK_OBJECT(field), "load_fault",
				     GTK_SIGNAL_FUNC(sql_mem_table_cb), table);
		  table->fields = g_slist_append(table->fields, field);
#ifdef debug_signal
		  g_print(">> 'TAB_FIELD_CREATED' from "
			  "sql_mem_table_load_from_xml_node\n");
#endif
		  gtk_signal_emit(GTK_OBJECT(table), 
				  sql_mem_table_signals[TAB_FIELD_CREATED], 
				  field);
#ifdef debug_signal
		  g_print("<< 'TAB_FIELD_CREATED' from "
			  "sql_mem_table_load_from_xml_node\n");
#endif
		}
	      if (field->name)
		g_free(field->name);
	      field->name = txt;
	      /*g_free(txt); NO because part of field object */
	      txt = xmlGetProp(node, "length");
	      if (txt)
		field->length = atoi(txt);
	      else
		field->length = -1;
	      txt = xmlGetProp(node, "notnull");
	      if (txt)
		switch (*txt) 
		  {
		  case 't':
		    field->null_allowed = FALSE;
		    break;
		  default:
		    field->null_allowed = TRUE;
		    break;
		  }
	      txt = xmlGetProp(node, "iskey");
	      if (txt)
		switch (*txt) 
		  {
		  case 't':
		    field->is_key = TRUE;
		    break;
		  default:
		    field->is_key = FALSE;
		    break;
		  }
	      txt = xmlGetProp(node, "type");
	      if (txt) 
		{
		  type = sql_access_get_type_from_name(db->srv, txt);
		  if (!type)
		    type = sql_access_get_type_from_name(db->srv, "unknown");
		  field->type = type;
		}
	      txt = xmlGetProp(node, "default");
	      if (txt) 
		field->default_val = g_strdup(txt);
	    }
	  used = TRUE;
	}
      
      node = node->next;
    }
  
  return ret_list;
}


/****************************************************************************/
/*                                                                          */
/* Object representing a field                                              */
/*                                                                          */
/****************************************************************************/

static void      sql_mem_field_class_init      (SqlMemFieldClass *class);
static void      sql_mem_field_init            (SqlMemField *field);
static void      sql_mem_field_destroy         (GtkObject *object);

/* No signals */
GtkType sql_mem_field_get_type(void) 
{
  static GtkType f_type = 0;
  
  if (!f_type)
    {
      GtkTypeInfo f_info =
      {
        "Sql_Mem_Field",
        sizeof (SqlMemField),	
        sizeof (SqlMemFieldClass),
        (GtkClassInitFunc) sql_mem_field_class_init,
        (GtkObjectInitFunc) sql_mem_field_init,
        (GtkArgSetFunc) NULL,
        (GtkArgGetFunc) NULL
      };
      
      f_type = gtk_type_unique (sql_mem_item_get_type (), &f_info);
    }
  
  return f_type;   
}

static void sql_mem_field_class_init(SqlMemFieldClass *class)
{
  GtkObjectClass *object_class;
  
  object_class = (GtkObjectClass*) class;   

  object_class->destroy = sql_mem_field_destroy;
}

static void sql_mem_field_init(SqlMemField *field)
{
  field->name = NULL;
  field->type = NULL;
  field->type = NULL;
  field->length = 0;
  field->null_allowed = TRUE;
  field->is_key = FALSE;
  field->default_val = NULL;
}


GtkObject* sql_mem_field_new(void)
{
  GtkObject *obj;

  obj = gtk_type_new(sql_mem_field_get_type());
  return obj;
}

static void sql_mem_field_destroy(GtkObject *object)
{
  SqlMemField *field;
  SqlMemItem *parent_class = NULL;

  parent_class = gtk_type_class (sql_mem_item_get_type());
  g_return_if_fail (object != NULL);
  g_return_if_fail (IS_SQL_MEM_FIELD (object));

  field = SQL_MEM_FIELD(object); 

  /* we free the name */
  if (field->name)
    g_free(field->name);
  field->name = NULL;
  field->type = NULL;
  if (field->default_val) {
    g_free(field->default_val);
    field->default_val = NULL;
  }
  
  /* for the parent class */
  if (GTK_OBJECT_CLASS (parent_class)->destroy)
    (* GTK_OBJECT_CLASS (parent_class)->destroy) (object); 
}


/****************************************/
/*                                      */
/* Managing the contents of the field   */
/*                                      */
/****************************************/

static void sql_mem_field_load_contents(SqlAccess *srv, SqlMemField *field, 
					Gda_Recordset* recset)
{
  gchar *str;
  Gda_Field *gfield;

  if (field->name) 
    g_free(field->name);
  gfield = gda_recordset_field_idx(recset, 0);
  field->name = gda_stringify_value(NULL, 0, gfield);;

  gfield = gda_recordset_field_idx(recset, 1);
  str = gda_stringify_value(NULL, 0, gfield);
  field->type = sql_access_get_type_from_name(srv, str);
  g_free(str);
  if (!field->type)
    field->type = sql_access_get_type_from_name(srv, "varchar");

  gfield = gda_recordset_field_idx(recset, 2);
  str = gda_stringify_value(NULL, 0, gfield);
  field->length = atoi(str);
  g_free(str);

  gfield = gda_recordset_field_idx(recset, 4);
  if (gda_field_boolean(gfield))
    field->null_allowed = TRUE;
  else
    field->null_allowed = FALSE;
  
  /* FIXME: field->is_key needs to be set */
  gfield = gda_recordset_field_idx(recset, 5);
  if (gda_field_boolean(gfield))
    field->is_key = TRUE;
  else
    field->is_key = FALSE;

  /* Default value */
  gfield = gda_recordset_field_idx(recset, 6);
  if (field->default_val)
    {
      g_free(field->default_val);
      field->default_val = NULL;
    }
  if (gfield->real_value->_u.v._u.lvc)
    {
      str = gda_stringify_value(NULL, 0, gfield);
      field->default_val = str;
    }
}


/****************************************/
/*                                      */
/* Dumping the field                    */
/*                                      */
/****************************************/
#ifdef debug
void sql_mem_field_dump(SqlMemField *field)
{
  /* prints all on one line */
/*  gchar*       sql_access_get_data_type (SqlAccess *srv, gchar *oid);*/

  g_print("%-20s %-10s(%d)\t%3d   ", field->name, 
	  field->type->sqlname, field->type->server_type, field->length);
  if (field->null_allowed)
    g_print("NULL allowed ");
  else
    g_print("NOT NULL     ");
  if (field->is_key)
    g_print("Key");
  if (field->default_val)
     g_print("Defaults to %s", field->default_val);
  g_print("\n");
}
#endif



/****************************************************************************/
/*                                                                          */
/* Object representing a sequence                                           */
/*                                                                          */
/****************************************************************************/

static void      sql_mem_seq_class_init      (SqlMemSeqClass *class);
static void      sql_mem_seq_init            (SqlMemSeq *seq);
static void      sql_mem_seq_destroy         (GtkObject *object);


/* No signals */
GtkType sql_mem_seq_get_type(void) 
{
  static GtkType f_type = 0;
  
  if (!f_type)
    {
      GtkTypeInfo f_info =
      {
        "Sql_Mem_Seq",
        sizeof (SqlMemSeq),	
        sizeof (SqlMemSeqClass),
        (GtkClassInitFunc) sql_mem_seq_class_init,
        (GtkObjectInitFunc) sql_mem_seq_init,
        (GtkArgSetFunc) NULL,
        (GtkArgGetFunc) NULL
      };
      
      f_type = gtk_type_unique (sql_mem_item_get_type (), &f_info);
    }
  
  return f_type;   
}

static void sql_mem_seq_class_init(SqlMemSeqClass *class)
{
  GtkObjectClass *object_class;
  
  object_class = (GtkObjectClass*) class;   

  object_class->destroy = sql_mem_seq_destroy;
}

static void sql_mem_seq_init(SqlMemSeq *seq)
{
  seq->name = NULL;
  seq->field_links = NULL;
  seq->comments = NULL;
  seq->owner = NULL;
}


GtkObject* sql_mem_seq_new(void)
{
  GtkObject *obj;

  obj = gtk_type_new(sql_mem_seq_get_type());
  return obj;
}

static void sql_mem_seq_destroy(GtkObject *object)
{
  SqlMemSeq *seq;
  SqlMemItem *parent_class = NULL;

  parent_class = gtk_type_class (sql_mem_item_get_type());
  g_return_if_fail (object != NULL);
  g_return_if_fail (IS_SQL_MEM_SEQ (object));
  
  seq = SQL_MEM_SEQ(object); 

  /* field links destroying */
  if (seq->field_links) {
    g_slist_free(seq->field_links);
    seq->field_links = NULL;
  }

  /* name destroying */
  if (seq->name)
    g_free(seq->name);
  seq->name = NULL;

  /* comments destroying */
  if (seq->comments)
    g_free(seq->comments);
  seq->comments = NULL;
  
  /* for the parent class */
  if (GTK_OBJECT_CLASS (parent_class)->destroy)
    (* GTK_OBJECT_CLASS (parent_class)->destroy) (object); 
}


/****************************************/
/*                                      */
/* Dumping the sequence                 */
/*                                      */
/****************************************/

void sql_mem_seq_dump_as_graph(SqlMemSeq *seq, SqlDb *db, FILE *st)
{
  GSList *list;

  /* sequence node */
  fprintf(st, "node: { title:\"seq:%s\"\n", seq->name);
  fprintf(st, "        label: \"%s\"\n", seq->name);
  fprintf(st, "        shape: ellipse\n");
  fprintf(st, "        color: pink\n");
  fprintf(st, "      }\n");

  /* links to fields */
  list = seq->field_links;
  while (list) {
    fprintf(st, "edge: { sourcename: \"table:%s\" targetname: \"seq:%s\""
	    " anchor: %d color: pink arrowstyle: none backarrowstyle: solid "
	    "backarrowsize: 10}\n", 
	    (sql_db_find_table_from_field(db, 
					  SQL_MEM_FIELD(list->data)))->name,
	    seq->name, 
	    sql_db_get_field_order(db, NULL,  
				   SQL_MEM_FIELD(list->data))+1);    

    list = g_slist_next(list);
  }
}


/* creates the SqlMemSeq.
   the SqlMemSeq object MUST not exist before. */
GSList *sql_mem_seq_load_from_xml_node(SqlDb *db,
				       xmlNodePtr tabletree, 
				       GSList *ref_list)
{
  gchar *txt, *txt2;
  xmlNodePtr node;
  GSList *ret_list = ref_list;
  SqlMemSeq *seq;
  SqlMemField *field;
  gboolean used;

  txt = gasql_xml_clean_string(xmlNodeGetContent(tabletree));
  
  seq = SQL_MEM_SEQ(sql_mem_seq_new());
  gtk_signal_connect(GTK_OBJECT(seq), "load_fault",
			   GTK_SIGNAL_FUNC(sql_db_cb), db);
  db->sequences = g_slist_append(db->sequences, seq);    
  seq->name = txt;
  txt = xmlGetProp(tabletree, "comments");
  if (txt)
    seq->comments = g_strdup(txt);

  txt = xmlGetProp(tabletree, "owner");
  if (txt)
    seq->owner = (gchar *)((find_user_name(db, txt))->data);
  
  node = tabletree->childs;
  while(node) {
    used = FALSE;
    if (!strcmp(node->name, "link")) {
      txt = xmlGetProp(node, "to_table");
      txt2 = xmlGetProp(node, "to_field");
      if (txt && txt2) {
	field = sql_db_find_field_from_names(db, txt, txt2);
	if (!field) {
	  seq->field_links = g_slist_append(seq->field_links, NULL);
	  ret_list = gasql_xml_append_ref_list(ret_list, g_strdup(txt),
					       g_strdup(txt2),
		     &(g_slist_last(seq->field_links)->data));
	}
	else 
	  if (!g_slist_find(seq->field_links, field))
	    seq->field_links = g_slist_append(seq->field_links, field);
		
      }
      used = TRUE;
    }
    node = node->next;
  }

  return ret_list;
}


/****************************************************************************/
/*                                                                          */
/* Object representing a field link                                         */
/*                                                                          */
/****************************************************************************/

static void      sql_mem_flink_class_init      (SqlMemFlinkClass *class);
static void      sql_mem_flink_init            (SqlMemFlink *flink);
static void      sql_mem_flink_destroy         (GtkObject *object);


/* No signals */
GtkType sql_mem_flink_get_type(void) 
{
  static GtkType f_type = 0;
  
  if (!f_type)
    {
      GtkTypeInfo f_info =
      {
        "Sql_Mem_Flink",
        sizeof (SqlMemFlink),	
        sizeof (SqlMemFlinkClass),
        (GtkClassInitFunc) sql_mem_flink_class_init,
        (GtkObjectInitFunc) sql_mem_flink_init,
        (GtkArgSetFunc) NULL,
        (GtkArgGetFunc) NULL
      };
      
      f_type = gtk_type_unique (sql_mem_item_get_type (), &f_info);
    }
  
  return f_type;   
}

static void sql_mem_flink_class_init(SqlMemFlinkClass *class)
{
  GtkObjectClass *object_class;
  
  object_class = (GtkObjectClass*) class;   

  object_class->destroy = sql_mem_flink_destroy;
}

static void sql_mem_flink_init(SqlMemFlink *flink)
{
  flink->name = NULL;
  flink->from = NULL;
  flink->to = NULL;
}


GtkObject* sql_mem_flink_new(void)
{
  GtkObject *obj;

  obj = gtk_type_new(sql_mem_flink_get_type());
  return obj;
}

static void sql_mem_flink_destroy(GtkObject *object)
{
  SqlMemFlink *flink;
  SqlMemItem *parent_class = NULL;

  parent_class = gtk_type_class (sql_mem_item_get_type());
  g_return_if_fail (object != NULL);
  g_return_if_fail (IS_SQL_MEM_FLINK (object));
  
  flink = SQL_MEM_FLINK(object); 

  /* name destroying */
  if (flink->name)
    g_free(flink->name);
  flink->name = NULL;
  
  /* for the parent class */
  if (GTK_OBJECT_CLASS (parent_class)->destroy)
    (* GTK_OBJECT_CLASS (parent_class)->destroy) (object); 
}


/****************************************/
/*                                      */
/* Dumping the link                     */
/*                                      */
/****************************************/

void sql_mem_flink_dump_as_graph(SqlMemFlink *flink, SqlDb *db, FILE *st)
{
  static guint num = 0; /* numbering of the created nodes */

  /* creating a new node */
  if (flink->name) 
    fprintf(st, "node: { title:\"fl%d\"\n"
	    "	label:\"%s\"\n"
	    "	borderwidth: 0\n"
	    "	color: aquamarine\n"
	    "      }\n", num, flink->name);
  else 
    fprintf(st, "node: { title:\"fl%d\"\n"
	    "	label:\"N/A\"\n"
	    "	shape: ellipse\n"
	    "	color: aquamarine\n"
	    "      }\n", num);

  /* creating the two edges */
  fprintf(st, "edge: { sourcename: \"table:%s\" targetname: \"fl%d\""
	  " anchor: %d arrowstyle: none backarrowstyle: solid "
	  "backarrowsize: 10 priority: 50}\n", 
	  (sql_db_find_table_from_field(db, flink->to))->name, num,
	  sql_db_get_field_order(db, NULL, flink->to)+1);

  fprintf(st, "edge: { sourcename: \"table:%s\" targetname: \"fl%d\""
	  " anchor: %d arrowstyle: none priority: 50}\n", 
	  (sql_db_find_table_from_field(db, flink->from))->name, num,
	  sql_db_get_field_order(db, NULL, flink->from)+1);

  /* prepares for the next one */
  num++;
}

/* creates the SqlMemFlink object.
   This object MUST NOT exist before. */
GSList *sql_mem_flink_load_from_xml_node(SqlDb *db,
					 xmlNodePtr tree, 
					 GSList *ref_list)
{
  gchar *ftabtxt, *ffieldtxt, *ttabtxt, *tfieldtxt;
  xmlNodePtr node;
  GSList *ret_list = ref_list;
  SqlMemFlink *flink;
  SqlMemField *ffield, *tfield;

  /* here we load only ONE field link */
  node = tree;
  if (!strcmp(node->name, "link")) {
    ftabtxt = xmlGetProp(node, "from_table");
    ffieldtxt = xmlGetProp(node, "from_field");
    if (ftabtxt && ffieldtxt) 
      ffield = sql_db_find_field_from_names(db, ftabtxt, ffieldtxt);
    else
      ffield = NULL;
    ttabtxt = xmlGetProp(node, "to_table");
    tfieldtxt = xmlGetProp(node, "to_field");
    if (ttabtxt && tfieldtxt) 
      tfield = sql_db_find_field_from_names(db, ttabtxt, tfieldtxt);
    else
      tfield = NULL;
    
    flink = SQL_MEM_FLINK(sql_mem_flink_new());

    if (!ffield || !tfield) {
      db->field_links = g_slist_append(db->field_links, flink);
      if (!ffield) 
	ret_list = gasql_xml_append_ref_list(ret_list, g_strdup(ftabtxt),
					     g_strdup(ffieldtxt),
					     (gpointer) (&(flink->from)));
      else
	flink->from = ffield;
      if (!tfield) 
	ret_list = gasql_xml_append_ref_list(ret_list, g_strdup(ttabtxt),
					     g_strdup(tfieldtxt),
					     (gpointer) (&(flink->to)));
      else
	flink->to = tfield;
    }
    else 
      if (!sql_db_find_flink_from_fields(db, ffield, tfield)) {
	flink->from = ffield;
	flink->to = tfield;
	db->field_links = g_slist_append(db->field_links, flink);
      }
  }
  
  return ret_list;
}
