/* sqlquery.c
 *
 * Copyright (C) 1999,2000 Vivien Malerba
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 */

#include <config.h>
#include "sqlquery.h"

static void      sql_query_class_init      (SqlQueryClass *class);
static void      sql_query_init            (SqlQuery *srv);
static void      sql_query_destroy         (GtkObject *object);

static GtkObject*   sql_query_new_id(gchar *name, gASQL_Main_Config *conf, 
				     guint id, gboolean ref);

static void remove_elt(SqlQuery *q, QueryElement *object); 

static void sql_query_table_dropped_cb(GtkObject *obj, 
				       SqlMemTable *table,
				       SqlQuery *q);
static void sql_query_field_dropped_cb(GtkObject *obj, 
				       SqlMemTable *table,
				       SqlMemField *field,
				       SqlQuery *q);
static void m_changed(SqlQuery *q, gpointer data);

/* get a pointer to the parents to be able to call their destructor */
static GtkObject *parent_class = NULL;

enum {
  ELT_CREATED,
  ELT_DROPPED,
  ELT_MODIF,
  CHANGED,
  NAME_CHANGED,
  AUTO_LINK_ADDED,
  AUTO_LINK_REMOVED,
  MODIF_TABLE_CHANGED,
  QUERY_REFRESHED,
  LAST_SIGNAL
};

static gint sql_query_signals[LAST_SIGNAL] = { 0, 0, 0, 0, 0, 0, 0, 0, 0 }; 


guint sql_query_get_type(void) 
{
  static guint f_type = 0;
  
  if (!f_type)
    {
      GtkTypeInfo f_info =
      {
        "Sql_Query",
        sizeof (SqlQuery),	
        sizeof (SqlQueryClass),
        (GtkClassInitFunc) sql_query_class_init,
        (GtkObjectInitFunc) sql_query_init,
        (GtkArgSetFunc) NULL,
        (GtkArgGetFunc) NULL
      };
      
      f_type = gtk_type_unique (gtk_object_get_type (), &f_info);
    }
  
  return f_type;   
}

static void sql_query_class_init      (SqlQueryClass *class)
{
  GtkObjectClass *object_class;
  
  object_class = (GtkObjectClass*) class;   

  sql_query_signals[ELT_CREATED] =
    gtk_signal_new ("elt_created",
                    GTK_RUN_FIRST,
                    object_class->type,
                    GTK_SIGNAL_OFFSET (SqlQueryClass, elt_created),
                    gtk_marshal_NONE__POINTER, GTK_TYPE_NONE, 1, 
		    GTK_TYPE_POINTER);
  sql_query_signals[ELT_DROPPED] =
    gtk_signal_new ("elt_dropped",
                    GTK_RUN_FIRST,
                    object_class->type,
                    GTK_SIGNAL_OFFSET (SqlQueryClass, elt_dropped),
                    gtk_marshal_NONE__POINTER, GTK_TYPE_NONE, 1, 
		    GTK_TYPE_POINTER);
  sql_query_signals[ELT_MODIF] =
    gtk_signal_new ("elt_modified",
                    GTK_RUN_FIRST,
                    object_class->type,
                    GTK_SIGNAL_OFFSET (SqlQueryClass, elt_modified),
                    gtk_marshal_NONE__POINTER, GTK_TYPE_NONE, 1, 
		    GTK_TYPE_POINTER);
  sql_query_signals[CHANGED] =
    gtk_signal_new ("changed",
                    GTK_RUN_FIRST,
                    object_class->type,
                    GTK_SIGNAL_OFFSET (SqlQueryClass, changed),
                    gtk_signal_default_marshaller, GTK_TYPE_NONE, 0);
  sql_query_signals[NAME_CHANGED] =
    gtk_signal_new ("name_changed",
                    GTK_RUN_FIRST,
                    object_class->type,
                    GTK_SIGNAL_OFFSET (SqlQueryClass, name_changed),
                    gtk_signal_default_marshaller, GTK_TYPE_NONE, 0);
  sql_query_signals[AUTO_LINK_ADDED] =
    gtk_signal_new ("auto_link_added",
                    GTK_RUN_FIRST,
                    object_class->type,
                    GTK_SIGNAL_OFFSET (SqlQueryClass, auto_link_added),
                    gtk_marshal_NONE__POINTER, GTK_TYPE_NONE, 1, 
		    GTK_TYPE_POINTER);
  sql_query_signals[AUTO_LINK_REMOVED] =
    gtk_signal_new ("auto_link_removed",
                    GTK_RUN_FIRST,
                    object_class->type,
                    GTK_SIGNAL_OFFSET (SqlQueryClass, auto_link_removed),
                    gtk_marshal_NONE__POINTER, GTK_TYPE_NONE, 1, 
		    GTK_TYPE_POINTER);
  sql_query_signals[MODIF_TABLE_CHANGED] =
    gtk_signal_new ("modif_table_changed",
                    GTK_RUN_FIRST,
                    object_class->type,
                    GTK_SIGNAL_OFFSET (SqlQueryClass, modif_table_changed),
                    gtk_marshal_NONE__POINTER, GTK_TYPE_NONE, 1, 
		    GTK_TYPE_POINTER);
  sql_query_signals[QUERY_REFRESHED] =
    gtk_signal_new ("query_refreshed",
                    GTK_RUN_FIRST,
                    object_class->type,
                    GTK_SIGNAL_OFFSET (SqlQueryClass, query_refreshed),
                    gtk_signal_default_marshaller, GTK_TYPE_NONE, 0);

  gtk_object_class_add_signals (object_class, sql_query_signals, LAST_SIGNAL);
  class->elt_created = (void (*) (SqlQuery *q, QueryElement *elt)) m_changed;
  class->elt_dropped = (void (*) (SqlQuery *q, QueryElement *elt)) m_changed;
  class->elt_modified = (void (*) (SqlQuery *q, QueryElement *elt)) m_changed;
  class->auto_link_added = (void (*) (SqlQuery *q, QueryLink *ql)) 
    m_changed;
  class->auto_link_removed = (void (*) (SqlQuery *q, QueryLink *ql)) 
    m_changed;
  class->changed = sql_query_refresh_res;
  class->name_changed = NULL;
  class->modif_table_changed = NULL;
  class->query_refreshed = NULL;

  object_class->destroy = sql_query_destroy;
  parent_class = gtk_type_class (gtk_object_get_type());
}

static void sql_query_init(SqlQuery *q)
{
  q->name = NULL;
  q->descr = NULL;
  q->objects = NULL;
  q->select_list = NULL;
  q->group_by_list = NULL;
  q->order_by_list = NULL;
  q->depend_list = NULL;
  q->text_only = FALSE;
  q->text_sql = NULL;
  q->modif_table = NULL;
  q->res = NULL;
  q->reg_users = NULL;
}

GtkObject* sql_query_new_noref(gchar *name, gASQL_Main_Config *conf)
{
  return sql_query_new_id(name, conf, 0, FALSE);
}

GtkObject* sql_query_new(gchar *name, gASQL_Main_Config *conf)
{
  return sql_query_new_id(name, conf, get_id_serial(conf), TRUE);
}

static GtkObject* sql_query_new_id(gchar *name, gASQL_Main_Config *conf, 
				   guint id, gboolean ref)
{
  GtkObject *obj;

  SqlQuery *q;
  obj =  gtk_type_new(sql_query_get_type());
  q = SQL_QUERY(obj);

  q->conf = conf;
  q->id = id; /* next value for the sequence */
  q->name = g_strdup(name);
  
  if (ref) {
    /* adding to the list of queries */
    conf->queries = g_slist_append(conf->queries, q);
    conf->save_up_to_date = FALSE;
  }

  /* connection to signals coming from SqlDb */
  gtk_signal_connect_while_alive(GTK_OBJECT(q->conf->db), "table_dropped",
				 GTK_SIGNAL_FUNC(sql_query_table_dropped_cb), 
				 q, obj);
  gtk_signal_connect_while_alive(GTK_OBJECT(q->conf->db), "field_dropped",
				 GTK_SIGNAL_FUNC(sql_query_field_dropped_cb), 
				 q, obj);
  gtk_signal_connect_while_alive(GTK_OBJECT(q->conf->db), "ff_link_created",
			       GTK_SIGNAL_FUNC(sql_query_auto_links_add_link), 
				 q, obj);
  gtk_signal_connect_while_alive(GTK_OBJECT(q->conf->db), "ff_link_dropped",
			       GTK_SIGNAL_FUNC(sql_query_auto_links_del_link), 
				 q, obj);
  return obj;
}

static void m_changed(SqlQuery *q, gpointer data)
{
#ifdef debug_signal
  g_print(">> 'CHANGED' from sqlquery->m_changed\n");
#endif
  gtk_signal_emit(GTK_OBJECT(q), sql_query_signals[CHANGED]);  
#ifdef debug_signal
  g_print("<< 'CHANGED' from sqlquery->m_changed\n");
#endif  
}

void sql_query_swap_objects(SqlQuery *q, gint row1, gint row2)
{
  GSList *list1, *list2;
  gpointer data;
  gint nrows;

  nrows = g_slist_length(q->objects);
  if ((row1<nrows) && (row2<nrows) && (row1>=0) && (row2>=0)) {
    list1 = g_slist_nth(q->objects, row1);
    list2 = g_slist_nth(q->objects, row2);
    /* swapping list1->data and list2->data */
    data = list1->data;
    list1->data = list2->data;
    list2->data = data;
  }
#ifdef debug_signal
  g_print(">> 'CHANGED' from sql_query_swap_objects\n");
#endif
  gtk_signal_emit(GTK_OBJECT(q), sql_query_signals[CHANGED]);  
#ifdef debug_signal
  g_print("<< 'CHANGED' from sql_query_swap_objects\n");
#endif
}

/* removes recursively everything referencing that object in all the 
   Query object' lists.
   Frees the memory associated to the object passed as argument */
static void remove_elt(SqlQuery *q, QueryElement *object) 
{
  GSList *list;
  QueryElement *qe;

  /* SELECT list */
  q->select_list = g_slist_remove(q->select_list, object);

  /* GROUP BY list */
  q->group_by_list = g_slist_remove(q->group_by_list, object);

  /* ORDER BY list */
  q->order_by_list = g_slist_remove(q->order_by_list, object);

  /* DEPEND list */
  q->depend_list = g_slist_remove(q->depend_list, object);

  /* objects list */
  /* the top level list of elements */
  q->objects = g_slist_remove(q->objects, object);

  /* elements referencing that object */
  list = q->objects;
  while (list) {
    qe = (QueryElement *) (list->data);
    if (qe->type != QUERY_ELT_FIELD) {
      if (g_slist_find(qe->args, object)) {
	remove_elt(q, list->data); /* del the element which refers to object */
	list = q->objects;
      }
      else 
	list = g_slist_next(list);
    }
    else 
      list = g_slist_next(list);
  }
#ifdef debug_signal
  g_print(">> 'ELT_DROPPED' from sqlquery->remove_elt\n");
#endif
  gtk_signal_emit(GTK_OBJECT(q), sql_query_signals[ELT_DROPPED], object);
#ifdef debug_signal
  g_print("<< 'ELT_DROPPED' from sqlquery->remove_elt\n");
#endif

  /* freeing the memory associated to the object */
  switch (object->type) {
  case QUERY_ELT_AGGREGATE:
  case QUERY_ELT_FUNCTION:
    if (object->args)
      g_slist_free(object->args);
  default:
    if (object->name)
      g_free(object->name);
    if (object->print_name)
      g_free(object->print_name);
    break;
  }
  g_free(object);
}

QueryElement* sql_query_get_element(SqlQuery *q, gchar *name)
{
  QueryElement *qe = NULL;
  GSList *list;

  list = q->objects;
  while (list && !qe) {
    if (!strcmp(((QueryElement *)(list->data))->name, name))
      qe = (QueryElement *)(list->data);
    else
      list = g_slist_next(list);
  }

  return qe;
}

void sql_query_set_modif_table(SqlQuery *q, SqlMemTable *t)
{
  if (q->modif_table != t) {
    q->modif_table = t;
#ifdef debug_signal
    g_print(">> 'CHANGED' from sql_query_set_modif_table\n");
#endif
    gtk_signal_emit(GTK_OBJECT(q), sql_query_signals[MODIF_TABLE_CHANGED], t);
#ifdef debug_signal
    g_print("<< 'CHANGED' from sql_query_set_modif_table\n");
#endif
  }
}

/***************************************************************/
/*                                                             */
/* Working with XML                                            */
/*                                                             */
/***************************************************************/

static gboolean foreach_hash_key_in_clean(gchar*	key,
					  gpointer	value,
					  gpointer	user_data)
{
  g_free(key);
  return TRUE;
}

/* before loading a SqlQuery, all the Tables, Fields, Functions, Aggregates
   must already have been loaded. */
SqlQuery* sql_query_build_from_xml_tree(gASQL_Main_Config *conf, 
					xmlNodePtr node)
{
  xmlNodePtr tree, subtree;
  gchar *str, *str2, *str3;
  guint id;
  gboolean ok = TRUE;
  SqlQuery *q;
  SqlMemField *field;
  GHashTable *htable;
  SqlDataType *dtype=NULL;
  GSList *list, *elts_ref = NULL;
  QueryElement *qe;
  
  str = gasql_xml_clean_string_ends(xmlNodeGetContent(node));
  id = atoi(xmlGetProp(node, "id"));
  q = SQL_QUERY(sql_query_new_id(str, conf, id, TRUE));
  g_free(str);
  str = xmlGetProp(node, "descr");
  if (str)
    sql_query_set_name(q, NULL, str);
  str = xmlGetProp(node, "modif_table");
  if (str) {
    SqlMemTable *t;

    t = sql_db_find_table_by_name(conf->db, str);
    if (t)
      q->modif_table = t;
  }
  /* 1st: establish the hash table name<->type of data. 
   and at the same time load all the table fields which will appear in the
   SqlQuery. */
  htable = g_hash_table_new(g_str_hash, g_str_equal);
  tree = node->childs;
  while (tree) {
    /* Field Element */
    if (!strcmp(tree->name, "FieldElt")) {
      str = xmlGetProp(tree, "table");
      str2 = xmlGetProp(tree, "field");
      field = sql_db_find_field_from_names(conf->db, str, str2);
      if (str && str2 && field) {
	dtype = field->type;
	g_hash_table_insert(htable, str2, dtype);
      }
      else {
	tree = NULL;
	ok = FALSE;
      }
    }
    /* Function Element */
    if (!strcmp(tree->name, "FuncElt")) {
      str = xmlGetProp(tree, "type");
      str2 = gasql_xml_clean_string(xmlNodeGetContent(tree));
      if (str && str2) {
	dtype = sql_data_type_get_from_name(conf->srv->data_types, str);
	if (dtype) {
	  g_hash_table_insert(htable, str2, dtype);
	}
	else {
	  ok = FALSE;
	  tree = NULL;
	}
      }		  
    }
    /* if SQL text only, loaded here! */
    if (!strcmp(tree->name, "sqltext")) {
      str2 = xmlNodeGetContent(tree);
      q->text_only = TRUE;
      q->text_sql = str2;
    }

    /* Dependencies list also loaded here */
    if (!strcmp(tree->name, "depend")) {
      SqlMemFlink *link = NULL;
      SqlMemField *field2;

      str = xmlGetProp(tree, "ftable");
      str2 = xmlGetProp(tree, "ffield");
      field = sql_db_find_field_from_names(conf->db, str, str2);
      str = xmlGetProp(tree, "ttable");
      str2 = xmlGetProp(tree, "tfield");
      field2 = sql_db_find_field_from_names(conf->db, str, str2);
      if (field && field2) 
	link = sql_db_find_flink_from_fields(conf->db, field, field2);
      if (link) {
	QueryLink *ql;

	ql = (QueryLink *) g_malloc(sizeof(QueryLink));
	ql->link = link;
	str = xmlGetProp(tree, "useit");
	if (str && (*str=='t'))
	  ql->useit = TRUE;
	else
	  ql->useit = FALSE;
	q->depend_list = g_slist_append(q->depend_list, ql);
      }
    }

    if (tree)
      tree = tree->next;
  }

  /* 2nd: loads the Query Elements which rely on a function or an
     aggregate (only if OK!) */
  if (ok) {
    tree = node->childs;
    while (tree) {
      /* Field Element */
      if (tree && !strcmp(tree->name, "FieldElt")) {
	str = xmlGetProp(tree, "table");
	str2 = xmlGetProp(tree, "field");
	field = sql_db_find_field_from_names(conf->db, str, str2);
	str = xmlGetProp(tree, "print");
	str2 = gasql_xml_clean_string_ends(xmlNodeGetContent(tree));
	if (str && str2 && field) {
	  sql_query_new_elt_field(q, str2, str, field);
	}
	else {
	  tree = NULL;
	  ok = FALSE;
	}
	if (str2)
	  g_free(str2);
      }

      /* Function Element */ 
      if (tree && !strcmp(tree->name, "FuncElt")) {
	SqlDataFunction *dfunc=NULL;
	gint nb, i;
	GSList *list2;
	GSList *elt_names=NULL, *types=NULL; /* to store the names and the 
						data types of the arguments
						for this function */
	subtree = tree->childs;
	while (subtree) { /* building the lists of arguments names and types */
	  if (!strcmp(subtree->name, "ARG")) {
	    str = gasql_xml_clean_string(xmlNodeGetContent(subtree));
	    if (str) 
	      dtype = SQL_DATA_TYPE(g_hash_table_lookup(htable, str));
	    if (str && dtype) {
	      elt_names = g_slist_append(elt_names, str);
	      types = g_slist_append(types, dtype);
	    }
	    else {
	      ok = FALSE;
	      tree = NULL;
	    }
	  }
	  subtree = subtree->next;
	}

	if (ok) { 
	  str3 = xmlGetProp(tree, "func"); /* function's name */
	  dfunc = sql_data_function_get_from_name(conf->srv->data_functions,
						  str3, types);
	  str3 = xmlGetProp(tree, "print");
	  str2 = gasql_xml_clean_string_ends(xmlNodeGetContent(tree));
	  if (str3 && str2 && dfunc) {
	    nb = g_slist_length(types);
	    list = NULL;
	    for (i=0; i<nb; i++)
	      list = g_slist_prepend(list, NULL);
	    qe = sql_query_new_elt_function(q, str2, str3, dfunc, list);
	    g_slist_free(list);
	    list = qe->args;
	    list2 = elt_names;
	    while (list) {
	      elts_ref = gasql_xml_append_ref_list(elts_ref, 
						   (gchar *)(list2->data), 
						   NULL,
						   &(list->data));
	      list2->data = NULL;
	      list = g_slist_next(list);
	      list2 = g_slist_next(list2);
	    }
	  }
	  else {
	    ok = FALSE;
	    tree = NULL;
	  }
	  if (str2)
	    g_free(str2);
	}

	/* always freeing what is remaining */
	while (elt_names) {
	  if (elt_names->data)
	    g_free(elt_names->data); /* freeing the strings */
	  elt_names = g_slist_remove_link(elt_names, elt_names);
	}
	g_slist_free(types);
	types = NULL;
      }

      /* Aggregate Element */
      if (tree && !strcmp(tree->name, "AggElt")) {
	SqlDataAggregate *dagg;
	
	dtype = NULL;
	subtree = tree->childs;
	while (subtree) {
	  if (!strcmp(subtree->name, "ARG")) {
	    /* str is the object's name which is the only argument */
	    str = gasql_xml_clean_string(xmlNodeGetContent(subtree));
	    if (str) 
	      dtype = SQL_DATA_TYPE(g_hash_table_lookup(htable, str));
	    if (!dtype) { 
	      ok = FALSE;
	      tree = NULL;
	    }
	  }
	  subtree = subtree->next;
	}

	if (dtype && ok) {
	  str3 = xmlGetProp(tree, "agg"); /*aggregate's name */
	  dagg = NULL;
	  if (str3)
	    dagg = sql_data_aggregate_get_from_name(conf->srv->data_aggregates,
						    str3, 
						    dtype);
	  str3 = xmlGetProp(tree, "print");
	  str2 = gasql_xml_clean_string_ends(xmlNodeGetContent(tree));
	  if (str3 && str && str2 && dagg) {
	    list = g_slist_append(NULL, NULL); /* creates a list with an
						empty entry to fill later
					        by cross referencing */
	    qe = sql_query_new_elt_aggregate(q, str2, str3, dagg, list);
	    /* DO not free str because used in the following ref! */
	    elts_ref = gasql_xml_append_ref_list(elts_ref, str, NULL,
						 &(qe->args->data));
	    str = NULL;
	    g_slist_free(list);
	  }
	  else {
	    ok = FALSE;
	    tree = NULL;
	  }

	  if (str2)
	    g_free(str2);
	}
	else {
	  ok = FALSE;
	  tree = NULL;
	}

	if (str)
	  g_free(str);
      }
      
      if (tree)
	tree = tree->next;
    }

    /* update the references */
    elts_ref = gasql_xml_update_ref1(elts_ref, q, 
			     GASQL_XML_UPDATE_REF1(sql_query_get_element));
    while (elts_ref) {
      elts_ref = gasql_xml_remove_ref_list(elts_ref, elts_ref);
    }
  }
  /* 3rd: destroying the hash table */
  g_hash_table_foreach_remove(htable, (GHRFunc)foreach_hash_key_in_clean,NULL);
  g_hash_table_destroy(htable);
  

  if (!ok) {
    conf->queries = g_slist_remove(conf->queries, q);
    gtk_object_destroy(GTK_OBJECT(q));
    q = NULL;
  }
  return q;
}

void sql_query_build_xml_tree(SqlQuery *q, gASQL_Main_Config *conf, 
			      xmlDocPtr doc)
{
  xmlNodePtr toptree, tree, subtree;
  gchar *str;
  GSList *list, *arglist;
  QueryElement *qe;
  QueryLink *ql;

  toptree = xmlNewChild(doc->root, NULL, "QUERY", q->name);
  str = g_strdup_printf("%d", q->id);
  xmlSetProp(toptree, "id", str);
  g_free(str);
  xmlSetProp(toptree, "descr", q->descr);
  if (q->modif_table)
    xmlSetProp(toptree, "modif_table", q->modif_table->name);
  /* list of objects */
  list = q->objects;
  while (list) {
    qe = (QueryElement *)(list->data);
    tree = NULL;
    switch (qe->type) {
    case QUERY_ELT_FIELD:
      tree = xmlNewChild(toptree, NULL, "FieldElt", qe->name);
      xmlSetProp(tree, "table", 
		 sql_db_find_table_from_field(conf->db, 
					      SQL_MEM_FIELD(qe->main))->name);
      xmlSetProp(tree, "field", 
		 SQL_MEM_FIELD(qe->main)->name);
      break;
    case QUERY_ELT_AGGREGATE:
      tree = xmlNewChild(toptree, NULL, "AggElt", qe->name);
      xmlSetProp(tree, "agg", SQL_DATA_AGGREGATE(qe->main)->sqlname);
      break;
    case QUERY_ELT_FUNCTION:
      tree = xmlNewChild(toptree, NULL, "FuncElt", qe->name);
      xmlSetProp(tree, "func", SQL_DATA_FUNCTION(qe->main)->sqlname);
      xmlSetProp(tree, "type", 
		 SQL_DATA_FUNCTION(qe->main)->result_type->sqlname);
      break;
    }
    xmlSetProp(tree, "print", qe->print_name);
    xmlSetProp(tree, "print", qe->print_name);
    /* arguments... */
    arglist = qe->args;
    while (arglist) {
      qe = (QueryElement *)(arglist->data);
      subtree = xmlNewChild(tree, NULL, "ARG", qe->name);
      arglist = g_slist_next(arglist);
    }
    list = g_slist_next(list);
  }
  /* list of automatic dependencies */
  list = q->depend_list;
  while (list) {
    ql = (QueryLink *)(list->data);
    tree = xmlNewChild(toptree, NULL, "depend", NULL);
    xmlSetProp(tree, "ftable", 
	       sql_db_find_table_from_field(conf->db, 
					    ql->link->from)->name);
    xmlSetProp(tree, "ffield", ql->link->from->name);
    xmlSetProp(tree, "ttable", 
	       sql_db_find_table_from_field(conf->db, 
					    ql->link->to)->name);
    xmlSetProp(tree, "tfield", ql->link->to->name);
    xmlSetProp(tree, "useit", (ql->useit)?"t":"f");
    list = g_slist_next(list);
  }
  /* if it is an SQL query only */
  if (q->text_only) 
    tree = xmlNewChild(toptree, NULL, "sqltext", q->text_sql);

  /* TOTO: save the remaining things from the SqlQuery object ... */
}

static void sql_query_table_dropped_cb(GtkObject *obj, 
				       SqlMemTable *table,
				       SqlQuery *q)
{
  
  GSList *list;
  QueryElement *qe;

  list = q->objects;
  while (list) {
    qe = (QueryElement *) (list->data);

    if (qe->type == QUERY_ELT_FIELD) {
      if (g_slist_find(table->fields, qe->main)) {
	remove_elt(q, qe);
	list = q->objects;
      }
      else
	list = g_slist_next(list);
    }
    else
      list = g_slist_next(list);
  }
}

static void sql_query_field_dropped_cb(GtkObject *obj, 
				       SqlMemTable *table,
				       SqlMemField *field,
				       SqlQuery *q)
{
  GSList *list;
  QueryElement *qe;

  list = q->objects;
  while (list) {
    qe = (QueryElement *) (list->data);
    
    if (qe->type == QUERY_ELT_FIELD)
      if (SQL_MEM_FIELD(qe->main) == field) {
	remove_elt(q, qe);
	list = q->objects;
      }
      else
	list = g_slist_next(list);   
    else
      list = g_slist_next(list);   
  }
}

static void sql_query_destroy (GtkObject *object)
{
  SqlQuery *q;
  GSList *list;
  guint sigid;

  g_return_if_fail (object != NULL);
  g_return_if_fail (IS_SQL_QUERY (object));

  q = SQL_QUERY(object); 

  /* list of dependencies */
  while (q->depend_list) {
    g_free(q->depend_list->data);
    q->depend_list = g_slist_remove_link(q->depend_list, q->depend_list);
  } 

  /* minor memory allocations */
  if (q->name)
    g_free(q->name);
  if (q->descr)
    g_free(q->descr);

  /* result of the SELECT query */
  list = q->reg_users;
  while (list) {
    sigid = GPOINTER_TO_UINT(gtk_object_get_data(GTK_OBJECT(list->data), 
						 "qsig"));  
    gtk_signal_disconnect(GTK_OBJECT(list->data), sigid);
    list = g_slist_remove_link(list, list);
  }
  q->reg_users = NULL;
  if (q->res) {
    gtk_object_destroy(GTK_OBJECT(q->res));
    q->res = NULL;
  }
  
  /* removing objects */
  while (q->objects) {
    remove_elt(q, (QueryElement *) (q->objects->data));
  } /* q->objects is now NULL and
       rem: other lists are cleared by the previous loop */

  if (GTK_OBJECT_CLASS (parent_class)->destroy)
    (* GTK_OBJECT_CLASS (parent_class)->destroy) (object); 
}


void sql_query_set_name(SqlQuery *q, gchar *name, gchar *descr)
{
  gboolean changed = FALSE;
  
  if (name) {
    if (q->name) {
      changed = strcmp(q->name, name) ? TRUE : FALSE;
      g_free(q->name);
    }
    else
      changed = TRUE;
    q->name = g_strdup(name);
  }
  
  if (descr) {
    if (q->descr) {
      changed = strcmp(q->descr, descr) ? TRUE : changed;
      g_free(q->descr);
    }
    else
      changed = TRUE;
    q->descr = g_strdup(descr);
  }

  if (changed) {
#ifdef debug_signal
    g_print(">> 'NAME_CHANGED' from sql_query_set_name\n");
#endif
    gtk_signal_emit(GTK_OBJECT(q), sql_query_signals[NAME_CHANGED]);
#ifdef debug_signal
    g_print("<< 'NAME_CHANGED' from sql_query_set_name\n");
#endif
  }
}

QueryElement* sql_query_new_elt_field(SqlQuery *q, 
				      gchar *name,
				      gchar *print_name,
				      SqlMemField *field)
{
  QueryElement *elt;

  elt = (QueryElement *) g_malloc(sizeof(QueryElement));
  elt->type = QUERY_ELT_FIELD;
  elt->main = field;
  elt->args = NULL;
  elt->name = g_strdup(name);
  if (*print_name != '\0')
    elt->print_name = g_strdup(print_name);
  else
    elt->print_name = NULL;

  q->objects = g_slist_append(q->objects, elt);
#ifdef debug_signal
  g_print(">> 'ELT_CREATED' from sql_query_new_elt_field\n");
#endif
  gtk_signal_emit(GTK_OBJECT(q), sql_query_signals[ELT_CREATED], elt);
#ifdef debug_signal
  g_print("<< 'ELT_CREATED' from sql_query_new_elt_field\n");
#endif
  return elt;
}

QueryElement* sql_query_new_elt_aggregate(SqlQuery *q,
					  gchar *name, 
					  gchar *print_name,
					  SqlDataAggregate *agg, 
					  GSList *args)
{
  QueryElement *elt;

  elt = (QueryElement *) g_malloc(sizeof(QueryElement));
  elt->type = QUERY_ELT_AGGREGATE;
  elt->main = agg;
  elt->args = g_slist_copy(args);
  elt->name = g_strdup(name);
  if (*print_name != '\0')
    elt->print_name = g_strdup(print_name);
  else
    elt->print_name = NULL;

  q->objects = g_slist_append(q->objects, elt);
#ifdef debug_signal
  g_print(">> 'ELT_CREATED' from sql_query_new_elt_aggregate\n");
#endif
  gtk_signal_emit(GTK_OBJECT(q), sql_query_signals[ELT_CREATED], elt);
#ifdef debug_signal
  g_print("<< 'ELT_CREATED' from sql_query_new_elt_aggregate\n");
#endif
  return elt;
}

QueryElement* sql_query_new_elt_function(SqlQuery *q,
					 gchar *name, 
					 gchar *print_name,
					 SqlDataFunction *func, 
					 GSList *args)
{
  QueryElement *elt;

  elt = (QueryElement *) g_malloc(sizeof(QueryElement));
  elt->type = QUERY_ELT_FUNCTION;
  elt->main = func;
  elt->args = g_slist_copy(args);
  elt->name = g_strdup(name);
  if (*print_name != '\0')
    elt->print_name = g_strdup(print_name);
  else
    elt->print_name = NULL;

  q->objects = g_slist_append(q->objects, elt);
#ifdef debug_signal
  g_print(">> 'ELT_CREATED' from sql_query_new_elt_function\n");
#endif
  gtk_signal_emit(GTK_OBJECT(q), sql_query_signals[ELT_CREATED], elt);
#ifdef debug_signal
  g_print("<< 'ELT_CREATED' from sql_query_new_elt_function\n");
#endif
  if (print_name) {
#ifdef debug_signal
    g_print(">> 'CHANGED' from sql_query_new_elt_function\n");
#endif
    gtk_signal_emit(GTK_OBJECT(q), sql_query_signals[CHANGED]);
#ifdef debug_signal
    g_print("<< 'CHANGED' from sql_query_new_elt_function\n");
#endif
  }
  return elt;
}

void sql_query_del_elt(SqlQuery *q, QueryElement* elt)
{
  remove_elt(q, elt);
}

static gboolean table_is_in_query(SqlQuery *q, SqlMemTable *table)
{
  gboolean found = FALSE;
  GSList *list;
  QueryElement *qe;

  list = q->objects;
  while (list && !found) {
    qe = (QueryElement *)(list->data);
    if (qe->type == QUERY_ELT_FIELD)
      if (sql_db_find_table_from_field(q->conf->db, 
				       SQL_MEM_FIELD(qe->main)) == table)
	found = TRUE;
    list = g_slist_next(list);
  }
  return found;
}

void sql_query_modif_elt(SqlQuery *q, QueryElement* elt,
			 gchar *name, gchar *print_name)
{
  if (name) {
    if (elt->name)
      g_free(elt->name);
    elt->name = g_strdup(name);
  }
  if (print_name) {
    if (elt->print_name)
      g_free(elt->print_name);
    if (*print_name == '\0')
      elt->print_name = NULL;
    else
      elt->print_name = g_strdup(print_name);
  }

  if (name || print_name) {
#ifdef debug_signal
    g_print(">> 'ELT_MODIF' from sql_query_modif_elt\n");
#endif
    gtk_signal_emit(GTK_OBJECT(q), sql_query_signals[ELT_MODIF], elt);
#ifdef debug_signal
    g_print("<< 'ELT_MODIF' from sql_query_modif_elt\n");
#endif
    if (print_name) {
#ifdef debug_signal
      g_print(">> 'CHANGED' from sql_query_modif_elt\n");
#endif
      gtk_signal_emit(GTK_OBJECT(q), sql_query_signals[CHANGED]);    
#ifdef debug_signal
      g_print("<< 'CHANGED' from sql_query_modif_elt\n");
#endif
    }
  }
}

/* returns a proper string SELECT... for the select query */

static gchar* object_to_str(SqlQuery *q, QueryElement *elt)
{
  /* WARNING: Allocates memory */
  gchar *str=NULL;
  GString *string;
  SqlMemTable *table;
  gboolean first;
  GSList *args;

  switch (elt->type) {
  case QUERY_ELT_FIELD:
    table = sql_db_find_table_from_field(q->conf->db,SQL_MEM_FIELD(elt->main));
    str = g_strdup_printf("%s.%s", table->name, 
			  SQL_MEM_FIELD(elt->main)->name);
    break;
  case QUERY_ELT_FUNCTION:
    string = g_string_new(SQL_DATA_FUNCTION(elt->main)->sqlname);
    first = TRUE;
    args = elt->args;
    g_string_append(string, "(");
    while (args) {
      if (first) 
	first = FALSE;
      else 
	g_string_append(string, ", ");

      str = object_to_str(q, (QueryElement *)(args->data));
      g_string_append(string, str);
      g_free(str);
	
      args = g_slist_next(args);
    }
    g_string_append(string, ")");

    str = string->str;
    g_string_free(string, FALSE);
    break;
  case QUERY_ELT_AGGREGATE:
    string = g_string_new(SQL_DATA_AGGREGATE(elt->main)->sqlname);
    args = elt->args;
    g_string_append(string, "(");
    str = object_to_str(q, (QueryElement *)(args->data));
    g_string_append(string, str);
    g_free(str);
    g_string_append(string, ")");

    str = string->str;
    g_string_free(string, FALSE);
    break;
  }

  return str;
}

static GSList* complement_tables_list(SqlQuery *q, 
				      GSList *list, QueryElement *elt)
{
  SqlMemTable *table;
  GSList *nlist, *args;

  nlist = list;
  switch (elt->type) {
  case QUERY_ELT_FIELD:
    table = sql_db_find_table_from_field(q->conf->db,SQL_MEM_FIELD(elt->main));
    if (! g_slist_find(nlist, table))
      nlist = g_slist_append(nlist, table);
    break;
  case QUERY_ELT_FUNCTION:
  case QUERY_ELT_AGGREGATE:
    args = elt->args;
    while (args) {
      nlist = complement_tables_list(q, nlist, 
				     (QueryElement *)(args->data));
      args = g_slist_next(args);
    }
  }

  return nlist;
}

/* returns the WHERE part of a query while taking care of the automatic links
   and any other WHERE constraint imposed to the query 
   The allocated string has to be freed */
static gchar* get_where_part_of_sql(SqlQuery *q, gboolean *any)
{
  GSList *links;
  gboolean first;
  GString *string;
  gchar *retval;
  QueryLink *ql;
  SqlMemTable *fromt, *tot;

  string = g_string_new(""); 
  links = q->depend_list;
  first = TRUE;
  while (links) {
    ql = (QueryLink *) (links->data);
    if (ql->useit) {
      fromt = sql_db_find_table_from_field(q->conf->db, ql->link->from);
      tot = sql_db_find_table_from_field(q->conf->db, ql->link->to);
      if (fromt && tot) { 
	if (first) 
	  first = FALSE;
	else
	  g_string_append(string, "AND ");
	g_string_sprintfa(string, "%s.%s = %s.%s\n",
			  fromt->name, ql->link->from->name,
			  tot->name, ql->link->to->name);
      }
    }
    links = g_slist_next(links);
  }
  retval = string->str;
  g_string_free(string, FALSE);
  *any = !first;
  return retval;
}

gchar* sql_query_get_select_query(SqlQuery *q)
{
  gchar *str, *retval = NULL;
  GString *string;
  GSList *elts, *tables;
  gboolean first, wherearg;

  if (q->text_only)
    retval = g_strdup(q->text_sql);
  else {
    string = g_string_new("SELECT\n");
    /* elements list */
    elts = q->objects;
    first = TRUE;
    while (elts) {
      if (((QueryElement *)(elts->data))->print_name) {
	if (first) 
	  first = FALSE;
	else
	  g_string_append(string, ",\n");

	str = object_to_str(q, (QueryElement *)(elts->data));
	g_string_sprintfa(string, "%s as %s", str,
			  ((QueryElement *)(elts->data))->print_name);
	g_free(str);
      }
      elts = g_slist_next(elts);
    }
    /* from list */
    elts = q->objects;
    tables = NULL;
    while (elts) {
      if (((QueryElement *)(elts->data))->print_name)
	tables = complement_tables_list(q, tables, 
					(QueryElement *)(elts->data));
      elts = g_slist_next(elts);
    }
    first = TRUE;
    elts = tables;
    while (elts) {
      if (first) 
	{
	  first = FALSE;
	  g_string_append(string, "\nFROM\n");
	}
      else
	g_string_append(string, ", ");
      g_string_append(string, SQL_MEM_TABLE(elts->data)->name);
      elts = g_slist_next(elts);
    }
    g_slist_free(tables);

    /* WHERE arguments */
    retval = get_where_part_of_sql(q, &wherearg);
    if (wherearg) {
      g_string_sprintfa(string, "\nWHERE\n%s", retval);
    }
    g_free(retval);

    retval = string->str;
    g_string_free(string, FALSE);
  }

  return retval;
}

gchar* sql_query_get_select_all_query_with_values(SqlQuery *q, gint row)
{
  GString *str;
  gchar *retval;
  gboolean wherearg, first;
  gint i;
  GSList *list;
  QueryElement *qe;
  SqlMemTable *t;
  gchar *strchar, *tmpstr;
  SqlDataDisplayFns *fns;
  Gda_Field *gfield;

  str = g_string_new("");
  g_string_sprintfa(str, "%s ", q->modif_table->name);

  /* WHERE list of automatic links */
  retval = get_where_part_of_sql(q, &wherearg);
  if (wherearg) {
    g_string_sprintfa(str, " WHERE\n%s", retval);
  }
  g_free(retval);

  /* WHERE part with the values provided by the SqlQueryRes */
  i = 0;
  first = !wherearg;
  list = q->objects;
  while (list) {
    qe = (QueryElement *)(list->data);
    if (qe->type == QUERY_ELT_FIELD) {
      t = sql_db_find_table_from_field(q->conf->db, SQL_MEM_FIELD(qe->main));
      if (!wherearg) {
	g_string_append(str, " WHERE");
	wherearg = TRUE;
      }

      fns = sql_access_get_object_display_fns(q->conf->srv,
			    GTK_OBJECT(SQL_MEM_FIELD(qe->main)->type));
      gfield = sql_query_res_get_gdafield(q->res, row, i);
      strchar = (fns->gdafield_to_sql)(gfield);
      tmpstr = sql_query_res_stringify(gfield);
      if (! tmpstr) 
	{
	  if (!first)
	    g_string_append(str, " AND");
	  else 
	    first = FALSE;
	  g_string_sprintfa(str, " %s.%s is null",
			    t->name, SQL_MEM_FIELD(qe->main)->name);
	}
      else 
	{
	  if (strchar)
	    {
	      if (!first)
		g_string_append(str, " AND");
	      else 
		first = FALSE;
	      g_string_sprintfa(str, " %s.%s=%s", 
				t->name,
				SQL_MEM_FIELD(qe->main)->name,
				strchar);
	    }
	}
      if (strchar)
	g_free(strchar);
      g_free(tmpstr);
    }
    i++;
    list = g_slist_next(list);
  }
  
  retval = str->str;
  g_string_free(str, FALSE);
  return retval;
}

void sql_query_set_text_sql(SqlQuery *q, gchar *sql)
{
  if (q->text_sql) {
    g_free(q->text_sql);
    q->text_sql = NULL;
    q->text_only = FALSE;
  }

  if (sql) {
    q->text_sql = g_strdup(sql);
    q->text_only = TRUE;
  }

#ifdef debug_signal
  g_print(">> 'CHANGED' from sql_query_set_text_sql\n");
#endif
  gtk_signal_emit(GTK_OBJECT(q), sql_query_signals[CHANGED]);
#ifdef debug_signal
  g_print("<< 'CHANGED' from sql_query_set_text_sql\n");
#endif
}

/**************************************************************/
/* CBs and functions to take care of the automatic links list */
/**************************************************************/
static void refresh_link(SqlQuery *q, QueryLink *ql)
{
  SqlMemTable *fromt, *tot;

  fromt = sql_db_find_table_from_field(q->conf->db, ql->link->from);
  tot = sql_db_find_table_from_field(q->conf->db, ql->link->to);
  if (fromt && tot) {
    if (table_is_in_query(q, fromt) && table_is_in_query(q, tot))
      ql->useit = TRUE;
    else
      ql->useit = FALSE;
  }
#ifdef debug_signal
  g_print(">> 'CHANGED' from sqldb->refresh_link\n");
#endif
  gtk_signal_emit(GTK_OBJECT(q), sql_query_signals[CHANGED]);
#ifdef debug_signal
  g_print("<< 'CHANGED' from sqldb->refresh_link\n");
#endif
}

/* CBs to be connected to SqlDb::"ff_link_created/dropped" */
void sql_query_auto_links_add_link(GtkObject *obj, SqlMemFlink *link, 
				   SqlQuery *q)
{
  SqlMemTable *fromt, *tot;
  GSList *list;
  QueryLink *ql;
  gboolean found;

  /* trying to find the link to see if needed to insert it */
  list = q->depend_list;
  found = FALSE;
  while (list && !found) {
    ql = (QueryLink *) (list->data);
    if (ql->link == link)
      found = TRUE;
    list = g_slist_next(list);
  }

  if (!found) {
    /* creating a new element */
    fromt = sql_db_find_table_from_field(q->conf->db, link->from);
    tot = sql_db_find_table_from_field(q->conf->db, link->to);
    if (fromt && tot) {
      ql = (QueryLink *) g_malloc(sizeof(QueryLink));
      ql->link = link;
      refresh_link(q, ql);
      /* insert it into the list */
      q->depend_list = g_slist_append(q->depend_list, ql);
#ifdef debug_signal
      g_print(">> 'AUTO_LINK_ADDED' from sql_query_auto_links_add_link\n");
#endif
      gtk_signal_emit(GTK_OBJECT(q), sql_query_signals[AUTO_LINK_ADDED],
		      ql);
#ifdef debug_signal
      g_print("<< 'AUTO_LINK_ADDED' from sql_query_auto_links_add_link\n");
#endif
    }
  }
}

void sql_query_auto_links_del_link(GtkObject *obj, SqlMemFlink *link, 
				   SqlQuery *q)
{
  GSList *list;
  QueryLink *ql;
  gboolean found;

  list = q->depend_list;
  found = FALSE;
  while(list && !found) {
    ql = (QueryLink *) (list->data);
    if (ql->link == link) {
      found = TRUE;
      q->depend_list = g_slist_remove_link(q->depend_list, list);
#ifdef debug_signal
      g_print(">> 'AUTO_LINK_REMOVED' from sql_query_auto_links_del_link\n");
#endif
      gtk_signal_emit(GTK_OBJECT(q), sql_query_signals[AUTO_LINK_REMOVED],
		      ql);
#ifdef debug_signal
      g_print("<< 'AUTO_LINK_REMOVED' from sql_query_auto_links_del_link\n");
#endif
      g_free(ql);
    }
    else
      list = g_slist_next(list);
  }
}

void sql_query_auto_links_refresh(SqlQuery *q)
{
  GSList *list;

  list = q->depend_list;
  while (list) {
    refresh_link(q, (QueryLink *)(list->data));
    list = g_slist_next(list);
  }
}


/************************************************************************/
/*                                                                      */
/* Result of the SELECT query and users of that result                  */
/*                                                                      */
/************************************************************************/
static void  sql_query_reg_user_destroy_cb(GtkObject *obj, SqlQuery *q);

void         sql_query_register_use_result(SqlQuery *q, GtkObject *obj)
{
  guint sigid;

  if (!q->res) {    /* making the query */
    gchar *query;
    query = sql_query_get_select_query(q);
    q->res = sql_access_do_query(q->conf->srv, query);
    g_free(query);
    /*g_print("res created\n");*/
  }
  sigid = gtk_signal_connect(obj, "destroy",
			     GTK_SIGNAL_FUNC(sql_query_reg_user_destroy_cb),q);
  gtk_object_set_data(obj, "qsig", GUINT_TO_POINTER(sigid));
  q->reg_users = g_slist_append(q->reg_users, obj);
  /*g_print("Object %p registered with signal %d\n", obj, sigid);*/
}

void         sql_query_unregister_use_result(SqlQuery *q, GtkObject *obj)
{
  GSList *list;
  guint sigid;

  list = g_slist_find(q->reg_users, obj);
  if (!list)
    return;

  q->reg_users = g_slist_remove_link(q->reg_users, list);
  sigid = GPOINTER_TO_UINT(gtk_object_get_data(obj, "qsig"));
  gtk_signal_disconnect(obj, sigid);
  /*g_print("Object %p unregistered with signal %d\n", obj, sigid);*/
  
  if (!q->reg_users) {
    /* time to destroy the result */
    if (q->res) {
      gtk_object_destroy(GTK_OBJECT(q->res));
      q->res = NULL;
      /*g_print("res destroyed\n");*/
    }
  }
}

void sql_query_refresh_res(SqlQuery *q)
{
  gchar *query;
  if (q->res) {
    gtk_object_destroy(GTK_OBJECT(q->res));

    query = sql_query_get_select_query(q);
    q->res = sql_access_do_query(q->conf->srv, query);
    g_free(query);
#ifdef debug_signal
    g_print(">> 'QUERY_REFRESHED' from sql_query_refresh_res\n");
#endif
    gtk_signal_emit(GTK_OBJECT(q), sql_query_signals[QUERY_REFRESHED]);  
#ifdef debug_signal
    g_print("<< 'QUERY_REFRESHED' from sql_query_refresh_res\n");
#endif
  }
}

static void  sql_query_reg_user_destroy_cb(GtkObject *obj, SqlQuery *q)
{
  sql_query_unregister_use_result(q, obj);
}


gint sql_query_is_field_printed(SqlQuery *q, SqlMemField *field)
{
  gint result=-1, i;
  GSList *list;
  QueryElement *elt;

  list = q->objects;
  i = 0;
  while (list && (result==-1)) {
    elt = (QueryElement *)(list->data);
    if ((elt->type==QUERY_ELT_FIELD) && (elt->main == field) && 
	(elt->print_name) && (*(elt->print_name) != '\0'))
      result = i;
    if ((elt->print_name) && (*(elt->print_name) != '\0'))
      i++;
    list = g_slist_next(list);
  }

  return result;
}

QueryElement *sql_query_get_query_element(SqlQuery *q, gint pos)
{
  QueryElement *qe=NULL;
  gint i;
  GSList *list;
  
  if (pos <0) return NULL;
  list = q->objects;
  i = -1;
  while (list && !qe) {
    if ((((QueryElement *)(list->data))->print_name) &&
	(*(((QueryElement *)(list->data))->print_name) != '\0')) {
      i++;
      if (i == pos)
	qe = (QueryElement *)(list->data);
    }
    list = g_slist_next(list);
  }

  return qe;
}

gint sql_query_get_queryres_pos(SqlQuery *q, QueryElement *qe)
{
  return g_slist_index(q->objects, qe);
}

gint sql_query_get_queryres_pos_f(SqlQuery *q, SqlMemField *field)
{
  GSList *list;
  gboolean found;
  list = q->objects;
  found = FALSE;
  while (list && !found) 
    {
      if ((((QueryElement *)(list->data))->type == QUERY_ELT_FIELD) && 
	  (((QueryElement *)(list->data))->main == field)) 
	found = TRUE;
      else
	list = g_slist_next(list);
    }
  if (!list)
    {
      g_warning("sql_query_get_queryres_pos_f: field not found");
      return -1;
    }
  else
    return g_slist_index(q->objects, list->data);
}
