/* sqlquery.h
 *
 * Copyright (C) 1999,2000 Vivien Malerba
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 */

#ifndef __SQL_QUERY__
#define __SQL_QUERY__

#include <gtk/gtkobject.h> 
#include "sqldb.h"
#include "gasql_conf.h"

#ifdef __cplusplus
extern "C" {
#endif /* __cplusplus */ 

#define SQL_QUERY(obj)          GTK_CHECK_CAST (obj, sql_query_get_type(), SqlQuery)
#define SQL_QUERY_CLASS(klass)  GTK_CHECK_CLASS_CAST (klass, sql_query_get_type (), SqlQueryClass) 
#define IS_SQL_QUERY(obj)       GTK_CHECK_TYPE (obj, sql_query_get_type ())


  typedef struct _Sqlquery        SqlQuery;
  typedef struct _SqlqueryClass   SqlQueryClass;

  /* these are static to the SqlQuery class */
  typedef struct _QueryElement     QueryElement;
  typedef struct _QueryOperation   QueryOperation;
  typedef struct _QueryCriteria    QueryCriteria;
  typedef struct _QueryLink        QueryLink;

  typedef enum {
    QUERY_ELT_FIELD,
    QUERY_ELT_AGGREGATE,
    QUERY_ELT_FUNCTION
  } QueryEltType;

  struct _QueryElement{
    QueryEltType type;
    /* QUERY_ELT_FIELD     => main is a ptr on SqlMemField object
       QUERY_ELT_AGGREGATE => main is a ptr on SqlDataAggregate object
       QUERY_ELT_FUNCTION  => main is a ptr on SqlDataFunction object */
    gpointer main;
    GSList *args; /* args is a list of pointers on other QueryElement structs
		   For Aggregate*/
    gchar *print_name;
    gchar *name;
  };

  struct _QueryLink{
    SqlMemFlink *link;
    gboolean useit;
  };

  /* struct for the object's data */
  struct _Sqlquery {
    GtkObject object;
    
    gASQL_Main_Config *conf;
    guint id; /* comes from conf->id_serial */
    gchar *name;
    gchar *descr;
    GSList *objects; /* pointers are to QueryElements */
    GSList *select_list;
    GSList *group_by_list;
    GSList *order_by_list;
    GSList *depend_list; /* dependencies imposed by the program.
			    Pointers to QueryLink structures */
    /* TODO: add the where clause tree */
    gboolean text_only;
    gchar *text_sql; /* if we have a text only query, otherwise NULL */

    /* result of the SELECT query and list of registered objects to use that
       result. */
    SqlQueryRes *res; 
    GSList *reg_users;
			 
    SqlMemTable *modif_table; /* table which will be modified if the user
				 wants to write some modifications from the 
				 result of the queries */
  };

  /* struct for the object's class */
  struct _SqlqueryClass {
    GtkObjectClass parent_class;
    
    void (* elt_created) (SqlQuery *q, QueryElement *new_elt);
    void (* elt_dropped) (SqlQuery *q, QueryElement *elt);
    void (* elt_modified) (SqlQuery *q, QueryElement *elt);
    void (* changed) (SqlQuery *q);
    void (* name_changed) (SqlQuery *q);
    void (* auto_link_added) (SqlQuery *q, QueryLink *ql);
    void (* auto_link_removed) (SqlQuery *q, QueryLink *ql);
    void (* modif_table_changed) (SqlQuery *q, SqlMemTable *table);
    void (* query_refreshed) (SqlQuery *q);
  };

  /* generic widget's functions */
  guint        sql_query_get_type(void);
  GtkObject*   sql_query_new(gchar *name, gASQL_Main_Config *conf);
  GtkObject*   sql_query_new_noref(gchar *name, gASQL_Main_Config *conf);
  SqlQuery*    sql_query_build_from_xml_tree(gASQL_Main_Config *conf, 
					     xmlNodePtr node);
  void         sql_query_build_xml_tree(SqlQuery *q, gASQL_Main_Config *conf, 
					xmlDocPtr doc);

  /* result of the SELECT query and list of registered objects to use that
     result. */
  void         sql_query_register_use_result(SqlQuery *q, GtkObject *obj);
  void         sql_query_unregister_use_result(SqlQuery *q, GtkObject *obj);
  void         sql_query_refresh_res(SqlQuery *q);

  /* if NULL passed as gchar*, nothing happens for that gchar* */
  void         sql_query_set_name(SqlQuery *q, gchar *name, gchar *descr);
  
  /* sets the text SQL (and toggles the query type) */
  void          sql_query_set_text_sql(SqlQuery *q, gchar *sql);

  /* inserts a new field element for the query and returns it */
  QueryElement* sql_query_new_elt_field(SqlQuery *q,
					gchar *name,
					gchar *print_name,
					SqlMemField *field);
  QueryElement* sql_query_new_elt_aggregate(SqlQuery *q,
					    gchar *name,
					    gchar *print_name,
					    SqlDataAggregate *agg, 
					    GSList *args);
  QueryElement* sql_query_new_elt_function(SqlQuery *q, 
					   gchar *name,
					   gchar *print_name,
					   SqlDataFunction *func, 
					   GSList *args);
  QueryElement* sql_query_get_element(SqlQuery *q, gchar *name);

  void          sql_query_del_elt(SqlQuery *q, QueryElement* elt);
  void          sql_query_swap_objects(SqlQuery *q, gint row1, gint row2);
  
  void          sql_query_modif_elt(SqlQuery *q, QueryElement* elt,
				    gchar *name, gchar *print_name);

  void          sql_query_auto_links_refresh(SqlQuery *q);
  void          sql_query_auto_links_add_link(GtkObject *obj, 
					      SqlMemFlink *link, SqlQuery *q);
  void          sql_query_auto_links_del_link(GtkObject *obj, 
					      SqlMemFlink *link, SqlQuery *q);

  /* tells if a particular field appears in the print list of the Query:
     returns -1 if not present and >=0 otherwise, the number being the column
     number of the SqlQueryRes of the SELECT query */
  gint          sql_query_is_field_printed(SqlQuery *q, SqlMemField *field);

  /* returns the QueryElement of the query given the column number it occupies
     int the SqlQueryRes of the SELECT row */
  QueryElement *sql_query_get_query_element(SqlQuery *q, gint pos);

  /* returns the position of a QueryElement in the resulting SqlQueryRes 
     resultat (opposite of function above) */
  gint          sql_query_get_queryres_pos_qe(SqlQuery *q, QueryElement *qe);
  gint          sql_query_get_queryres_pos_f(SqlQuery *q, SqlMemField *field);

  /* sets which table is modified from this query if an update is initiated 
     from the result of the query */
  void          sql_query_set_modif_table(SqlQuery *q, SqlMemTable *t);

  /* getting the SQL version of the different queries */
  /* memory is allocated and will have to be freed */
  gchar*       sql_query_get_select_query(SqlQuery *q);

  /* getting the SQL query to select all the fields of a table given
     the entry number in the SqlQueryRes for the SELECT version of the query 
     WARNING: returns only the part of the query after the FROM statement,
     so it can be used with DELETE, UPDATE as well */
  gchar*       sql_query_get_select_all_query_with_values(SqlQuery *q, 
							  gint row);
#ifdef __cplusplus
}
#endif /* __cplusplus */   

#endif
