/* sqlqueryexec.c
 *
 * Copyright (C) 1999,2000 Vivien Malerba
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 */

#include <config.h>
#include "sqlqueryexec.h"
#include "key.xpm"
#include "pen.xpm"

static void      sql_query_exec_class_init      (SqlQueryExecClass *class);
static void      sql_query_exec_init            (SqlQueryExec *qe);
static void      sql_query_exec_destroy         (GtkObject *object);
static void      sql_query_exec_post_init       (SqlQueryExec *qe);

static void      ask_for_query_refresh          (SqlQueryExec *qe);
static void      refresh_query                  (SqlQuery *q,
						 SqlQueryExec *qe);

/* various CB */
static void      sql_query_name_changed_cb      (GtkObject *obj, 
						 SqlQueryExec *qw);
static void object_clist_select_row_cb(GtkWidget *wid, 
				       gint row, gint column, 
				       GdkEventButton *event,
				       SqlQueryExec *qe);
static void sql_access_conn_to_close_cb(GtkObject *obj, SqlQueryExec *qe);

/* functions to have a little menu at the top right to choose which table
   is to be modified */
static void cb_update_menu_select(GtkWidget *item, SqlMemTable *t);
static GtkWidget *make_menu_item( gchar         *name,
				  GtkSignalFunc  callback,
				  gpointer       data );
static void sql_query_exec_update_modify_tables(SqlQueryExec *qe);

/* coolumn titles management */
static void set_column_title(GtkCList *clist, gASQL_Main_Config *conf,
			     gint col, gboolean key, gboolean write);
static void update_titles_contents(SqlQueryExec *qe);

/* element editing/removing */
static void edit_element_cb(GtkObject *obj, SqlQueryExec *qe);
static void delete_element_cb(GtkObject *obj, SqlQueryExec *qe);

/***************************************************************************/
/*                                                                         */
/* Main functions for the object                                           */
/*                                                                         */
/***************************************************************************/

/* get a pointer to the parents to be able to call their destructor */
static GtkObject *parent_class = NULL;

enum {
  SELECTION_CHANGED,
  LAST_SIGNAL
};

static gint sql_query_exec_signals[LAST_SIGNAL] = { 0 };

guint sql_query_exec_get_type(void) 
{
  static guint f_type = 0;
  
  if (!f_type)
    {
      GtkTypeInfo f_info =
      {
        "Sql_Query_Exec",
        sizeof (SqlQueryExec),	
        sizeof (SqlQueryExecClass),
        (GtkClassInitFunc) sql_query_exec_class_init,
        (GtkObjectInitFunc) sql_query_exec_init,
        (GtkArgSetFunc) NULL,
        (GtkArgGetFunc) NULL
      };
      
      f_type = gtk_type_unique (gtk_vbox_get_type (), &f_info);
    }
  
  return f_type;   
}

static void sql_query_exec_class_init(SqlQueryExecClass *class)
{
  GtkObjectClass *object_class;

  object_class = (GtkObjectClass*) class;   
  sql_query_exec_signals[SELECTION_CHANGED] =
    gtk_signal_new ("selection_changed",
                    GTK_RUN_FIRST,
                    object_class->type,
                    GTK_SIGNAL_OFFSET (SqlQueryExecClass, selection_changed),
                    gtk_signal_default_marshaller, GTK_TYPE_NONE, 0);

  gtk_object_class_add_signals (object_class, sql_query_exec_signals, 
				LAST_SIGNAL);
  class->selection_changed = NULL;

  parent_class = gtk_type_class (gtk_vbox_get_type());
  object_class->destroy = sql_query_exec_destroy;
}

static void sql_query_exec_init(SqlQueryExec *qe)
{
  qe->sel_row = -1; 
  qe->clist = NULL;
  qe->modif_table = NULL;
  qe->tables_opt_menu = NULL;
  qe->is_error = FALSE;
}


static void sql_query_exec_post_init(SqlQueryExec *qe)
{
  GtkWidget *hb, *label;
  GtkWidget *sw, *te;
  GtkWidget *opt, *menu;

  /* hbox for the top text */
  hb = gtk_hbox_new(FALSE, 0);
  gtk_box_pack_start(GTK_BOX(qe), hb, FALSE, TRUE, GNOME_PAD);
  gtk_widget_show(hb);

  /* Label of the query */
  label = gtk_label_new(_("Query name:"));
  gtk_box_pack_start(GTK_BOX(hb), label, FALSE, TRUE, GNOME_PAD);
  gtk_widget_show(label);

  te = gtk_entry_new();
  gtk_entry_set_text(GTK_ENTRY(te), qe->q->name); 
  gtk_box_pack_start(GTK_BOX(hb), te, FALSE, TRUE, GNOME_PAD);
  gtk_widget_set_sensitive(te, FALSE);
  gtk_widget_show(te);
  qe->name_te = te;

  /* which table to update */
  if (! qe->q->text_only) {
    label = gtk_label_new(_("Table to modify:"));
    gtk_box_pack_start(GTK_BOX(hb), label, FALSE, TRUE, GNOME_PAD);
    gtk_widget_show(label);
    
    opt = gtk_option_menu_new();
    menu = gtk_menu_new();
    qe->tables_opt_menu = opt;
    qe->modif_table = qe->q->modif_table;
    
    gtk_box_pack_start (GTK_BOX(hb), opt, TRUE, TRUE, 0);
    gtk_widget_show (opt);
  }

  /* main component: CList */
  sw = gtk_scrolled_window_new(NULL, NULL);
  qe->sw = sw;
  gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(sw),
				 GTK_POLICY_AUTOMATIC, GTK_POLICY_ALWAYS);
  gtk_box_pack_start(GTK_BOX(qe), sw, TRUE, TRUE, 0);
  gtk_widget_show(sw);

  /* the CList is created in the next function */
  refresh_query(qe->q, qe);
  sql_query_exec_update_modify_tables(qe);
  sql_query_set_modif_table(qe->q, qe->modif_table);
}

GtkWidget* sql_query_exec_new(SqlQuery *q)
{
  GtkWidget *obj;
  SqlQueryExec *qe;

  obj =  GTK_WIDGET(gtk_type_new(sql_query_exec_get_type()));
  qe = SQL_QUERY_EXEC(obj);
  qe->q = q;
  sql_query_register_use_result(q, GTK_OBJECT(obj));
  sql_query_exec_post_init(qe);

  /* connection to signals from SqlDb */
  /* WARNING: no connection to the signal SqlQuery::"changed" because
     we already connect to the "query_refreshed" one which is more generic
     and when the SqlQuery changes, it also emits the "query_refreshed" 
     signal */
  gtk_signal_connect_while_alive(GTK_OBJECT(q), "query_refreshed",
				 GTK_SIGNAL_FUNC(refresh_query),
				 qe, GTK_OBJECT(qe));
  gtk_signal_connect_while_alive(GTK_OBJECT(q), "name_changed",
				 GTK_SIGNAL_FUNC(sql_query_name_changed_cb),
				 qe, GTK_OBJECT(qe));
  gtk_signal_connect_while_alive(GTK_OBJECT(q->conf->srv), "conn_to_close",
				 GTK_SIGNAL_FUNC(sql_access_conn_to_close_cb),
				 qe, GTK_OBJECT(qe));
  return obj;
}

static void sql_access_conn_to_close_cb(GtkObject *obj, SqlQueryExec *qe)
{

}

static void sql_query_exec_destroy(GtkObject *object)
{
  SqlQueryExec *qe;
  GnomeDialog *dlg;
  gint i;
  
  g_return_if_fail (object != NULL);
  g_return_if_fail (IS_SQL_QUERY_EXEC (object));

  qe = SQL_QUERY_EXEC(object); 

  /* closing any EDIT dialog which remained opened */
  for (i=0; i< GTK_CLIST(qe->clist)->rows; i++) {
    dlg = gtk_clist_get_row_data(GTK_CLIST(qe->clist), i);
    if (dlg) {
      gnome_dialog_close(dlg);
      gtk_clist_set_row_data(GTK_CLIST(qe->clist), i, NULL);
    }
  }
  /* closing the NEW dialog if opened */
  dlg = gtk_object_get_data(GTK_OBJECT(qe->clist), "newdlg");
  if (dlg) 
    gnome_dialog_close(dlg);


  if (GTK_OBJECT_CLASS (parent_class)->destroy)
    (* GTK_OBJECT_CLASS (parent_class)->destroy) (object); 
}


/*************************************************************************/
/*                                                                       */
/* Table to modify selection management                                  */
/*                                                                       */
/*************************************************************************/
static void cb_update_menu_select( GtkWidget     *item,
				   SqlMemTable *t )
{
  SqlQueryExec *qe;

  qe = 	gtk_object_get_data(GTK_OBJECT(item), "qe");
  if (qe->modif_table != t) {
    qe->modif_table = t;
    update_titles_contents(qe);
    sql_query_set_modif_table(qe->q, t);
  }
}


/* returns the menuitem */
static GtkWidget *make_menu_item( gchar         *name,
				  GtkSignalFunc  callback,
				  gpointer       data )
{
  GtkWidget *item;
  
  item = gtk_menu_item_new_with_label (name);
  gtk_signal_connect (GTK_OBJECT (item), "activate",
		      callback, data);
  gtk_widget_show (item);
  
  return(item);
}

/* do update the titles of the tabular, after setting the possible 
   tables which can be updated from here */
static void sql_query_exec_update_modify_tables(SqlQueryExec *qe)
{
  GList *list;
  GSList *elts;
  QueryElement *elt;
  SqlMemTable *table, *seltab, *firsttab;
  GtkWidget *item, *menu;

  if (! qe->tables_opt_menu) return;

  /* emptying the list */
  seltab = qe->modif_table;
  gtk_option_menu_remove_menu(GTK_OPTION_MENU(qe->tables_opt_menu));

  /* new list */
  menu = gtk_menu_new();
  list = NULL;
  elts = qe->q->objects;
  firsttab = NULL;
  while (elts) {
    elt = (QueryElement *)(elts->data);
    if (elt->type == QUERY_ELT_FIELD) {
      table = sql_db_find_table_from_field(qe->q->conf->db, 
					   SQL_MEM_FIELD(elt->main));
      if (! g_list_find(list, table)) {
	if (!firsttab) firsttab = table;
	list = g_list_append(list, table);
	item = make_menu_item (table->name,
			       GTK_SIGNAL_FUNC (cb_update_menu_select),
			       table);
	gtk_object_set_data(GTK_OBJECT(item), "qe", qe);
	gtk_menu_append(GTK_MENU(menu), item);
      }
    }
    elts = g_slist_next(elts);
  }
  gtk_option_menu_set_menu(GTK_OPTION_MENU(qe->tables_opt_menu), menu);
  
  /* modify the selection if necessary */
  if (g_list_find(list, seltab)) {
    gint pos;

    pos = g_list_index(list, seltab);
    gtk_option_menu_set_history(GTK_OPTION_MENU(qe->tables_opt_menu), pos);
  }
  else
    qe->modif_table = firsttab;
  g_list_free(list);
}

/*************************************************************************/
/*                                                                       */
/* Column titles management                                              */
/*                                                                       */
/*************************************************************************/
static void set_column_title(GtkCList *clist, gASQL_Main_Config *conf,
			     gint col, gboolean key, gboolean write)
{
  GtkWidget *align, *label=NULL, *button, *hbox;
  GdkPixmap* icon = NULL;
  GdkBitmap* mask = NULL;
  GtkWidget *pixmapwid;
  GtkStyle* style = gtk_widget_get_style(GTK_WIDGET(clist));

  button = (clist->column)[col].button;

  align = GTK_BIN(button)->child;

  if (GTK_BIN(align)->child)/*here the contents of the alignment is destroyed*/
    gtk_container_remove(GTK_CONTAINER(align), GTK_BIN(align)->child);

  hbox = gtk_hbox_new(FALSE, 0);
  gtk_container_add(GTK_CONTAINER(align), hbox);
  if (key) {
    icon = mask = NULL;

    /* the icon must be inserted in the button (clist->column)[col].button */
    icon = gdk_pixmap_create_from_xpm_d(conf->app->window, &mask,
					&style->bg[GTK_STATE_NORMAL], 
					(gchar **)icon_key);
    pixmapwid = gtk_pixmap_new (icon, mask);
    
    gtk_box_pack_start(GTK_BOX(hbox), pixmapwid, FALSE, FALSE, 0);
    gtk_widget_show(pixmapwid);
  }

  if (write) {
    icon = mask = NULL;

    /* the icon must be inserted in the button (clist->column)[col].button */
    icon = gdk_pixmap_create_from_xpm_d(conf->app->window, &mask,
					&style->bg[GTK_STATE_NORMAL], 
					(gchar **)icon_pen);
    pixmapwid = gtk_pixmap_new (icon, mask);
    
    gtk_box_pack_start(GTK_BOX(hbox), pixmapwid, FALSE, FALSE, 0);
    gtk_widget_show(pixmapwid);
  }

  label = gtk_label_new(clist->column[col].title);
  gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, FALSE, 0);
  gtk_widget_show(label);
  gtk_widget_show(hbox);
}

static void update_titles_contents(SqlQueryExec *qe)
{
  GSList *list;
  gint i;
  gboolean keyb, writeb;
  SqlMemTable *t;

  if (!qe->q->text_only && !qe->is_error) {
    QueryElement *qee;
    i = 0;
    list = qe->q->objects;
    while (list) {
      qee = (QueryElement *)(list->data);
      if ((qee->type == QUERY_ELT_FIELD) &&
	  (qee->print_name)) {
	if (SQL_MEM_FIELD(qee->main)->is_key) keyb = TRUE;
	else keyb = FALSE;

	t = qe->modif_table ? sql_db_find_table_from_field(qe->q->conf->db, 
					  SQL_MEM_FIELD(qee->main)) : NULL;
	if (qe->modif_table && t && (qe->modif_table == t))
	  writeb = TRUE;
	else
	  writeb = FALSE;
	set_column_title(GTK_CLIST(qe->clist), qe->q->conf, i, keyb, writeb);
	i++;
      }
      list = g_slist_next(list);
    }
  }

}


/*************************************************************************/
/*                                                                       */
/* Element editing                                                       */
/*                                                                       */
/*************************************************************************/
static void edit_element_dlg_close_cb(GtkObject *obj, SqlQueryExec *qe)
{
  gint row;

  row = GPOINTER_TO_INT(gtk_object_get_data(GTK_OBJECT(obj), "row"));
  gtk_clist_set_row_data(GTK_CLIST(qe->clist), row, NULL);
}

static void edit_element_cb(GtkObject *obj, SqlQueryExec *qe)
{
  GtkWidget *dlg;

  if (qe->sel_row < 0) /* should never happen */
    return;

  dlg = gtk_clist_get_row_data(GTK_CLIST(qe->clist), qe->sel_row);
  if (!dlg) {
    dlg = sql_query_form_new_dlg(qe->q, qe->sel_row+1, FALSE);
    gtk_object_set_data(GTK_OBJECT(dlg), "row", GINT_TO_POINTER(qe->sel_row));
    gtk_clist_set_row_data(GTK_CLIST(qe->clist), qe->sel_row, dlg);
    gtk_widget_show(dlg);
    gtk_signal_connect_while_alive(GTK_OBJECT(dlg), "destroy",
				   GTK_SIGNAL_FUNC(edit_element_dlg_close_cb), 
				   qe, GTK_OBJECT(qe));
  }
  else
    gdk_window_raise(dlg->window);
}

static void dummy_cb(gint reply, gpointer data)
{
}

static void delete_element_cb(GtkObject *obj, SqlQueryExec *qe)
{
  GtkWidget *dlg;
  gchar *str, *querystr; 
  GString *query;
  GString *question;
  SqlQueryRes *res;
  GSList *links, *list;
  SqlMemFlink *link;
  SqlMemTable *t;
  gboolean cont;
  gint i;
  SqlDataDisplayFns *fns;
  Gda_Field *gfield;

  if (qe->sel_row < 0) /* should never happen */
    return;
  
  /* Integrity check */
  cont = TRUE;
  question = NULL;
  query = NULL;
  links = qe->q->conf->db->field_links;
  while(links) {
    link = SQL_MEM_FLINK(links->data);
    if (g_slist_find(qe->q->modif_table->fields, link->from)) 
      {
	t = sql_db_find_table_from_field(qe->q->conf->db, link->to);
	query = g_string_new("SELECT count(");
	g_string_sprintfa(query, "%s) from %s where %s.%s=%s.%s",
			  link->to->name, t->name, 
			  t->name, link->to->name, qe->q->modif_table->name,
			  link->from->name);
	/* going through fields of modif_table to insert them in WHERE 
	   with their values */
	list = qe->q->modif_table->fields;
	while (list)
	  {
	    if ((i=sql_query_is_field_printed
		 (qe->q, SQL_MEM_FIELD(list->data))) >= 0)
	      {
		fns = sql_access_get_object_display_fns(qe->q->conf->srv,
			    GTK_OBJECT(SQL_MEM_FIELD(list->data)->type));
		gfield = sql_query_res_get_gdafield(qe->q->res, 
						    qe->sel_row+1, i);
		str = (fns->gdafield_to_sql)(gfield); 
		if (str && (*str != '\0')) 
		  g_string_sprintfa(query, " AND %s.%s = %s",
				    qe->q->modif_table->name, 
				    SQL_MEM_FIELD(list->data)->name,
				    str);
		if (str)
		  g_free(str);
	      }
	    list = g_slist_next(list);
	  }
	
	res = sql_access_do_query(qe->q->conf->srv, query->str);
	g_string_free(query, TRUE);
	if (res) 
	  {
	    str = sql_query_res_get_item(res, 1, 0);
	    i = atoi(str);
	    g_free(str);
	    if (i>0) 
	      {
		if (!question)
		  question = g_string_new(_("Integrity problem:\n"
					    "If you remove this item, "
					    "the following\n"
					    "dependencies will fail:"));
		g_string_sprintfa(question, 
				  "\n%s.%s --> %s.%s (%d)",
				  qe->q->modif_table->name, link->from->name,
				  t->name, link->to->name, i);
	      }
	    gtk_object_destroy(GTK_OBJECT(res));
	  }
      }
    links = g_slist_next(links);
  }
  if (question) {
    g_string_sprintfa(question, _("\nDo you still want to remove this item?"));
    dlg = gnome_question_dialog(question->str, 
				(GnomeReplyCallback)(dummy_cb), NULL);
    g_string_free(question, TRUE);
    i = gnome_dialog_run_and_close(GNOME_DIALOG(dlg));
    cont = FALSE;
    if (!i)
      cont = TRUE;
  }

  /* A last confirmation */
  if (!cont) return;
  str = sql_query_get_select_all_query_with_values(qe->q, qe->sel_row+1);
  querystr = g_strdup_printf("SELECT * FROM %s", str);
  g_print("Query is: #%s#\n", querystr);
  res = sql_access_do_query(qe->q->conf->srv, querystr);
  g_free(querystr);
  if (res) {
    g_print("Res is %p\n", res);
    i = sql_query_res_get_nbtuples(res);
    gtk_object_destroy(GTK_OBJECT(res));
    question = g_string_new("");
    if (i>1)
      g_string_sprintfa(question, _("You are about to remove %d rows.\nDo you "
				    "want to continue?\n"), i);
    else
      g_string_sprintfa(question, _("You are about to remove one row.\nDo you "
				    "want to continue?\n"));

    dlg = gnome_question_dialog(question->str, 
				(GnomeReplyCallback)(dummy_cb), NULL);
    g_string_free(question, TRUE);
    i = gnome_dialog_run_and_close(GNOME_DIALOG(dlg));
    if (!i) {
      querystr = g_strdup_printf("DELETE FROM %s", str);
      g_print("Query is: #%s#\n", querystr);
      res = sql_access_do_query(qe->q->conf->srv, querystr);
      g_free(querystr);
    }
  }
  g_free(str);
}

static void new_element_dlg_close_cb(GtkObject *obj, SqlQueryExec *qe)
{
  gtk_object_set_data(GTK_OBJECT(qe->clist), "newdlg", NULL);
}

static void new_element_cb(GtkObject *obj, SqlQueryExec *qe)
{
  GtkWidget *dlg;

  dlg = gtk_object_get_data(GTK_OBJECT(qe->clist), "newdlg");
  if (!dlg) {
    dlg = sql_query_form_new_dlg(qe->q, -1, TRUE);
    gtk_widget_show(dlg);
    gtk_object_set_data(GTK_OBJECT(qe->clist), "newdlg", dlg);
    gtk_signal_connect_while_alive(GTK_OBJECT(dlg), "destroy",
				   GTK_SIGNAL_FUNC(new_element_dlg_close_cb), 
				   qe, GTK_OBJECT(qe));
  }
  else
    gdk_window_raise(dlg->window);
}

/*************************************************************************/
/*                                                                       */
/* Query Update                                                          */
/*                                                                       */
/*************************************************************************/
/* Callback to handle the button presses over the CList */
static gint press_handler_event(GtkWidget *widget,
				GdkEventButton *event, SqlQueryExec *qe)
{
  gint row, col;

  if (qe->q->text_only)
    return FALSE;
  if (GTK_IS_CLIST(widget)) {
    /* setting the right selection */
    gtk_clist_get_selection_info(GTK_CLIST(widget),
				 event->x, event->y,
				 &row, &col);
    if ((qe->sel_row != row) && (event->button != 1)) 
      gtk_clist_select_row(GTK_CLIST(widget), row, col);    

    /* depending on what the clicks were */
    if (event->type==GDK_2BUTTON_PRESS) {
      edit_element_cb(NULL, qe);
      return TRUE;
    }

    if (event->button == 3) {
      GtkWidget *menu, *wid;

      menu = gtk_menu_new();
      wid = gtk_menu_item_new_with_label(_("Modify"));
      gtk_signal_connect(GTK_OBJECT(wid), "activate",
			 GTK_SIGNAL_FUNC(edit_element_cb), qe);
      gtk_menu_append(GTK_MENU(menu), wid);
      gtk_widget_show(wid);
      wid = gtk_menu_item_new_with_label(_("Delete"));
      gtk_signal_connect(GTK_OBJECT(wid), "activate",
			 GTK_SIGNAL_FUNC(delete_element_cb), qe);
      gtk_menu_append(GTK_MENU(menu), wid);
      gtk_widget_show(wid);
      gtk_menu_popup (GTK_MENU(menu), NULL, NULL, NULL, NULL,
		      event->button, event->time);

      return TRUE;
    }
  }

  return TRUE;
}

static void ask_for_query_refresh(SqlQueryExec *qe)
{
  sql_query_refresh_res(qe->q);
}

static void refresh_query(SqlQuery *q, SqlQueryExec *qe)
{
  gchar **entry;
  SqlQueryRes *res;
  GtkWidget *clist;
  gint i, j, nb, nrows;
  SqlDataType **types;
  SqlDataDisplayFns **fns;
  GSList *list;
  QueryElement *elt;
  Gda_Field *gdaf;
  
  /* TOTO:if the query has not changed, let's try to keep the visual 
     as common as possible to what we already had */

  /* build titles array */
  res = qe->q->res;
  clist = NULL;
  if (res) {
    nb = sql_query_res_get_nbcols(res);
    if (nb>0) {
      entry = (gchar **) g_malloc(sizeof(gchar *)*nb);
      /* column titles */
      for (i=0; i<nb; i++) 
	entry[i] = sql_query_res_get_col_name(res, i);
      clist = gtk_clist_new_with_titles(nb, entry);
      for (i=0; i<nb; i++) 
	gtk_clist_set_column_auto_resize(GTK_CLIST(clist), i, TRUE);
      gtk_signal_connect(GTK_OBJECT(clist),
			 "button_press_event",
			 GTK_SIGNAL_FUNC(press_handler_event), qe);

      /* types of data being displayed in the clist */
      types = (SqlDataType **) g_malloc(sizeof(gpointer)*nb);
      fns = (SqlDataDisplayFns **) g_malloc(sizeof(gpointer)*nb);
      list = qe->q->objects;
      j = 0;
      while (list) {
	elt = (QueryElement *)(list->data);
	if (elt->print_name)
	  switch (elt->type) {
	  case QUERY_ELT_FIELD:
	    types[j] = SQL_MEM_FIELD(elt->main)->type;
	    break;
	  case QUERY_ELT_AGGREGATE:
	    types[j] = NULL;
	    break;
	  case QUERY_ELT_FUNCTION:
	    types[j] = SQL_DATA_FUNCTION(elt->main)->result_type;
	    break;
	  }

	else
	  types[j] = NULL;
	fns[j] = sql_access_get_object_display_fns(q->conf->srv, 
					       GTK_OBJECT(types[j]));
	j++;
	list = g_slist_next(list);
      }

      /* CList content */
      nrows = sql_query_res_get_nbtuples(res);
      for (i=1; i<=nrows; i++) {
	for (j=0; j<nb; j++) { 
	  if (types[j]) 
	    {
	      gchar *str;
	      gdaf = sql_query_res_get_gdafield(res, i, j);
	      str = (fns[j]->gdafield_to_str) (gdaf);
	      if (str)
		entry[j] = str;
	      else
		entry[j] = g_strdup("");
	    }
	  else
	    entry[j] = sql_query_res_get_item(res, i, j);
	}
	gtk_clist_append(GTK_CLIST(clist), entry);
	for (j=0; j<nb; j++) 
	  g_free(entry[j]);
      }
      g_free(entry);
      qe->is_error = FALSE;

      g_free(types);
      g_free(fns);
    }
  }
  if (!clist) {
    clist = gtk_clist_new(1);
      qe->is_error = TRUE;
  }

  gtk_clist_set_reorderable(GTK_CLIST(clist), FALSE);
  gtk_clist_set_selection_mode(GTK_CLIST(clist), GTK_SELECTION_SINGLE);
  if (qe->clist) {
    gtk_container_remove(GTK_CONTAINER(qe->sw), qe->clist);
    /* REM: the above function seems to destroy the widget which was inside
       the container */
    /*gtk_object_destroy(GTK_OBJECT(qe->clist));*/
  }
  gtk_container_add(GTK_CONTAINER(qe->sw), clist);
  qe->clist = clist;

  /* update the tables which can be modified */
  sql_query_exec_update_modify_tables(qe);

  /* set the keys and other icons in titles */
  update_titles_contents(qe);
  gtk_widget_set_usize(clist, 400, 300);
  gtk_widget_show(clist);

  gtk_clist_column_titles_passive(GTK_CLIST(clist));

  qe->sel_row = -1; /* because new CList */
#ifdef debug_signal
  g_print(">> 'SELECTION_CHANGED' from sqlqueryexec->refresh_query\n");
#endif
  gtk_signal_emit(GTK_OBJECT(qe), sql_query_exec_signals[SELECTION_CHANGED]);
#ifdef debug_signal
  g_print("<< 'SELECTION_CHANGED' from sqlqueryexec->refresh_query\n");
#endif
  if (!qe->q->text_only)
    gtk_signal_connect(GTK_OBJECT(clist), "select_row",
		       GTK_SIGNAL_FUNC(object_clist_select_row_cb), qe);

}

static void object_clist_select_row_cb(GtkWidget *wid, 
				       gint row, gint column, 
				       GdkEventButton *event,
				       SqlQueryExec *qe)
{
  qe->sel_row = row;
#ifdef debug_signal
  g_print(">> 'SELECTION_CHANGED' from "
	  "sqlqueryexec->object_clist_select_row_cb\n");
#endif
  gtk_signal_emit(GTK_OBJECT(qe), sql_query_exec_signals[SELECTION_CHANGED]);
#ifdef debug_signal
  g_print("<< 'SELECTION_CHANGED' from "
	  "sqlqueryexec->object_clist_select_row_cb\n");
#endif
}


static void sql_query_name_changed_cb(GtkObject *obj, 
				      SqlQueryExec *qe)
{
  gtk_entry_set_text(GTK_ENTRY(qe->name_te), qe->q->name);
}

/****************************************************************************/
/*                                                                          */
/* Functions related to the Dialog in which is packed the SqlQueryExec obj  */
/*                                                                          */
/****************************************************************************/
static void main_dialog_clicked_cb(GnomeDialog * dialog, 
				   gint button_number, 
				   SqlQueryExec *qe);

static void main_dialog_auto_close(GtkObject *obj, GnomeDialog *dlg)
{
  SqlQueryExec *qe = gtk_object_get_data(GTK_OBJECT(dlg), "SqlQueryExec");
  main_dialog_clicked_cb(dlg, 0, qe);
}

static void query_modif_table_cb(SqlQuery *q, SqlMemTable *t, 
				 GtkWidget *dlg)
{
  gboolean sens;

  sens = FALSE;
  if (t)
    sens = TRUE;

  gnome_dialog_set_sensitive(GNOME_DIALOG(dlg), 2, sens);
}

static void query_insert_into_table_cb(SqlQueryExec *qe, GtkWidget *dlg)
{
  if (qe->sel_row >= 0)
    gnome_dialog_set_sensitive(GNOME_DIALOG(dlg), 3, TRUE);
  else
    gnome_dialog_set_sensitive(GNOME_DIALOG(dlg), 3, FALSE);
}

GtkWidget* sql_query_exec_new_dlg(SqlQuery *q)
{
  GtkWidget *dlg, *wid;
  gchar *str;

  str = g_strdup_printf(_("Execution of query %s"), q->name);
  dlg = gnome_dialog_new(str, GNOME_STOCK_BUTTON_CLOSE, _("Refresh"), 
			 _("Add an entry"), _("Modify"), _("Export"), NULL);
  gtk_window_set_policy(GTK_WINDOW(dlg), TRUE, TRUE, FALSE);
  gnome_dialog_set_default(GNOME_DIALOG(dlg), 0); /* Close Button */
  g_free(str);

  /* widgets at the top */
  wid = sql_query_exec_new(q);
  gtk_box_pack_start(GTK_BOX(GNOME_DIALOG(dlg)->vbox), 
		     wid, TRUE, TRUE, 0);
  gtk_widget_show(wid);
  
  gtk_signal_connect(GTK_OBJECT(dlg), "clicked",
		     GTK_SIGNAL_FUNC(main_dialog_clicked_cb),
		     wid);
  gtk_object_set_data(GTK_OBJECT(dlg), "SqlQueryExec", wid);

  /* Sensitiveness of the buttons */
  gnome_dialog_set_sensitive(GNOME_DIALOG(dlg), 2, FALSE);
  gnome_dialog_set_sensitive(GNOME_DIALOG(dlg), 3, FALSE);
  query_modif_table_cb(q, q->modif_table, dlg);
  query_insert_into_table_cb(SQL_QUERY_EXEC(wid), dlg);

  /* signals */
  gtk_signal_connect_while_alive(GTK_OBJECT(q->conf->srv), "conn_closed",
				 GTK_SIGNAL_FUNC(main_dialog_auto_close), dlg,
				 GTK_OBJECT(dlg));
  gtk_signal_connect_while_alive(GTK_OBJECT(q), "modif_table_changed",
				 GTK_SIGNAL_FUNC(query_modif_table_cb), dlg,
				 GTK_OBJECT(dlg));
  gtk_signal_connect(GTK_OBJECT(wid), "selection_changed",
		     GTK_SIGNAL_FUNC(query_insert_into_table_cb), dlg);

  return dlg;
}

static void main_dialog_clicked_cb(GnomeDialog * dialog, 
				   gint button_number, 
				   SqlQueryExec *qe)
{
  switch (button_number) {
  case 0: /* Close the Dlg */
    gnome_dialog_close(dialog);
    break;
  case 1: /* Refresh */
    ask_for_query_refresh(qe);
    break;
  case 2: /* add an entry */
    new_element_cb(NULL, qe);
    break;
  case 3: /* modify an entry */
    edit_element_cb(NULL, qe);
    break;
  case 4: /* export */
    break; 
  }
}

