/* sqlqueryform.c
 *
 * Copyright (C) 1999,2000 Vivien Malerba
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 */

#include <config.h>
#include "sqlqueryform.h"
#include "choicecombo.h"

static void      sql_query_form_class_init      (SqlQueryFormClass *class);
static void      sql_query_form_init            (SqlQueryForm *qf);
static void      sql_query_form_destroy         (GtkObject *object);
static void      sql_query_form_post_init       (SqlQueryForm *qf);


static void sql_access_conn_to_close_cb(GtkObject *obj, SqlQueryForm *qf);
static void sql_query_changed_cb(GtkObject *obj, SqlQueryForm *qf);
static void sql_query_modif_table_changed_cb(GtkObject *obj, SqlMemTable *t,
					     SqlQueryForm *qf);

/* modification of widgets is connected to this function */
static void sql_query_form_modif_cb(GtkObject *obj, SqlQueryForm *form);
static void sql_query_form_modif_2_cb(GtkObject *obj, gpointer ptr,
				      SqlQueryForm *form);


/* structures for each node in the form */
typedef enum {
  NODE_FIELD,
  NODE_LINK
} QueryNodeType;

typedef struct _ANode ANode;
struct _ANode {
  GtkWidget *widget; /* if NODE_LINK, only the nodes with possible othernodes
			will have widget!=NULL, to be able to have ONE ONLY
			ANode with a widget for a SqlMemFlinkink */
  QueryNodeType type;
  gint pos; /* num of column in the SqlQueryRes tabular */
  gboolean modified;
  QueryElement *elt; /* CAUTION: if NODE_LINK, elt will point to 
			the QueryElement which is represented, not 
			what is displayed in the ChoiceCombo widget */
  SqlMemFlink *link;
  GSList *choicelist; /* only for NODE_LINK */
  GSList *othernodes; /* list of the other nodes sharing the same table and
			 which are NODE_LINK */
};

/* structure to hold the information for each possible entry in a 
   ChoiceCombo */
typedef struct _ChoiceEntry ChoiceEntry;
struct _ChoiceEntry {
  gchar *linkval;
  gchar *printval;
};


/***************************************************************************/
/*                                                                         */
/* Main functions for the object                                           */
/*                                                                         */
/***************************************************************************/

/* get a pointer to the parents to be able to call their destructor */
static GtkObject *parent_class = NULL;

enum {
  CHANGED,
  OUTDATED,
  LAST_SIGNAL
};

static gint sql_query_form_signals[LAST_SIGNAL] = { 0, 0 };

guint sql_query_form_get_type(void) 
{
  static guint f_type = 0;
  
  if (!f_type)
    {
      GtkTypeInfo f_info =
      {
        "Sql_Query_Form",
        sizeof (SqlQueryForm),	
        sizeof (SqlQueryFormClass),
        (GtkClassInitFunc) sql_query_form_class_init,
        (GtkObjectInitFunc) sql_query_form_init,
        (GtkArgSetFunc) NULL,
        (GtkArgGetFunc) NULL
      };
      
      f_type = gtk_type_unique (gtk_vbox_get_type (), &f_info);
    }
  
  return f_type;   
}

static void sql_query_form_class_init(SqlQueryFormClass *class)
{
  GtkObjectClass *object_class;

  object_class = (GtkObjectClass*) class;   
  sql_query_form_signals[CHANGED] =
    gtk_signal_new ("changed",
                    GTK_RUN_FIRST,
                    object_class->type,
                    GTK_SIGNAL_OFFSET (SqlQueryFormClass, changed),
                    gtk_signal_default_marshaller, GTK_TYPE_NONE, 0);
  sql_query_form_signals[OUTDATED] =
    gtk_signal_new ("outdated",
                    GTK_RUN_FIRST,
                    object_class->type,
                    GTK_SIGNAL_OFFSET (SqlQueryFormClass, outdated),
                    gtk_signal_default_marshaller, GTK_TYPE_NONE, 0);

  gtk_object_class_add_signals (object_class, sql_query_form_signals, 
				LAST_SIGNAL);
  class->changed = NULL;
  class->outdated = NULL;

  parent_class = gtk_type_class (gtk_vbox_get_type());
  object_class->destroy = sql_query_form_destroy;
}

static void sql_query_form_init(SqlQueryForm *qf)
{

}

GtkWidget* sql_query_form_new(SqlQuery *q, gint numrow)
{
  GtkWidget *obj;
  SqlQueryForm *qf;

  obj =  GTK_WIDGET(gtk_type_new(sql_query_form_get_type()));
  qf = SQL_QUERY_FORM(obj);
  qf->q = q;
  sql_query_register_use_result(q, GTK_OBJECT(obj));
  qf->numrow = numrow;
  qf->modified = FALSE;
  qf->outdated = FALSE;
  sql_query_form_post_init(qf);

  /* connection to signals from SqlQuery */
  gtk_signal_connect_while_alive(GTK_OBJECT(q), "changed",
				 GTK_SIGNAL_FUNC(sql_query_changed_cb),
				 qf, GTK_OBJECT(qf));
  gtk_signal_connect_while_alive(GTK_OBJECT(q), "modif_table_changed",
			     GTK_SIGNAL_FUNC(sql_query_modif_table_changed_cb),
				 qf, GTK_OBJECT(qf));
  gtk_signal_connect_while_alive(GTK_OBJECT(q->conf->srv), "conn_to_close",
				 GTK_SIGNAL_FUNC(sql_access_conn_to_close_cb),
				 qf, GTK_OBJECT(qf));
  gtk_signal_connect_while_alive(GTK_OBJECT(q), "destroy",
				 GTK_SIGNAL_FUNC(sql_query_changed_cb),
				 qf, GTK_OBJECT(qf));
  return obj;
}


static gint m_GCompareFunc (ANode *a, ANode *b)
{
  if (a->pos < b->pos)
    return -1;
  if (a->pos > b->pos)
    return 1;
  return 0;
}


/* FIXME: NON TESTED */
void sql_query_form_refill_form(SqlQueryForm *qf)
{
  GSList *inform, *list2;
  ANode *node;
  gchar *str;
  GString *string;
  SqlDataDisplayFns *fns;
  Gda_Field *gfield;

  /* will change what is displayed in each widget of the form according
     to the numrow */
  g_warning("Function sql_query_form_refill_form called but not yet tested\n");
  inform = qf->inform;
  while (inform) {
    node = (ANode *)(inform->data);
    switch (node->type) {
    case NODE_FIELD:
      fns = sql_access_get_object_display_fns(qf->q->conf->srv,
	    GTK_OBJECT(SQL_MEM_FIELD(node->elt->main)->type));
      gfield = sql_query_res_get_gdafield(qf->q->res, qf->numrow, node->pos);
      (fns->gdafield_to_widget_update)(DATA_DISPLAY(node->widget), gfield);
      break;
    case NODE_LINK:
      fns = sql_access_get_object_display_fns(qf->q->conf->srv,
					  GTK_OBJECT(SQL_MEM_FIELD
						     (node->elt->main)->type));
      gfield = sql_query_res_get_gdafield(qf->q->res, qf->numrow, node->pos);
      str = (fns->gdafield_to_str)(gfield);
      if (str)
	{
	  string = g_string_new(str);
	  g_free(str);
	}
      else
	string = g_string_new("");
      list2 = node->othernodes;
      while (list2) {
	fns = sql_access_get_object_display_fns(qf->q->conf->srv,
		GTK_OBJECT(SQL_MEM_FIELD(((ANode *)
					  (list2->data))->elt->main)->type));
	gfield = sql_query_res_get_gdafield(qf->q->res, qf->numrow, 
					    ((ANode *)(list2->data))->pos);
	str = (fns->gdafield_to_str)(gfield);
	if (str)
	  {
	    g_string_sprintfa(string, " / %s", str);
	    g_free(str);
	  }
	else
	    g_string_append(string, " / -");
	list2 = g_slist_next(list2);
      }
      choice_combo_set_selection_txt(CHOICE_COMBO(node->widget), string->str);
      g_string_free(string, FALSE);
      break;
    }
    inform = g_slist_next(inform);
  }
}

static void sql_query_form_post_init(SqlQueryForm *qf)
{
  GtkWidget *table, *wid;
  GSList *list, *inform, *choicelist, *list2;
  gint i, nbrows;
  gboolean takeit;
  ANode *node;
  QueryElement *elt;
  QueryLink *qlink;
  ChoiceEntry *ce;
  gchar *str;
  SqlMemTable *t;
  SqlQueryRes *res;
  SqlDataDisplayFns *fns;
  Gda_Field *gfield;

  /* determining which element will be in the form */
  inform = NULL;
  i = 0;
  list = qf->q->objects;
  /* choose among the SqlQuery the objects (in q->objects) which
     will appear on the form for modification */
  while (list) {
    elt = (QueryElement *) (list->data);
    takeit = FALSE;
    
    if (elt->print_name) {
      if ((elt->type == QUERY_ELT_FIELD)) {
	SqlMemTable *t;
	t = sql_db_find_table_from_field(qf->q->conf->db, 
					 SQL_MEM_FIELD(elt->main));
	if (qf->q->modif_table && (t == qf->q->modif_table))
	  takeit = TRUE;
      }

      if (takeit) {
	node = (ANode *) g_malloc(sizeof(ANode));
	qf->nodes = g_slist_append(qf->nodes, node);
	node->pos = i;
	node->type = NODE_FIELD;
	node->elt = elt;
	node->link = NULL;
	node->modified = FALSE;
	node->othernodes = NULL;
	node->choicelist = NULL;
	node->widget = NULL;
	inform = g_slist_append(inform, node);
      }
    }
    i++;
    list = g_slist_next(list);
  }
  /* select some other fields of the table to be modified, but this
     time through a link (using the ChoiceCombo widget). These are fields
     belonging to the table to modify, for which some links from other tables
     come TO that field, and those links are declared as being used. */
  list = qf->q->depend_list;
  while (list) {
    qlink = (QueryLink *) (list->data);
    if (qlink->useit) {
      if (g_slist_find(qf->q->modif_table->fields, qlink->link->to)) {
	SqlMemTable *t;
	GSList *list2;
	/* does any field of the table containing the FROM field 
	   appear in the list of printed elements? */
	t = sql_db_find_table_from_field(qf->q->conf->db, qlink->link->from);
	if (t) {
	  list2 = t->fields;
	  while (list2) {
	    if ((i=sql_query_is_field_printed(qf->q, 
					   SQL_MEM_FIELD(list2->data)))>-1) {
	      node = (ANode *) g_malloc(sizeof(ANode));
	      qf->nodes = g_slist_append(qf->nodes, node);
	      node->pos = i;
	      node->type = NODE_LINK;
	      node->elt = sql_query_get_query_element(qf->q, i);
	      node->link = qlink->link;
	      node->modified = FALSE;
	      node->othernodes = NULL;
	      node->choicelist = NULL;
	      node->widget = NULL;
	      inform = g_slist_append(inform, node);
	    }
	    list2 = g_slist_next(list2);
	  }
	}
      }
    }
    list = g_slist_next(list);
  }

  /* reordering all the items to make them appear in the same order as they
     were specified in the SELECT query */
  inform = g_slist_sort(inform, (GCompareFunc) m_GCompareFunc);

  /* grouping all the NODE_LINK "ANodes" which actually refer to the same 
     table to avoid stupid display */
  nbrows = g_slist_length(inform);
  list = inform;
  while (list) {
    SqlMemTable *t1, *t2;

    node = (ANode *)(list->data);
    if ((node->type == NODE_LINK)  && (!node->modified)){
      t1 = sql_db_find_table_from_field(qf->q->conf->db,
					node->link->from);
      list2 = g_slist_next(list);
      while (list2) {
	ANode *node2;
	node2 = (ANode *)(list2->data);
	if ((node2->type == NODE_LINK) && (!node2->modified)){
	  t2 = sql_db_find_table_from_field(qf->q->conf->db,
		    node2->link->from);
	  if (t1 == t2) { /* lists need to be grouped */
	    node->othernodes = g_slist_append(node->othernodes, node2);
	    node2->modified = TRUE;
	  }
	}
	list2 = g_slist_next(list2); 
      }
    }
    list = g_slist_next(list);
  }
  /* removing any NODE_LINK "ANode" for which linklist is NULL */
  list = inform;
  while (list) {
    node = (ANode *)(list->data);
    if ((node->type == NODE_LINK) && (node->modified)){
      inform = g_slist_remove_link(inform, list);
      node->modified = FALSE;
      list = inform;
    }
    list = g_slist_next(list);
  }
  qf->inform = inform;

  /* Creating the table and the widgets in that table */
  nbrows = g_slist_length(inform);
  if (nbrows==0) {
    wid = gtk_label_new(_("There is nothing to be modified\n"
			  "in this query"));
    gtk_box_pack_start(GTK_BOX(qf), wid, FALSE, TRUE, GNOME_PAD);
    gtk_widget_show(wid);
  }
  else {
    table = gtk_table_new(nbrows, 2, FALSE);
    gtk_table_set_row_spacings(GTK_TABLE(table), GNOME_PAD);
    gtk_box_pack_start(GTK_BOX(qf), table, FALSE, TRUE, GNOME_PAD);
    gtk_widget_show(table);
    
    /* filling the table */
    list = inform;
    i = 0;
    while (list) {
      GString *string, *title, *order;

      node = (ANode *)(list->data);
      switch (node->type) {
      case NODE_FIELD: /*** FIELD ***/
	str = g_strdup_printf("%s: ", node->elt->name);
	wid = gtk_label_new(str);
	g_free(str);
	gtk_table_attach_defaults(GTK_TABLE(table), wid,
				  0, 1, i, i+1);
	gtk_widget_show(wid);
	fns = sql_access_get_object_display_fns(qf->q->conf->srv,
		    GTK_OBJECT(SQL_MEM_FIELD(node->elt->main)->type));
	gfield = sql_query_res_get_gdafield(qf->q->res, qf->numrow, node->pos);
	wid = GTK_WIDGET((fns->gdafield_to_widget)(gfield));
	gtk_table_attach_defaults(GTK_TABLE(table), wid,
				  1, 2, i, i+1);
	if (IS_DATA_DISPLAY(wid))
	  gtk_signal_connect(GTK_OBJECT(wid), "contents_modified",
			     GTK_SIGNAL_FUNC(sql_query_form_modif_cb), qf);
	else
	  g_warning("the returned widget is not of type DataDisplay");
	node->widget = wid;
	break;
      case NODE_LINK: /*** LINK ***/
	title = g_string_new(node->elt->name);

	/* build the list for that combo */
	/* table t common for the entire linklist */ 
	t = sql_db_find_table_from_field(qf->q->conf->db, 
					 node->link->from);
	order = g_string_new(SQL_MEM_FIELD(node->elt->main)->name);
	string = g_string_new("SELECT DISTINCT ");
	g_string_sprintfa(string, "%s, %s",
			  node->link->from->name, 
			  SQL_MEM_FIELD(node->elt->main)->name); 
	list2 = node->othernodes;
	while (list2) {
	  g_string_sprintfa(title, "/\n%s", 
			    ((ANode *)(list2->data))->elt->name);
	  g_string_sprintfa(string, ", %s", 
			    SQL_MEM_FIELD(((ANode *)(list2->data))->elt->main)->name);
	  g_string_sprintfa(order, ", %s", 
			    SQL_MEM_FIELD(((ANode *)(list2->data))->elt->main)->name);
	  list2 = g_slist_next(list2);
	}
	g_string_sprintfa(string, " FROM %s ORDER BY %s", t->name,
			  order->str);
	g_string_free(order, TRUE);
	res = sql_access_do_query(qf->q->conf->srv, string->str);
	g_string_free(string, TRUE);
	choicelist = NULL;
	if (res) {
	  gint i, nrows;
	  gint ncols=0, j;

	  ncols = g_slist_length(node->othernodes);
	  nrows = sql_query_res_get_nbtuples(res);
	  for (i=1; i<=nrows; i++) {
	    ce = (ChoiceEntry *) g_malloc(sizeof(ChoiceEntry));
	    fns = sql_access_get_object_display_fns(qf->q->conf->srv,
		       GTK_OBJECT(SQL_MEM_FIELD(node->elt->main)->type));
	    gfield = sql_query_res_get_gdafield(res, i, 0); 
	    ce->linkval = (fns->gdafield_to_str)(gfield);
	    if (!ce->linkval)
	      ce->linkval = g_strdup("");
	    gfield = sql_query_res_get_gdafield(res, i, 1); 
	    str = (fns->gdafield_to_str)(gfield);
	    if (str)
	      {
		string = g_string_new(str);
		g_free(str);
	      }
	    else
	      string = g_string_new("");
	    list2 = node->othernodes;
	    for (j=1; j<=ncols; j++) {
	      fns = sql_access_get_object_display_fns(qf->q->conf->srv,
		    GTK_OBJECT(SQL_MEM_FIELD(
		      ((ANode *)(list2->data))->elt->main)->type));
	      gfield = sql_query_res_get_gdafield(res, i, 1+j); 
	      str = (fns->gdafield_to_str)(gfield);
	      if (str)
		{
		  g_string_sprintfa(string, " / %s", str);
		  g_free(str);
		}
	      else
		g_string_append(string, " / -");
	      list2 = g_slist_next(list2);
	    }
	    ce->printval = string->str;
	    g_string_free(string, FALSE);
	    choicelist = g_slist_append(choicelist, ce);
	  }
	  sql_query_res_free(res);
	}

	g_string_sprintfa(title, ": ");
	wid = gtk_label_new(title->str);
	g_string_free(title, TRUE);
	gtk_table_attach_defaults(GTK_TABLE(table), wid,
				  0, 1, i, i+1);
	gtk_widget_show(wid);
	wid = choice_combo_new();

	/* setting the ChoiceCombo contents */
	choice_combo_set_content(CHOICE_COMBO(wid), choicelist,
				 GTK_STRUCT_OFFSET(ChoiceEntry, printval));
	node->choicelist = choicelist;
	/* choosing the right item to display in the ChoiceCombo */
	fns = sql_access_get_object_display_fns(qf->q->conf->srv,
			 GTK_OBJECT(SQL_MEM_FIELD(node->elt->main)->type));
	gfield = sql_query_res_get_gdafield(qf->q->res, qf->numrow, node->pos);
	str = (fns->gdafield_to_str)(gfield);
	if (str)
	  {
	    string = g_string_new(str);
	    g_free(str);
	  }
	else
	  string = g_string_new("");
	list2 = node->othernodes;
	while (list2) {
	  fns = sql_access_get_object_display_fns(qf->q->conf->srv,
	        GTK_OBJECT(SQL_MEM_FIELD(((ANode *)
					  (list2->data))->elt->main)->type));
	  gfield = sql_query_res_get_gdafield(qf->q->res, qf->numrow, 
					      ((ANode *)(list2->data))->pos);
	  str = (fns->gdafield_to_str)(gfield);
	  if (str)
	    {
	      g_string_sprintfa(string, " / %s", str);
	      g_free(str);
	    }
	  else
	    g_string_append(string, " / -");
	  list2 = g_slist_next(list2);
	}
	choice_combo_set_selection_txt(CHOICE_COMBO(wid), string->str);
	g_string_free(string, FALSE);

	gtk_table_attach_defaults(GTK_TABLE(table), wid,
				  1, 2, i, i+1);
	gtk_signal_connect(GTK_OBJECT(wid), "selection_changed",
			   GTK_SIGNAL_FUNC(sql_query_form_modif_2_cb), qf);
	node->widget = wid;
	break;
      default:
	wid = NULL; /* should never happen! */
	break;
      }

      gtk_object_set_data(GTK_OBJECT(wid), "node", node);
      gtk_widget_show(wid);
      list = g_slist_next(list);
      i++;
    }
  }
}

static void sql_access_conn_to_close_cb(GtkObject *obj, SqlQueryForm *qe)
{
  gtk_object_destroy(GTK_OBJECT(qe));
}

static void sql_query_form_destroy(GtkObject *object)
{
  SqlQueryForm *qf;
  GSList *list;

  g_return_if_fail (object != NULL);
  g_return_if_fail (IS_SQL_QUERY_FORM (object));

  qf = SQL_QUERY_FORM(object); 

  /* destroying the "ANode"s structure */
  if (qf->inform)
    g_slist_free(qf->inform);
  while (qf->nodes) {
    /* destroying the choicelist */
    list = ((ANode*)(qf->nodes->data))->choicelist;
    while(list) {
      g_free(((ChoiceEntry *)(list->data))->linkval);
      g_free(((ChoiceEntry *)(list->data))->printval);
      list = g_slist_remove_link(list, list);
    }
    g_free(qf->nodes->data);
    qf->nodes = g_slist_remove_link(qf->nodes, qf->nodes);
  }
  
  if (GTK_OBJECT_CLASS (parent_class)->destroy)
    (* GTK_OBJECT_CLASS (parent_class)->destroy) (object); 
}

static void sql_query_changed_cb(GtkObject *obj, SqlQueryForm *qf)
{
  GList *list;

  qf->outdated = TRUE;
  list = gtk_container_children(GTK_CONTAINER(qf));
  gtk_widget_set_sensitive(GTK_WIDGET(list->data), FALSE);
#ifdef debug_signal
  g_print(">> 'OUTDATED' from sqlqueryform->sql_query_changed_cb\n");
#endif
  gtk_signal_emit(GTK_OBJECT(qf), sql_query_form_signals[OUTDATED]);  
#ifdef debug_signal
  g_print("<< 'OUTDATED' from sqlqueryform->sql_query_changed_cb\n");
#endif
}

static void sql_query_modif_table_changed_cb(GtkObject *obj, SqlMemTable *t,
					     SqlQueryForm *qf)
					     
{
  sql_query_changed_cb(obj, qf);
}

/* called whenever an entry in the QueryForm is modified */
static void sql_query_form_modif_cb(GtkObject *obj, SqlQueryForm *form)
{
  ANode *node;
  GSList *list;

  node = gtk_object_get_data(GTK_OBJECT(obj), "node");
  node->modified = TRUE;
  list = node->othernodes;
  while (list) {
    ((ANode *)(list->data))->modified = TRUE;
    list = g_slist_next(list);
  }
#ifdef debug_signal
  g_print(">> 'CHANGED' from sql_query_form_modif_cb\n");
#endif
  gtk_signal_emit(GTK_OBJECT(form), sql_query_form_signals[CHANGED]);
#ifdef debug_signal
  g_print("<< 'CHANGED' from sql_query_form_modif_cb\n");
#endif
}

static void sql_query_form_modif_2_cb(GtkObject *obj, gpointer ptr,
				      SqlQueryForm *form)
{
  sql_query_form_modif_cb(obj, form);
}

/*************************************************************************/
/*                                                                       */
/* TO COMMIT THE CHANGES                                                 */
/*                                                                       */
/*************************************************************************/
static void dummy_cb(gint reply, gpointer data)
{
}

static void sql_query_form_commit_update(SqlQueryForm *qf)
{
  GSList *list;
  ANode *node;
  gchar *str;
  GString *query, *where, *tmpstr;
  gboolean firstup, firstwh;
  SqlQueryRes *res;  
  SqlDataDisplayFns *fns;
  Gda_Field *gfield;

  query = g_string_new("UPDATE ");
  where = g_string_new("WHERE ");
  g_string_sprintfa(query, "%s set ", qf->q->modif_table->name);
  firstup = TRUE;
  firstwh = TRUE;
  list = qf->nodes;
  while (list) {
    node = (ANode *)(list->data);
    
    switch (node->type) {
    case NODE_FIELD: /* node node->elt->name */
      if (node->modified) {
	/* UPDATE */
	/* set to str */
	fns = sql_access_get_object_display_fns(qf->q->conf->srv,
		    GTK_OBJECT(SQL_MEM_FIELD(node->elt->main)->type));
	str = (fns->widget_to_sql)(DATA_DISPLAY(node->widget));
	if (str)
	  {
	    if (firstup) 
	      {
		firstup = FALSE;
		g_string_sprintfa(query, "%s = %s", 
				  SQL_MEM_FIELD(node->elt->main)->name, str);
	      }
	    else 
	      g_string_sprintfa(query, ", %s = %s", 
				SQL_MEM_FIELD(node->elt->main)->name, str);
	    g_free(str);
	  }
      }
      /* WHERE */
      fns = sql_access_get_object_display_fns(qf->q->conf->srv,
		  GTK_OBJECT(SQL_MEM_FIELD(node->elt->main)->type));
      gfield = sql_query_res_get_gdafield(qf->q->res, qf->numrow, node->pos);
      str = (fns->gdafield_to_sql)(gfield);

      if (!str || !strcmp(str, "''")) { /* str == '' */
	if (firstwh)
	  firstwh = FALSE;
	else
	  g_string_sprintfa(where, "AND ");

	if (str)
	  g_string_sprintfa(where, "(%s = %s OR %s is null) ", 
			    SQL_MEM_FIELD(node->elt->main)->name, str,
			    SQL_MEM_FIELD(node->elt->main)->name);
	else
	  g_string_sprintfa(where, "%s is null ", 
			    SQL_MEM_FIELD(node->elt->main)->name);
      }
      else {                    /* str != '' */
	if (firstwh) {
	  firstwh = FALSE;
	  g_string_sprintfa(where, "%s = %s ", 
			    SQL_MEM_FIELD(node->elt->main)->name, str);
	}
	else
	  g_string_sprintfa(where, "AND %s = %s ", 
			    SQL_MEM_FIELD(node->elt->main)->name, str);
      }
      if (str) 
	g_free(str);
      break;
    case NODE_LINK:
      if (node->widget) {
	/* node node->link->to->name */
	if (node->modified) {
	  ChoiceEntry *ce;
	  ce = choice_combo_get_selection(CHOICE_COMBO(node->widget));
	  /* set to ce->linkval */
	  if (firstup) {
	    firstup = FALSE;
	    g_string_sprintfa(query, "%s = %s", node->link->to->name, 
			      ce->linkval);
	  }
	  else
	    g_string_sprintfa(query, ", %s = %s", node->link->to->name, 
			      ce->linkval);
	}
      }
      break;
    }
    
    list = g_slist_next(list);
  }
  g_string_sprintfa(query, " %s", where->str);
  /*g_print("\nQUERY: %s\n", query->str);*/
  tmpstr = g_string_new("SELECT * FROM ");
  g_string_sprintfa(tmpstr, "%s %s", qf->q->modif_table->name, where->str);
  g_string_free(where, TRUE);
  
  /*g_print("\nSELECT QUERY: %s\n", tmpstr->str);*/
  res = sql_access_do_query(qf->q->conf->srv, tmpstr->str);
  g_string_free(tmpstr, TRUE);
  if (res) {
    GtkWidget *dlg;
    gint i;

    i = sql_query_res_get_nbtuples(res);
    if (i>1)
      str = g_strdup_printf(_("%d tuples will be modified.\n"
			      "Do You want to continue modification?"), i);
    else
      str = g_strdup(_("One tuple will be modified.\n"
		       "Do You want to continue modification?"));
    dlg = gnome_question_dialog(str, (GnomeReplyCallback)dummy_cb, NULL);
    g_free(str);
    gtk_object_destroy(GTK_OBJECT(res));
    i = gnome_dialog_run_and_close(GNOME_DIALOG(dlg));
    if (!i) { /* saving */
      /*g_print("Query is #%s#\n", query->str);*/
      sql_access_do_query(qf->q->conf->srv, query->str);
      gtk_object_destroy(GTK_OBJECT(qf));
    }
  }
  
  g_string_free(query, TRUE);
}

static gboolean sql_query_form_commit_insert(SqlQueryForm *qf)
{
  GSList *list, *fields;
  ANode *node;
  gchar *str;
  GString *query, *vals, *errstr;
  gboolean firstin, firstv, retval=TRUE;
  SqlMemField *field;
  SqlMemSeq *seq;
  SqlDataDisplayFns *fns;

  fields = NULL;
  query = g_string_new("INSERT INTO ");
  vals = g_string_new("VALUES (");
  g_string_sprintfa(query, "%s (", qf->q->modif_table->name);
  firstin = TRUE;
  firstv = TRUE;
  list = qf->nodes;
  while (list) {
    node = (ANode *)(list->data);
    
    switch (node->type) {
    case NODE_FIELD: /* node node->elt->name */
      /* WHAT TO INSERT INTO */
      /* set to str */
      if (node->modified) 
	{
	  fns = sql_access_get_object_display_fns(qf->q->conf->srv,
	      GTK_OBJECT(SQL_MEM_FIELD(node->elt->main)->type));
	  str = (fns->widget_to_sql)(DATA_DISPLAY(node->widget));
	  if (str)
	    {
	      fields = g_slist_append(fields, node->elt->main);
	      if (firstin) 
		{
		  firstin = FALSE;
		  g_string_sprintfa(query, "%s", 
				    SQL_MEM_FIELD(node->elt->main)->name);
		}
	      else 
		g_string_sprintfa(query, ", %s", 
				  SQL_MEM_FIELD(node->elt->main)->name);
	      /* VALUES */
	      if (firstv) 
		{
		  firstv = FALSE;
		  g_string_sprintfa(vals, "%s", str);
		}
	      else
		g_string_sprintfa(vals, ", %s", str);
	      g_free(str);
	    }
	}
      break;
    case NODE_LINK:
      if (node->widget) 
	{
	  /* node node->link->to->name */
	  if (node->modified) 
	    {
	      ChoiceEntry *ce;
	      ce = choice_combo_get_selection(CHOICE_COMBO(node->widget));
	      fields = g_slist_append(fields, node->link->to);
	      /* set to ce->linkval */
	      if (firstin) 
		{
		  firstin = FALSE;
		  g_string_sprintfa(query, "%s", node->link->to->name);
		}
	      else
		g_string_sprintfa(query, ", %s", node->link->to->name);
	      if (firstv) 
		{
		  firstv = FALSE;
		  g_string_sprintfa(vals, "%s", ce->linkval);
		}
	      else
		g_string_sprintfa(vals, ", %s", ce->linkval);
	    }
	}
      break;
    }
    list = g_slist_next(list);
  }

  /* checking if the INSERT query is OK */
  errstr = g_string_new("");
  list = qf->q->modif_table->fields;
  firstv = TRUE;
  while (list) {
    field = SQL_MEM_FIELD(list->data);

    if (!g_slist_find(fields, field))
      {
	/* fields with a sequence link */
	if ((seq = sql_db_find_sequence_to_field(qf->q->conf->db, field))) 
	  {
	    g_string_sprintfa(query, ", %s", field->name);
	    g_string_sprintfa(vals, ", nextval('%s')", seq->name);
	    fields = g_slist_append(fields, field);
	  }
	
	/* default values:
	   ANY field with a default value is inserted into the query, so that
	   default values which are not hard_coded into the DB, but are there 
	   in gasql are taken into account. */
	if ((! g_slist_find(fields, field)) && (field->default_val)) 
	  {
	    g_string_sprintfa(query, ", %s", field->name);
	    g_string_sprintfa(vals, ", %s", field->default_val);
	    fields = g_slist_append(fields, field);
	  }

	if (! field->null_allowed) 
	  {
	    if (! g_slist_find(fields, field)) 
	      {
		gboolean error=TRUE;
		
		/* somehow other possibilities to modify the query go here */
		
		if (error) 
		  {
		    if (firstv) 
		      {
			firstv = FALSE;
			g_string_sprintfa(errstr, _("The field '%s' must "
						    "have a non NULL"
						    " value and is missing."),
					  field->name);
		      }
		    else
		      g_string_sprintfa(errstr, _("\nThe field '%s' must "
						  "have a non NULL"
						  " value and is missing."),
					field->name);
		  }
	      }
	  }
      }
    list = g_slist_next(list);
  }
  g_slist_free(fields);
    
  if (*(errstr->str) != '\0') {
    gnome_error_dialog(errstr->str);
    retval = FALSE;
  }
  else {
    g_string_sprintfa(query, ") %s)", vals->str);
    sql_access_do_query(qf->q->conf->srv, query->str);
  }
  g_string_free(errstr, TRUE);
  g_string_free(vals, TRUE);
  g_string_free(query, TRUE);

  return retval;
}

/* prepares the queries, ... and ask if really commit the changes */
void sql_query_form_commit(SqlQueryForm *qf)
{
  if (!qf->outdated) {
    if (qf->numrow < 0) {
      if (sql_query_form_commit_insert(qf))
	gtk_object_destroy(GTK_OBJECT(qf));
    }
    else
      sql_query_form_commit_update(qf);
  }
}



/****************************************************************************/
/*                                                                          */
/* Functions related to the Dialog in which is packed the SqlQueryForm obj  */
/*                                                                          */
/****************************************************************************/
static void form_changed_cb(SqlQueryForm *form, GnomeDialog *dlg);
static void form_outdated_cb(SqlQueryForm *form, GnomeDialog *dlg);
static void form_destroy_cb(GtkObject *obj, GnomeDialog *dlg);
static void main_dialog_clicked_cb(GnomeDialog * dialog, 
				   gint button_number, 
				   SqlQueryForm *qf);

static void main_dialog_auto_close(GtkObject *obj, GnomeDialog *dlg)
{
  SqlQueryForm *qf = gtk_object_get_data(GTK_OBJECT(dlg), "SqlQueryForm");
  main_dialog_clicked_cb(dlg, 0, qf);
}

GtkWidget* sql_query_form_new_dlg(SqlQuery *q, gint numrow, gboolean is_new)
{
  GtkWidget *dlg, *wid;

  if (is_new)
    dlg = gnome_dialog_new(_("New entry"), 
			   GNOME_STOCK_BUTTON_CLOSE, _("Commit"), NULL); 
  else
    dlg = gnome_dialog_new(_("Modification of entry"), 
			   GNOME_STOCK_BUTTON_CLOSE, _("Commit"), NULL); 
  gtk_window_set_policy(GTK_WINDOW(dlg), TRUE, TRUE, FALSE);
  gnome_dialog_set_default(GNOME_DIALOG(dlg), 0); /* Close Button */

  /* widgets at the top */
  wid = sql_query_form_new(q, numrow);
  gtk_box_pack_start(GTK_BOX(GNOME_DIALOG(dlg)->vbox), 
		     wid, TRUE, TRUE, 0);
  gtk_widget_show(wid);
  
  gtk_signal_connect_while_alive(GTK_OBJECT(wid), "destroy",
				 GTK_SIGNAL_FUNC(form_destroy_cb),
				 dlg, GTK_OBJECT(dlg));
  gtk_signal_connect(GTK_OBJECT(dlg), "clicked",
		     GTK_SIGNAL_FUNC(main_dialog_clicked_cb),
		     wid);
  gtk_object_set_data(GTK_OBJECT(dlg), "SqlQueryForm", wid);

  /* Sensitiveness of the buttons */
  gnome_dialog_set_sensitive(GNOME_DIALOG(dlg), 1, FALSE);

  /* signals */
  gtk_signal_connect_while_alive(GTK_OBJECT(q->conf->srv), "conn_closed",
				 GTK_SIGNAL_FUNC(main_dialog_auto_close), dlg,
				 GTK_OBJECT(dlg));
  gtk_signal_connect(GTK_OBJECT(wid), "changed",
		     GTK_SIGNAL_FUNC(form_changed_cb), dlg);
  gtk_signal_connect(GTK_OBJECT(wid), "outdated",
		     GTK_SIGNAL_FUNC(form_outdated_cb), dlg);

  return dlg;
}

static void form_destroy_cb(GtkObject *obj, GnomeDialog *dlg)
{
  gtk_object_destroy(GTK_OBJECT(dlg));
}

static void main_dialog_clicked_cb(GnomeDialog * dialog, 
				   gint button_number, 
				   SqlQueryForm *qf)
{
  switch (button_number) {
  case 0: /* Close the Dlg */
    gnome_dialog_close(dialog);
    break;
  case 1: /* Commit */
    sql_query_form_commit(qf);
    break;
  }
}

static void form_changed_cb(SqlQueryForm *form, GnomeDialog *dlg)
{
  gnome_dialog_set_sensitive(GNOME_DIALOG(dlg), 1, TRUE);
}

static void form_outdated_cb(SqlQueryForm *form, GnomeDialog *dlg)
{
  gnome_dialog_set_sensitive(GNOME_DIALOG(dlg), 1, FALSE);
}
