/* sqlquerywid.c
 *
 * Copyright (C) 1999 Vivien Malerba
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 */

#include <config.h>
#include "sqlquerywid.h"
#include "choicecombo.h"

static void      sql_query_wid_class_init      (SqlQueryWidClass *class);
static void      sql_query_wid_init            (SqlQueryWid *qw);
static void      sql_query_wid_destroy         (GtkObject *object);
static void      sql_query_wid_post_init(SqlQueryWid *qw);


static void sql_query_wid_elt_created_cb(GtkObject *obj, 
					 QueryElement *elt,
					 SqlQueryWid *qw);
static void sql_query_wid_elt_dropped_cb(GtkObject *obj, 
					 QueryElement *elt,
					 SqlQueryWid *qw);
static void sql_query_wid_elt_modified_cb(GtkObject *obj, 
					  QueryElement *elt,
					  SqlQueryWid *qw);
static void sql_query_wid_changed_cb(GtkObject *obj, 
				     SqlQueryWid *qw);

/* Global interface CBs */
static void query_name_or_descr_changed_cb(GtkWidget *wid, SqlQueryWid *qw);
static void tree_table_sel_changed_cb(GtkWidget *wid, SqlMemTable *t,
				      SqlQueryWid *qw);
static void wid_rel_show_table_cb(GtkObject *obj, SqlMemTable *t, 
				  SqlQueryWid *qw);
static void wid_rel_hide_table_cb(GtkObject *obj, SqlMemTable *t, 
				  SqlQueryWid *qw);

/* Objects page CBs */
static void obj_add_button(GtkWidget *widget, SqlQueryWid *qw);
static void obj_del_button(GtkWidget *widget, SqlQueryWid *qw);
static void object_clist_select_row_cb(GtkWidget *wid, 
				       gint row, gint column, 
				       GdkEventButton *event,
				       SqlQueryWid *qw);
static void object_edit_cb(GtkWidget *widget, SqlQueryWid *qw);
static void object_move_up_button_click_cb(GtkWidget *button, SqlQueryWid *qw);
static void object_move_dn_button_click_cb(GtkWidget *button, SqlQueryWid *qw);

/* Automatic dependencies page CBs */
static void depend_clist_select_row_cb(GtkWidget *wid, 
				       gint row, gint column, 
				       GdkEventButton *event,
				       SqlQueryWid *qw);
static void depend_update_one_entry(GtkWidget *button, SqlQueryWid *qw);
static void depend_add_to_list(GtkObject *obj, QueryLink *ql,
			       SqlQueryWid *qw);
static void depend_remove_from_list(GtkObject *obj, QueryLink *ql,
				    SqlQueryWid *qw);
static void refresh_depend(GtkWidget *bttn, SqlQueryWid *qw);
static void use_all_depend(GtkWidget *bttn, SqlQueryWid *qw);
static void use_none_depend(GtkWidget *bttn, SqlQueryWid *qw);
static void sql_query_wid_auto_links_add_table(SqlQueryWid *qw,SqlMemTable *t);
static void sql_query_wid_auto_links_del_table(SqlQueryWid *qw,SqlMemTable *t);


/* Text SQL only CBs */
static void text_only_tbutton_cb(GtkWidget *widget, SqlQueryWid *qw);
static void text_only_validate_cb(GtkWidget *widget, SqlQueryWid *qw);


static void object_edit_dialog_clicked_cb(GnomeDialog * dialog, 
					  gint button_number, 
					  SqlQueryWid *qw);

/***************************************************************************/
/*                                                                         */
/* Main functions for the object                                           */
/*                                                                         */
/***************************************************************************/

/* get a pointer to the parents to be able to call their destructor */
static GtkObject *parent_class = NULL;

guint sql_query_wid_get_type(void) 
{
  static guint f_type = 0;
  
  if (!f_type)
    {
      GtkTypeInfo f_info =
      {
        "Sql_Query_Wid",
        sizeof (SqlQueryWid),	
        sizeof (SqlQueryWidClass),
        (GtkClassInitFunc) sql_query_wid_class_init,
        (GtkObjectInitFunc) sql_query_wid_init,
        (GtkArgSetFunc) NULL,
        (GtkArgGetFunc) NULL
      };
      
      f_type = gtk_type_unique (gtk_vbox_get_type (), &f_info);
    }
  
  return f_type;   
}

static void sql_query_wid_class_init(SqlQueryWidClass *class)
{
  GtkObjectClass *object_class;
  
  object_class = (GtkObjectClass*) class;   
  parent_class = gtk_type_class (gtk_vbox_get_type());
  object_class->destroy = sql_query_wid_destroy;
}

static void sql_query_wid_init(SqlQueryWid *qw)
{
  qw->sel_table = NULL; /* last selection froml SqlWidDbTree */
  qw->obj_add_dlg = NULL; /* no dialog out right now */
  qw->depend_sel = -1;
}

static void button_tree_add_table_cb(GtkWidget *wid, SqlQueryWid *qw)
{
  if (qw->sel_table)
    sql_wid_db_rel_show_table(SQL_WID_DB_REL(qw->relview), qw->sel_table);
}

static void button_tree_del_table_cb(GtkWidget *wid, SqlQueryWid *qw)
{
  if (qw->sel_table)
    sql_wid_db_rel_hide_table(SQL_WID_DB_REL(qw->relview), qw->sel_table);
}

static void text_only_text_changed_cb(GtkWidget *widget,
				      SqlQueryWid *qw) 
{
  gtk_widget_set_sensitive(qw->text_valid_btn, TRUE);
}

static void sql_query_wid_post_init(SqlQueryWid *qw)
{
  GtkWidget *hb, *tree, *label, *vb, *button, *clist, *bb, *hp ,*vp;
  GtkWidget *table, *sw, *te, *text, *rel;
  GtkWidget *arrow, *wid;
  gchar *titles[]={N_("Object type"), N_("Object Name"), N_("Object content"), 
		   N_("Print Name")};
  gchar *titles2[]={N_("Used"), N_("Link From"), N_("Link To")}; 

  gchar *relname;
  gint i;
  GSList *list;
  
  table = gtk_table_new(2, 2, FALSE);
  gtk_box_pack_start(GTK_BOX(qw), table, FALSE, TRUE, GNOME_PAD);
  gtk_widget_show(table);

  /* Label of the query */
  label = gtk_label_new(_("Query name:"));
  gtk_table_attach(GTK_TABLE(table), label, 0, 1, 0, 1, 
		   0, 0, GNOME_PAD, 0);
  gtk_widget_show(label);

  te = gtk_entry_new();
  gtk_entry_set_text(GTK_ENTRY(te), qw->q->name); 
  gtk_signal_connect(GTK_OBJECT(te), "changed",
		     GTK_SIGNAL_FUNC(query_name_or_descr_changed_cb), qw);
  gtk_table_attach_defaults(GTK_TABLE(table), te, 1, 2, 0, 1);
  gtk_widget_show(te);
  qw->name = te;

  /* description of the query */
  label = gtk_label_new(_("Query description:"));
  gtk_table_attach(GTK_TABLE(table), label, 0, 1, 1, 2,
		   0, 0, GNOME_PAD, 0);
  gtk_widget_show(label);

  te = gtk_entry_new();
  if (qw->q->descr) 
    gtk_entry_set_text(GTK_ENTRY(te), qw->q->descr); 
  gtk_signal_connect(GTK_OBJECT(te), "changed",
		     GTK_SIGNAL_FUNC(query_name_or_descr_changed_cb), qw);
  gtk_table_attach_defaults(GTK_TABLE(table), te, 1, 2, 1, 2);
  gtk_widget_show(te);
  qw->descr = te;

  /* Main VPaned */
  vp = gtk_vpaned_new();
  gtk_box_pack_start(GTK_BOX(qw), vp, TRUE, TRUE, 0);
  gtk_container_set_border_width(GTK_CONTAINER(vp), GNOME_PAD);
  gtk_widget_show(vp);

  /* Top Hpaned */
  hp = gtk_hpaned_new();
  gtk_paned_pack1(GTK_PANED(vp), hp, TRUE, TRUE);
  gtk_container_set_border_width(GTK_CONTAINER(hp), GNOME_PAD);
  gtk_widget_show(hp);

  /* SqlWidDbTree */
  hb = gtk_hbox_new(FALSE, 0);
  gtk_paned_add1(GTK_PANED(hp), hb);
  gtk_widget_show(hb);

  tree = sql_wid_db_tree_new(qw->q->conf->db);
  sql_wid_db_tree_set_mode(SQL_WID_DB_TREE(tree) , 
			   SQL_WID_DB_TREE_TABLES |SQL_WID_DB_TREE_TABLES_SEL);
  gtk_signal_connect(GTK_OBJECT(tree), "table_selected",
		     GTK_SIGNAL_FUNC(tree_table_sel_changed_cb), qw);
  gtk_box_pack_start(GTK_BOX(hb), tree, TRUE, TRUE, 0);
  gtk_widget_show(tree);
  qw->tree = tree;

  bb = gtk_vbutton_box_new();
  gtk_button_box_set_child_size(GTK_BUTTON_BOX(bb), 15, 15);
  gtk_button_box_set_layout(GTK_BUTTON_BOX(bb), GTK_BUTTONBOX_SPREAD);
  gtk_box_pack_start(GTK_BOX(hb), bb, FALSE, TRUE, 0);
  gtk_widget_show(bb);

  button = gtk_button_new();
  gtk_container_add(GTK_CONTAINER(bb), button);
  gtk_widget_show(button);
  gtk_signal_connect(GTK_OBJECT(button), "clicked",
		     GTK_SIGNAL_FUNC(button_tree_add_table_cb), qw);

  arrow = gtk_arrow_new(GTK_ARROW_RIGHT, GTK_SHADOW_OUT);
  gtk_container_add(GTK_CONTAINER(button), arrow);
  gtk_widget_show(arrow);

  button = gtk_button_new();
  gtk_container_add(GTK_CONTAINER(bb), button);
  gtk_widget_show(button);
  gtk_signal_connect(GTK_OBJECT(button), "clicked",
		     GTK_SIGNAL_FUNC(button_tree_del_table_cb), qw);

  arrow = gtk_arrow_new(GTK_ARROW_LEFT, GTK_SHADOW_OUT);
  gtk_container_add(GTK_CONTAINER(button), arrow);
  gtk_widget_show(arrow);


  /* SqlWidDbRel Widget */
  relname = g_strdup_printf("Q%d", qw->q->id); 
  rel = sql_wid_db_rel_new(qw->q->conf, relname);
  g_free(relname);
  gtk_paned_add2(GTK_PANED(hp), rel);  
  gtk_widget_show(rel);
  qw->relview = rel;

  /* notebook */
  wid = gtk_notebook_new();
  gtk_paned_pack2(GTK_PANED(vp), wid, TRUE, TRUE);  
  gtk_notebook_set_tab_pos(GTK_NOTEBOOK(wid), GTK_POS_TOP);
  gtk_widget_show(wid);
  qw->notebook = wid;
  
  /* notebook Objects page */
  wid = gtk_label_new(_("Objects"));
  gtk_widget_show(wid);

  hb = gtk_hbox_new(FALSE, 0);
  gtk_container_set_border_width(GTK_CONTAINER(hb), GNOME_PAD);
  gtk_notebook_append_page(GTK_NOTEBOOK(qw->notebook), hb, wid);
  gtk_widget_show(hb);

  bb = gtk_vbutton_box_new();
  gtk_button_box_set_child_size(GTK_BUTTON_BOX(bb), 15, 15);
  gtk_container_set_border_width(GTK_CONTAINER(bb), GNOME_PAD);
  gtk_button_box_set_layout(GTK_BUTTON_BOX(bb), GTK_BUTTONBOX_SPREAD);
  gtk_box_pack_start(GTK_BOX(hb), bb, FALSE, TRUE, 0);
  gtk_widget_show(bb);

  button = gtk_button_new();
  gtk_container_add(GTK_CONTAINER(bb), button);
  gtk_widget_show(button);
  gtk_signal_connect(GTK_OBJECT(button), "clicked",
		     GTK_SIGNAL_FUNC(object_move_up_button_click_cb), qw);
  arrow = gtk_arrow_new(GTK_ARROW_UP, GTK_SHADOW_OUT);
  gtk_container_add(GTK_CONTAINER(button), arrow);
  gtk_widget_show(arrow);

  button = gtk_button_new();
  gtk_container_add(GTK_CONTAINER(bb), button);
  gtk_widget_show(button);
  gtk_signal_connect(GTK_OBJECT(button), "clicked",
		     GTK_SIGNAL_FUNC(object_move_dn_button_click_cb), qw);
  arrow = gtk_arrow_new(GTK_ARROW_DOWN, GTK_SHADOW_OUT);
  gtk_container_add(GTK_CONTAINER(button), arrow);
  gtk_widget_show(arrow);


  sw = gtk_scrolled_window_new(NULL, NULL);
  gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(sw),
				 GTK_POLICY_AUTOMATIC, GTK_POLICY_ALWAYS);
  gtk_box_pack_start(GTK_BOX(hb), sw, TRUE, TRUE, 0);
  gtk_widget_show(sw);

  clist = gtk_clist_new_with_titles(4, titles);
  gtk_clist_set_selection_mode(GTK_CLIST(clist), GTK_SELECTION_SINGLE);
  gtk_container_add(GTK_CONTAINER(sw), clist);
  gtk_widget_set_usize(clist, 400, 150);
  gtk_widget_show(clist);
  qw->clist = clist;
  for (i=0; i<4; i++) {
    gtk_clist_set_column_title(GTK_CLIST(clist), i, _(titles[i])); 
    gtk_clist_set_column_auto_resize(GTK_CLIST(clist), i, TRUE); 
  }
  gtk_clist_column_titles_passive(GTK_CLIST(clist));
  gtk_clist_set_reorderable(GTK_CLIST(clist), FALSE);
  gtk_signal_connect(GTK_OBJECT(clist), "select_row",
		     GTK_SIGNAL_FUNC(object_clist_select_row_cb), qw);

  bb = gtk_vbutton_box_new();
  gtk_container_set_border_width(GTK_CONTAINER(bb), GNOME_PAD);
  gtk_button_box_set_layout(GTK_BUTTON_BOX(bb), GTK_BUTTONBOX_SPREAD);
  gtk_box_pack_start(GTK_BOX(hb), bb, FALSE, TRUE, 0);
  gtk_widget_show(bb);
  
  wid = gtk_button_new_with_label(_("Add object"));
  gtk_container_add(GTK_CONTAINER(bb), wid);  
  gtk_widget_show(wid);
  qw->add_button = wid;
  gtk_signal_connect(GTK_OBJECT(wid), "clicked",
		     GTK_SIGNAL_FUNC(obj_add_button), qw);

  wid = gtk_button_new_with_label(_("Edit object"));
  gtk_container_add(GTK_CONTAINER(bb), wid);  
  gtk_widget_set_sensitive(wid, FALSE);
  gtk_widget_show(wid);
  qw->edit_button = wid;
  gtk_signal_connect(GTK_OBJECT(wid), "clicked",
		     GTK_SIGNAL_FUNC(object_edit_cb), qw);

  wid = gtk_button_new_with_label(_("Remove object"));
  gtk_container_add(GTK_CONTAINER(bb), wid);  
  gtk_widget_show(wid);
  qw->del_button = wid;
  gtk_widget_set_sensitive(qw->del_button, FALSE);
  gtk_signal_connect(GTK_OBJECT(wid), "clicked",
		     GTK_SIGNAL_FUNC(obj_del_button), qw);

  /* notebook Depend list */
  label = gtk_label_new(_("Automatic dependencies"));
  gtk_widget_show(label);

  hb = gtk_hbox_new(FALSE, 0);
  gtk_container_set_border_width(GTK_CONTAINER(hb), GNOME_PAD);
  gtk_notebook_append_page(GTK_NOTEBOOK(qw->notebook), hb, label);
  gtk_widget_show(hb);

  sw = gtk_scrolled_window_new(NULL, NULL);
  gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(sw),
				 GTK_POLICY_AUTOMATIC, GTK_POLICY_ALWAYS);
  gtk_box_pack_start(GTK_BOX(hb), sw, TRUE, TRUE, 0);
  gtk_widget_show(sw);

  wid = gtk_clist_new_with_titles(3, titles2);
  gtk_clist_set_selection_mode(GTK_CLIST(wid), GTK_SELECTION_SINGLE);
  gtk_clist_column_titles_passive(GTK_CLIST(wid));
  for (i=0; i<3; i++) {
    gtk_clist_set_column_title(GTK_CLIST(wid), i, _(titles2[i])); 
    gtk_clist_set_column_auto_resize(GTK_CLIST(wid), i, TRUE); 
  }
  gtk_container_add(GTK_CONTAINER(sw), wid);
  gtk_widget_show(wid);
  gtk_signal_connect(GTK_OBJECT(wid), "select_row",
		     GTK_SIGNAL_FUNC(depend_clist_select_row_cb), qw);
  qw->depend_clist = wid;

  vb = gtk_vbutton_box_new();
  gtk_container_set_border_width(GTK_CONTAINER(vb), GNOME_PAD);
  gtk_button_box_set_layout(GTK_BUTTON_BOX(vb), GTK_BUTTONBOX_SPREAD);
  gtk_box_pack_start(GTK_BOX(hb), vb, FALSE, FALSE, 0);
  gtk_widget_show(vb);

  wid = gtk_check_button_new_with_label(_("Use Link"));
  gtk_container_add(GTK_CONTAINER(vb), wid);  
  gtk_widget_show(wid);
  gtk_widget_set_sensitive(wid, FALSE);
  gtk_signal_connect(GTK_OBJECT(wid), "toggled",
		     GTK_SIGNAL_FUNC(depend_update_one_entry), qw);
  qw->use_button = wid;

  wid = gtk_button_new_with_label(_("Refresh"));
  gtk_container_add(GTK_CONTAINER(vb), wid);  
  gtk_signal_connect(GTK_OBJECT(wid), "clicked",
		     GTK_SIGNAL_FUNC(refresh_depend), qw);
  gtk_widget_show(wid);

  wid = gtk_button_new_with_label(_("Use All"));
  gtk_container_add(GTK_CONTAINER(vb), wid);  
  gtk_signal_connect(GTK_OBJECT(wid), "clicked",
		     GTK_SIGNAL_FUNC(use_all_depend), qw);
  gtk_widget_show(wid);

  wid = gtk_button_new_with_label(_("Use None"));
  gtk_signal_connect(GTK_OBJECT(wid), "clicked",
		     GTK_SIGNAL_FUNC(use_none_depend), qw);
  gtk_container_add(GTK_CONTAINER(vb), wid);  
  gtk_widget_show(wid);

  /* notebook Where page */
  label = gtk_label_new(_("Where clauses"));
  gtk_widget_show(label);

  wid = gtk_label_new("Not Yet...");
  gtk_notebook_append_page(GTK_NOTEBOOK(qw->notebook), wid, label);
  gtk_widget_show(wid);

  /* notebook Options page */
  label = gtk_label_new(_("Options"));
  gtk_widget_show(label);

  hb = gtk_hbox_new(TRUE, 0);
  gtk_container_set_border_width(GTK_CONTAINER(hb), GNOME_PAD);
  gtk_notebook_append_page(GTK_NOTEBOOK(qw->notebook), hb, label);
  gtk_widget_show(hb);
  
  vb = gtk_vbox_new(FALSE, 0);
  gtk_box_pack_start(GTK_BOX(hb), vb, FALSE, TRUE, 0);
  gtk_widget_show(vb);

  wid = gtk_label_new(_("Order by:"));
  gtk_box_pack_start(GTK_BOX(vb), wid, FALSE, TRUE, 0);
  gtk_widget_show(wid);

  wid = gtk_combo_new();
  gtk_box_pack_start(GTK_BOX(vb), wid, FALSE, TRUE, 0);
  gtk_widget_show(wid);
  qw->ordersel = wid;

  bb = gtk_hbutton_box_new();
  gtk_container_set_border_width(GTK_CONTAINER(bb), GNOME_PAD);
  gtk_button_box_set_child_size(GTK_BUTTON_BOX(bb), 15, 15);
  gtk_button_box_set_layout(GTK_BUTTON_BOX(bb), GTK_BUTTONBOX_SPREAD);
  gtk_box_pack_start(GTK_BOX(vb), bb, FALSE, TRUE, 0);
  gtk_widget_show(bb);

  button = gtk_button_new();
  gtk_container_add(GTK_CONTAINER(bb), button);
  gtk_widget_show(button);

  arrow = gtk_arrow_new(GTK_ARROW_UP, GTK_SHADOW_OUT);
  gtk_container_add(GTK_CONTAINER(button), arrow);
  gtk_widget_show(arrow);

  button = gtk_button_new();
  gtk_container_add(GTK_CONTAINER(bb), button);
  gtk_widget_show(button);

  arrow = gtk_arrow_new(GTK_ARROW_DOWN, GTK_SHADOW_OUT);
  gtk_container_add(GTK_CONTAINER(button), arrow);
  gtk_widget_show(arrow);

  sw = gtk_scrolled_window_new(NULL, NULL);
  gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(sw),
				 GTK_POLICY_AUTOMATIC, GTK_POLICY_ALWAYS);
  gtk_box_pack_start(GTK_BOX(vb), sw, TRUE, TRUE, 0);
  gtk_widget_show(sw);
  
  wid = gtk_clist_new(1);
  gtk_clist_set_selection_mode(GTK_CLIST(wid), GTK_SELECTION_SINGLE);
  gtk_clist_set_reorderable(GTK_CLIST(wid), TRUE);
  gtk_container_add(GTK_CONTAINER(sw), wid);
  gtk_widget_show(wid);
  qw->orderlist = wid;

  vb = gtk_vbox_new(FALSE, 0);
  gtk_box_pack_start(GTK_BOX(hb), vb, FALSE, TRUE, 0);
  gtk_widget_show(vb);

  wid = gtk_label_new(_("Group by:"));
  gtk_box_pack_start(GTK_BOX(vb), wid, FALSE, TRUE, 0);
  gtk_widget_show(wid);

  wid = gtk_combo_new();
  gtk_box_pack_start(GTK_BOX(vb), wid, FALSE, TRUE, 0);
  gtk_widget_show(wid);
  qw->groupsel = wid;

  bb = gtk_hbutton_box_new();
  gtk_button_box_set_child_size(GTK_BUTTON_BOX(bb), 15, 15);
  gtk_container_set_border_width(GTK_CONTAINER(bb), GNOME_PAD);
  gtk_button_box_set_layout(GTK_BUTTON_BOX(bb), GTK_BUTTONBOX_SPREAD);
  gtk_box_pack_start(GTK_BOX(vb), bb, FALSE, TRUE, 0);
  gtk_widget_show(bb);

  button = gtk_button_new();
  gtk_container_add(GTK_CONTAINER(bb), button);
  gtk_widget_show(button);
  /*gtk_signal_connect(GTK_OBJECT(button), "clicked",
		     GTK_SIGNAL_FUNC(button_add_table_cb), dlg);*/

  arrow = gtk_arrow_new(GTK_ARROW_UP, GTK_SHADOW_OUT);
  gtk_container_add(GTK_CONTAINER(button), arrow);
  gtk_widget_show(arrow);

  button = gtk_button_new();
  gtk_container_add(GTK_CONTAINER(bb), button);
  gtk_widget_show(button);
/*  gtk_signal_connect(GTK_OBJECT(button), "clicked",
		     GTK_SIGNAL_FUNC(button_del_table_cb), dlg);*/

  arrow = gtk_arrow_new(GTK_ARROW_DOWN, GTK_SHADOW_OUT);
  gtk_container_add(GTK_CONTAINER(button), arrow);
  gtk_widget_show(arrow);

  sw = gtk_scrolled_window_new(NULL, NULL);
  gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(sw),
				 GTK_POLICY_AUTOMATIC, GTK_POLICY_ALWAYS);
  gtk_box_pack_start(GTK_BOX(vb), sw, TRUE, TRUE, 0);
  gtk_widget_show(sw);
  
  wid = gtk_clist_new(1);
  gtk_clist_set_selection_mode(GTK_CLIST(wid), GTK_SELECTION_SINGLE);
  gtk_clist_set_reorderable(GTK_CLIST(wid), TRUE);
  gtk_container_add(GTK_CONTAINER(sw), wid);
  gtk_widget_show(wid);
  qw->grouplist = wid;


  /* Text notebook page */
  label = gtk_label_new(_("Text SQL"));
  gtk_widget_show(label);

  hb = gtk_hbox_new(FALSE, GNOME_PAD);
  gtk_container_set_border_width(GTK_CONTAINER(hb), GNOME_PAD);
  gtk_notebook_append_page(GTK_NOTEBOOK(qw->notebook), hb, label);
  gtk_widget_show(hb);

  sw = gtk_scrolled_window_new(NULL, NULL);
  gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(sw),
				 GTK_POLICY_AUTOMATIC, GTK_POLICY_ALWAYS);
  gtk_box_pack_start(GTK_BOX(hb), sw, TRUE, TRUE, 0);
  gtk_widget_show(sw);

  text = gtk_text_new(NULL, NULL);
  gtk_text_set_editable(GTK_TEXT(text), TRUE);
  gtk_text_set_word_wrap(GTK_TEXT(text), TRUE);
  gtk_container_add(GTK_CONTAINER(sw), text);
  gtk_widget_show(text);
  qw->text = text;
  gtk_text_set_editable(GTK_TEXT(qw->text), FALSE);
  gtk_signal_connect(GTK_OBJECT(text), "changed",
		     GTK_SIGNAL_FUNC(text_only_text_changed_cb), qw);

  vb = gtk_vbox_new(FALSE, GNOME_PAD);
  gtk_box_pack_start(GTK_BOX(hb), vb, FALSE, FALSE, 0);
  gtk_widget_show(vb);


  wid = gtk_check_button_new_with_label(_("Text Query Only"));
  gtk_box_pack_start(GTK_BOX(vb), wid, FALSE, FALSE, 0);
  gtk_widget_show(wid);
  qw->texttoggle = wid;

  wid = gtk_button_new_with_label(_("Validate"));
  gtk_box_pack_start(GTK_BOX(vb), wid, FALSE, FALSE, 0);
  gtk_widget_set_sensitive(wid, FALSE);
  gtk_widget_show(wid);
  qw->text_valid_btn = wid;
  gtk_signal_connect(GTK_OBJECT(wid), "clicked",
		     GTK_SIGNAL_FUNC(text_only_validate_cb), qw);

  /****************************************************/
  /* To take care of the state of the SqlQuery object */
  /****************************************************/

  /* if the query is text only */
  gtk_text_set_editable(GTK_TEXT(qw->text), qw->q->text_only);
  for (i=0; i<4; i++) {
    wid =  gtk_notebook_get_nth_page(GTK_NOTEBOOK(qw->notebook), i);
    gtk_widget_set_sensitive(wid, ! qw->q->text_only);
  }
  if (qw->q->text_only) {
    gtk_notebook_set_page(GTK_NOTEBOOK(qw->notebook), 4);
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(qw->texttoggle), TRUE);
    gtk_text_set_point(GTK_TEXT(qw->text), 0);
    gtk_text_insert(GTK_TEXT(qw->text), NULL, NULL, NULL, qw->q->text_sql, -1);
  }

  /* insertion of all the objects */
  list = qw->q->objects;
  while (list) {
    sql_query_wid_elt_created_cb(GTK_OBJECT(qw->q), 
				 (QueryElement *) (list->data), qw);
    list = g_slist_next(list);
  }

  /* putting the automatic links in the right CList */
  list = qw->q->depend_list;
  while (list) {
    depend_add_to_list(NULL, (QueryLink *) (list->data), qw);
    list = g_slist_next(list);
  }

  /* to tell the SqlQueryWid to write the right SQL text in the text widget */
  sql_query_wid_changed_cb(GTK_OBJECT(qw->q), qw);   
 
  /* late signal connexions */
  gtk_signal_connect(GTK_OBJECT(qw->texttoggle), "toggled",
		     GTK_SIGNAL_FUNC(text_only_tbutton_cb), qw);
}

GtkWidget* sql_query_wid_new(SqlQuery *q)
{
  GtkWidget *obj;
  SqlQueryWid *qw;

  obj =  GTK_WIDGET(gtk_type_new(sql_query_wid_get_type()));
  qw = SQL_QUERY_WID(obj);
  qw->q = q;
  sql_query_wid_post_init(qw);
  qw->edited_objects = NULL;
  /* connection to signals from SqlDb */
  gtk_signal_connect_while_alive(GTK_OBJECT(q), "elt_created",
				 GTK_SIGNAL_FUNC(sql_query_wid_elt_created_cb), 
				 qw, GTK_OBJECT(qw));
  gtk_signal_connect_while_alive(GTK_OBJECT(q), "elt_dropped",
				 GTK_SIGNAL_FUNC(sql_query_wid_elt_dropped_cb),
				 qw, GTK_OBJECT(qw));
  gtk_signal_connect_while_alive(GTK_OBJECT(q), "elt_modified",
				 GTK_SIGNAL_FUNC(sql_query_wid_elt_modified_cb), 
				 qw, GTK_OBJECT(qw));
  gtk_signal_connect_while_alive(GTK_OBJECT(q), "changed",
				 GTK_SIGNAL_FUNC(sql_query_wid_changed_cb),
				 qw, GTK_OBJECT(qw));

  gtk_signal_connect_while_alive(GTK_OBJECT(qw->relview), "table_shown",
				 GTK_SIGNAL_FUNC(wid_rel_show_table_cb), qw,
				 GTK_OBJECT(qw));

  gtk_signal_connect_while_alive(GTK_OBJECT(qw->relview), "table_hidden",
				 GTK_SIGNAL_FUNC(wid_rel_hide_table_cb), qw,
				 GTK_OBJECT(qw));

  gtk_signal_connect_while_alive(GTK_OBJECT(qw->q), "auto_link_added",
				 GTK_SIGNAL_FUNC(depend_add_to_list), qw,
				 GTK_OBJECT(qw));

  gtk_signal_connect_while_alive(GTK_OBJECT(qw->q), "auto_link_removed",
				 GTK_SIGNAL_FUNC(depend_remove_from_list), qw,
				 GTK_OBJECT(qw));

  return obj;
}

static void sql_query_wid_destroy(GtkObject *object)
{
  SqlQueryWid *qw;
  GSList *list;

  g_return_if_fail (object != NULL);
  g_return_if_fail (IS_SQL_QUERY_WID (object));

  qw = SQL_QUERY_WID(object); 

  /* removing the ADD dialog if opened */
  if (qw->obj_add_dlg)
    gtk_object_destroy(GTK_OBJECT(qw->obj_add_dlg));
  /* removing any Element Object modify dialog still opened */
  list = qw->edited_objects;
  while (list) {
    object_edit_dialog_clicked_cb(GNOME_DIALOG(list->data), 
				  1 /* CANCEL BUTTON */, qw);
    list = qw->edited_objects;
  }
  
  if (GTK_OBJECT_CLASS (parent_class)->destroy)
    (* GTK_OBJECT_CLASS (parent_class)->destroy) (object); 
}

static void object_clist_select_row_cb(GtkWidget *wid, 
				       gint row, gint column, 
				       GdkEventButton *event,
				       SqlQueryWid *qw)
{
  gtk_widget_set_sensitive(GTK_WIDGET(qw->del_button), TRUE);
  gtk_widget_set_sensitive(GTK_WIDGET(qw->edit_button), TRUE);
}


/***********************************************************************/
/*                                                                     */
/* Top of the interface callbacks                                      */
/*                                                                     */
/***********************************************************************/
static void query_name_or_descr_changed_cb(GtkWidget *wid, SqlQueryWid *qw)
{
  sql_query_set_name(qw->q, gtk_entry_get_text(GTK_ENTRY(qw->name)),
		     gtk_entry_get_text(GTK_ENTRY(qw->descr)));
}

static void tree_table_sel_changed_cb(GtkWidget *wid, SqlMemTable *t,
				      SqlQueryWid *qw)
{
  qw->sel_table = t;
}

static void wid_rel_show_table_cb(GtkObject *obj, SqlMemTable *t, 
				  SqlQueryWid *qw)
{
  sql_query_wid_auto_links_add_table(qw, t);
}

static void wid_rel_hide_table_cb(GtkObject *obj, SqlMemTable *t, 
				  SqlQueryWid *qw)
{
  sql_query_wid_auto_links_del_table(qw, t);
}

/***********************************************************************/
/*                                                                     */
/* Notebook Objects Page CBs                                           */
/*                                                                     */
/***********************************************************************/

static void object_add_button_dlg_clicked_cb(GnomeDialog * dialog, 
					     gint button_number, 
					     SqlQueryWid *qw);

static void object_add_type_choice_toggle_cb(GtkWidget *wid,
					     GnomeDialog *dlg);

static void object_add_close_dlg(SqlQueryWid *qw);

static void object_add_field_sel_cb(GtkWidget *widget, 
				    SqlMemTable *t, SqlMemField *f,
				    GnomeDialog *dlg);

static void object_add_set_function_list(SqlQueryWid *qw);

static void object_move_up_button_click_cb(GtkWidget *button, SqlQueryWid *qw)
{
  gint sel_row, nrows;
  GList *sel;

  sel = GTK_CLIST(qw->clist)->selection;
  if (sel)
    sel_row = GPOINTER_TO_INT(sel->data);    
  else
    sel_row = -1;
  nrows = GTK_CLIST(qw->clist)->rows;

  if ((sel_row!=-1) && (sel_row>0)) {
    gtk_clist_swap_rows(GTK_CLIST(qw->clist), sel_row, sel_row-1);
    sql_query_swap_objects(qw->q, sel_row, sel_row-1);
  }
}

static void object_move_dn_button_click_cb(GtkWidget *button, SqlQueryWid *qw)
{
  gint sel_row, nrows;
  GList *sel;

  sel = GTK_CLIST(qw->clist)->selection;
  if (sel)
    sel_row = GPOINTER_TO_INT(sel->data);    
  else
    sel_row = -1;
  nrows = GTK_CLIST(qw->clist)->rows;

  if ((sel_row!=-1) && (sel_row<nrows-1)) {
    gtk_clist_swap_rows(GTK_CLIST(qw->clist), sel_row, sel_row+1);
    sql_query_swap_objects(qw->q, sel_row, sel_row+1);
  }
}

static void object_add_user_fns_toggled_cb(GtkWidget *tbutton,
					   SqlQueryWid *qw)
{
  if (GTK_TOGGLE_BUTTON(tbutton)->active)
    gtk_object_set_data(GTK_OBJECT(qw->obj_add_dlg), "func_choice_user_bool", 
			qw->obj_add_dlg);
  else 
    gtk_object_set_data(GTK_OBJECT(qw->obj_add_dlg), "func_choice_user_bool", 
			NULL);
  object_add_set_function_list(qw);
}

static void object_add_system_fns_toggled_cb(GtkWidget *tbutton,
					     SqlQueryWid *qw)
{
  if (GTK_TOGGLE_BUTTON(tbutton)->active) 
    gtk_object_set_data(GTK_OBJECT(qw->obj_add_dlg), "func_choice_sys_bool", 
			qw->obj_add_dlg);
  else 
    gtk_object_set_data(GTK_OBJECT(qw->obj_add_dlg), "func_choice_sys_bool", 
			NULL);
  object_add_set_function_list(qw);
}

static void object_add_func_changed_cb(GtkWidget *wid, 
				       gpointer listdata, SqlQueryWid *qw)
{
  SqlDataFunction *f=NULL;
  gint i, limit;
  GtkWidget **args_tab;
  GSList *nlist, *elts;
  gboolean *ok_state, activate_ok = TRUE;

  args_tab = (GtkWidget **) gtk_object_get_data(GTK_OBJECT(qw->obj_add_dlg),
						"fn_args_tab");
  /* showing / hiding of arguments for that function */
  if (listdata) {
    f = SQL_DATA_FUNCTION(listdata);
    limit = g_slist_length(f->args);
  }
  else {
    limit = 0;
    activate_ok = FALSE;
  }

  for (i=0; i<limit ;i++) {
    gtk_widget_set_sensitive(args_tab[2*i], TRUE);
    gtk_widget_set_sensitive(args_tab[2*i+1], TRUE);
    /* possible choices for that argument */
    elts = qw->q->objects;
    nlist = NULL;
    while (elts) {
      SqlDataType *type;

      switch (((QueryElement *)(elts->data))->type) {
      case QUERY_ELT_FIELD: /* Table Field */
	type = SQL_MEM_FIELD(((QueryElement *)(elts->data))->main)->type;
	if (type == (g_slist_nth(f->args, i))->data) {
	  nlist = g_slist_append(nlist, elts->data);
	}
	break;
      case QUERY_ELT_FUNCTION: /* Function */
	type = SQL_DATA_FUNCTION(((QueryElement *)(elts->data))->main)->result_type;
	if (type == (g_slist_nth(f->args, i))->data) {
	  nlist = g_slist_append(nlist, elts->data);
	}
	break;
      case QUERY_ELT_AGGREGATE:
	break;
      }
      elts = g_slist_next(elts);
    }
    choice_combo_set_content(CHOICE_COMBO(args_tab[2*i+1]), nlist, 
			     GTK_STRUCT_OFFSET(QueryElement, name));
    if (!nlist) 
      activate_ok = FALSE;
    g_slist_free(nlist);
  }
  for (i=limit; i<8 ;i++) {
    gtk_widget_set_sensitive(args_tab[2*i], FALSE);
    gtk_widget_set_sensitive(args_tab[2*i+1], FALSE);
    choice_combo_set_content(CHOICE_COMBO(args_tab[2*i+1]), NULL, 0);
  }
  ok_state = gtk_object_get_data(GTK_OBJECT(qw->obj_add_dlg), "ok_state");
  ok_state[1] = activate_ok;
  gnome_dialog_set_sensitive(GNOME_DIALOG(qw->obj_add_dlg), 0, ok_state[1]);  
}

static void object_add_agg_changed_cb(GtkWidget *wid, 
				      gpointer listdata, SqlQueryWid *qw)
{
  SqlDataAggregate *agg = NULL;
  GtkWidget *arg_wid;
  GSList *nlist = NULL, *elts, *rlist;
  gboolean *ok_state, activate_ok = TRUE;

  /* showing / hiding of arguments for that function */
  if (listdata) 
    agg = SQL_DATA_AGGREGATE(listdata);
  else 
    activate_ok = FALSE;

  arg_wid = gtk_object_get_data(GTK_OBJECT(qw->obj_add_dlg), 
				"agg_choice_obj");  
  if (activate_ok) {
    /* rlist is the list of acceptable types for that aggregate name */
    rlist = NULL;
    nlist = qw->q->conf->srv->data_aggregates;
    while (nlist) {
      if (!g_slist_find(rlist, SQL_DATA_AGGREGATE(nlist->data)->arg_type) &&
	  !strcmp(SQL_DATA_AGGREGATE(nlist->data)->sqlname, agg->sqlname))
	rlist=g_slist_append(rlist, SQL_DATA_AGGREGATE(nlist->data)->arg_type);
      nlist = g_slist_next(nlist);
    }

    /* possible choices for that argument */
    elts = qw->q->objects;
    nlist = NULL;
    while (elts) {
      SqlDataType *type;
      
      switch (((QueryElement *)(elts->data))->type) {
      case QUERY_ELT_FIELD: /* Table Field */
	type = SQL_MEM_FIELD(((QueryElement *)(elts->data))->main)->type;
	if (g_slist_find(rlist, type)) {
	  nlist = g_slist_append(nlist, elts->data);
	}
	break;
      case QUERY_ELT_FUNCTION: /* Function */
	type = SQL_DATA_FUNCTION(((QueryElement *)(elts->data))->main)->result_type;
	if (g_slist_find(rlist, type)) {
	  nlist = g_slist_append(nlist, elts->data);
	}
	break;
      case QUERY_ELT_AGGREGATE:
	break;
      }
      elts = g_slist_next(elts);
    }
    g_slist_free(rlist);
  }
  choice_combo_set_content(CHOICE_COMBO(arg_wid), nlist, 
			   GTK_STRUCT_OFFSET(QueryElement, name));
  if (!nlist) 
    activate_ok = FALSE;
  else
    g_slist_free(nlist);
  
  ok_state = gtk_object_get_data(GTK_OBJECT(qw->obj_add_dlg), "ok_state");
  ok_state[2] = activate_ok;
  gnome_dialog_set_sensitive(GNOME_DIALOG(qw->obj_add_dlg), 0, ok_state[2]);  
}

static void object_add_set_function_list(SqlQueryWid *qw)
{
  gboolean sys, user;
  GSList *list, *lptr;
  GtkWidget *combo;

  combo = GTK_WIDGET(gtk_object_get_data(GTK_OBJECT(qw->obj_add_dlg), 
			      "func_choice_combo"));
  sys = gtk_object_get_data(GTK_OBJECT(qw->obj_add_dlg), 
			    "func_choice_sys_bool") ? TRUE : FALSE;
  user = gtk_object_get_data(GTK_OBJECT(qw->obj_add_dlg), 
			     "func_choice_user_bool") ? TRUE : FALSE;
  if (sys && user)
    choice_combo_set_content(CHOICE_COMBO(combo), 
			     qw->q->conf->srv->data_functions, 
			     GTK_STRUCT_OFFSET(SqlDataFunction, sqlname));  
  else {
    list = NULL;
    lptr = qw->q->conf->srv->data_functions;
    while (lptr) {
      if ((SQL_DATA_FUNCTION(lptr->data)->is_user && user) ||
	  (!SQL_DATA_FUNCTION(lptr->data)->is_user && sys))
	list = g_slist_append(list, lptr->data);
      lptr = g_slist_next(lptr);
    }
    choice_combo_set_content(CHOICE_COMBO(combo), list, 
			     GTK_STRUCT_OFFSET(SqlDataFunction, sqlname));  
    g_slist_free(list);
  }
}

static void object_add_set_agg_list(SqlQueryWid *qw)
{
  GtkWidget *caggs;
  GSList *list, *nlist;

  caggs = GTK_WIDGET(gtk_object_get_data(GTK_OBJECT(qw->obj_add_dlg), 
					 "agg_choice_agg"));
  /* aggregates lists (one per aggregate name) */
  list = qw->q->conf->srv->data_aggregates;
  nlist = NULL;
  while (list) {
    if (! nlist || 
	(strcmp(SQL_DATA_AGGREGATE(g_slist_last(nlist)->data)->sqlname,
		SQL_DATA_AGGREGATE(list->data)->sqlname)))
      nlist = g_slist_append(nlist, list->data);
    list = g_slist_next(list);
  }
  choice_combo_set_content(CHOICE_COMBO(caggs), nlist, 
			   GTK_STRUCT_OFFSET(SqlDataAggregate, sqlname));
  g_slist_free(nlist);
}

static void object_add_data_functions_add_or_del_cb(GtkObject *obj, 
						    SqlDataFunction *func,
						    SqlQueryWid *qw)
{
  object_add_set_function_list(qw);
}


static void object_add_data_aggregates_add_or_del_cb(GtkObject *obj, 
						     SqlDataAggregate *agg,
						     SqlQueryWid *qw)
{
  object_add_set_agg_list(qw);
}


static void obj_add_button(GtkWidget *widget, SqlQueryWid *qw)
{
  GtkWidget *dlg, *wid, *hb, *vb, *sw, *button, *table;
  GtkWidget *rb1, *rb2, *rb3;
  gchar *str;
  GSList *grp;
  gboolean *ok_state;
  GtkWidget **args_tab;
  gint i;
  
  if (! qw->obj_add_dlg) {
    ok_state = (gboolean *) g_malloc(sizeof(gboolean)*3);
    ok_state[0] = FALSE;
    ok_state[1] = FALSE;
    ok_state[2] = FALSE;
    args_tab = (GtkWidget **) g_malloc(sizeof(gpointer)*16);

    str = g_strdup_printf(_("Adding an object to the query %s"), qw->q->name);
    dlg = gnome_dialog_new(str, GNOME_STOCK_BUTTON_OK, 
			   GNOME_STOCK_BUTTON_CANCEL, NULL);
    gtk_widget_set_usize(dlg, 400, 350); 
    gtk_window_set_policy(GTK_WINDOW(dlg), TRUE, TRUE, FALSE);
    g_free(str);
    qw->obj_add_dlg = dlg;
    gtk_object_set_data(GTK_OBJECT(dlg), "ok_state", ok_state);
    gtk_object_set_data(GTK_OBJECT(dlg), "fn_args_tab", args_tab);
    gnome_dialog_set_sensitive(GNOME_DIALOG(dlg), 0, ok_state[0]);

    /* top frame */
    hb = gtk_hbox_new(FALSE, GNOME_PAD);
    gtk_box_pack_start(GTK_BOX(GNOME_DIALOG(dlg)->vbox), 
		       hb, FALSE, TRUE, 0);
    gtk_widget_show(hb);
    
    vb = gtk_vbox_new(TRUE, 0);
    gtk_box_pack_start(GTK_BOX(hb), vb, FALSE, TRUE, 0);
    gtk_widget_show(vb);
    
    button = gtk_radio_button_new_with_label(NULL, _("Add a table field"));
    gtk_box_pack_start(GTK_BOX(vb), button, TRUE, TRUE, 0);
    gtk_widget_show(button);
    grp = gtk_radio_button_group(GTK_RADIO_BUTTON(button));
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(button), TRUE);
    gtk_signal_connect(GTK_OBJECT(button), "toggled",
		       GTK_SIGNAL_FUNC(object_add_type_choice_toggle_cb), dlg);
    rb1 = button;
    gtk_object_set_data(GTK_OBJECT(dlg), "rb1", rb1);
    
    button = gtk_radio_button_new_with_label(grp, _("Add a function"));
    grp = gtk_radio_button_group(GTK_RADIO_BUTTON(button));
    gtk_box_pack_start(GTK_BOX(vb), button, TRUE, TRUE, 0);
    gtk_widget_show(button);
    gtk_signal_connect(GTK_OBJECT(button), "toggled",
		       GTK_SIGNAL_FUNC(object_add_type_choice_toggle_cb), dlg);
    rb2 = button;
    gtk_object_set_data(GTK_OBJECT(dlg), "rb2", rb2);
    
    button = gtk_radio_button_new_with_label(grp, _("Add an aggregate"));
    gtk_box_pack_start(GTK_BOX(vb), button, TRUE, TRUE, 0);
    gtk_widget_show(button);
    gtk_signal_connect(GTK_OBJECT(button), "toggled",
		       GTK_SIGNAL_FUNC(object_add_type_choice_toggle_cb), dlg);
    rb3 = button;
    gtk_object_set_data(GTK_OBJECT(dlg), "rb3", rb3);
    
    table = gtk_table_new(2, 2, FALSE);
    gtk_container_set_border_width(GTK_CONTAINER(table), GNOME_PAD);
    gtk_box_pack_start(GTK_BOX(hb), table, FALSE, FALSE, GNOME_PAD);
    gtk_table_set_row_spacings(GTK_TABLE(table), GNOME_PAD);
    gtk_widget_show(table);
    
    wid = gtk_label_new(_("Object Name:"));
    gtk_table_attach_defaults(GTK_TABLE(table), wid, 0, 1, 0, 1);
    gtk_widget_show(wid);
    
    wid = gtk_entry_new();
    gtk_table_attach_defaults(GTK_TABLE(table), wid, 1, 2, 0, 1);
    gtk_widget_show(wid);
    gtk_object_set_data(GTK_OBJECT(dlg), "obj_name_wid", wid);
    
    wid = gtk_label_new(_("Print Name:"));
    gtk_table_attach_defaults(GTK_TABLE(table), wid, 0, 1, 1, 2);
    gtk_widget_show(wid);
    
    wid = gtk_entry_new();
    gtk_table_attach_defaults(GTK_TABLE(table), wid, 1, 2, 1, 2);
    gtk_widget_show(wid);
    gtk_object_set_data(GTK_OBJECT(dlg), "obj_pname_wid", wid);
    
    /* lower part of the DLG */
    sw = gtk_scrolled_window_new(NULL, NULL);
    gtk_box_pack_start(GTK_BOX(GNOME_DIALOG(dlg)->vbox), 
		       sw, TRUE, TRUE, 0);
    gtk_widget_show(sw);
    
    /* add all in the SW inside this VB */
    vb = gtk_vbox_new(FALSE, GNOME_PAD);
    gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(sw), 
				   GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
    gtk_scrolled_window_add_with_viewport(GTK_SCROLLED_WINDOW(sw), vb);  
    gtk_widget_show(vb);

    /* for the field types */
    wid = sql_wid_db_tree_new(qw->q->conf->db);
    gtk_widget_set_usize(GTK_WIDGET(wid), 200, 100);
    sql_wid_db_tree_set_mode(SQL_WID_DB_TREE(wid), SQL_WID_DB_TREE_FIELDS | 
			     SQL_WID_DB_TREE_FIELDS_SEL);
    gtk_signal_connect(GTK_OBJECT(wid), "field_selected",
		       GTK_SIGNAL_FUNC(object_add_field_sel_cb), dlg);

    gtk_box_pack_start(GTK_BOX(vb), wid, TRUE, TRUE, 0);
    gtk_widget_show(wid);
    gtk_object_set_data(GTK_OBJECT(rb1), "sh_wid", wid);
    
    
    /* for the functions */
    hb = gtk_hbox_new(FALSE, GNOME_PAD);
    gtk_box_pack_start(GTK_BOX(vb), hb, FALSE, FALSE, 0);
    gtk_object_set_data(GTK_OBJECT(rb2), "sh_wid", hb);
    
    button = gtk_check_button_new_with_label(_("User defined functions"));
    gtk_box_pack_start(GTK_BOX(hb), button, FALSE, FALSE, 0);
    gtk_widget_show(button);
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(button), TRUE); /*default*/
    gtk_object_set_data(GTK_OBJECT(dlg), "func_choice_user_bool", dlg);
    gtk_signal_connect(GTK_OBJECT(button), "toggled",
		       GTK_SIGNAL_FUNC(object_add_user_fns_toggled_cb), qw);
    
    button = gtk_check_button_new_with_label(_("System functions"));
    gtk_box_pack_start(GTK_BOX(hb), button, FALSE, FALSE, 0);
    gtk_widget_show(button);
    gtk_signal_connect(GTK_OBJECT(button), "toggled",
		       GTK_SIGNAL_FUNC(object_add_system_fns_toggled_cb), qw);
    
    table = gtk_table_new(9, 2, FALSE);
    gtk_box_pack_start(GTK_BOX(vb), table, FALSE, FALSE, GNOME_PAD);
    gtk_table_set_row_spacings(GTK_TABLE(table), GNOME_PAD);
    gtk_object_set_data(GTK_OBJECT(rb2), "sh_wid2", table);
    
    wid = gtk_label_new(_("Function:"));
    gtk_table_attach_defaults(GTK_TABLE(table), wid, 0, 1, 0, 1);
    gtk_widget_show(wid);
    
    wid = choice_combo_new();
    gtk_table_attach_defaults(GTK_TABLE(table), wid, 1, 2, 0, 1);
    gtk_widget_show(wid);

    gtk_object_set_data(GTK_OBJECT(dlg), "func_choice_combo", wid);
    gtk_signal_connect(GTK_OBJECT(wid), "selection_changed",
		       GTK_SIGNAL_FUNC(object_add_func_changed_cb), qw);

    for (i=0; i<8; i++) {
      str = g_strdup_printf(_("Argument #%d:"), i+1);
      wid = gtk_label_new(str);
      g_free(str);
      gtk_table_attach_defaults(GTK_TABLE(table), wid, 0, 1, 1+i, 2+i); 
      args_tab[2*i] = wid;      
      gtk_widget_show(wid);

      wid = choice_combo_new();
      gtk_table_attach_defaults(GTK_TABLE(table), wid, 1, 2, 1+i, 2+i); 
      args_tab[2*i+1] = wid;
      gtk_widget_show(wid);
    }

    object_add_set_function_list(qw);
    
    /* for the aggregates */
    table = gtk_table_new(2, 2, FALSE);
    gtk_box_pack_start(GTK_BOX(vb), table, FALSE, FALSE, GNOME_PAD);
    gtk_table_set_row_spacings(GTK_TABLE(table), GNOME_PAD);
    gtk_object_set_data(GTK_OBJECT(rb3), "sh_wid", table);
    
    wid = gtk_label_new(_("Aggregate:"));
    gtk_table_attach_defaults(GTK_TABLE(table), wid, 0, 1, 0, 1);
    gtk_widget_show(wid);
    
    wid = choice_combo_new();
    gtk_table_attach_defaults(GTK_TABLE(table), wid, 1, 2, 0, 1);
    gtk_widget_show(wid);
    gtk_object_set_data(GTK_OBJECT(dlg), "agg_choice_agg", wid);
    gtk_signal_connect(GTK_OBJECT(wid), "selection_changed",
		       GTK_SIGNAL_FUNC(object_add_agg_changed_cb), qw);
    
    wid = gtk_label_new(_("Apply to field:"));
    gtk_table_attach_defaults(GTK_TABLE(table), wid, 0, 1, 1, 2);
    gtk_widget_show(wid);
    
    wid = choice_combo_new();
    gtk_table_attach_defaults(GTK_TABLE(table), wid, 1, 2, 1, 2);
    gtk_object_set_data(GTK_OBJECT(dlg), "agg_choice_obj", wid);
    gtk_widget_show(wid);

    object_add_set_agg_list(qw);
    
    /* DLG signal */
    gtk_signal_connect(GTK_OBJECT(dlg), "clicked",
		       GTK_SIGNAL_FUNC(object_add_button_dlg_clicked_cb),
		       qw); 
    /* connection to signals from SqlAccess */
    gtk_signal_connect_while_alive(GTK_OBJECT(qw->q->conf->srv), 
		 "data_function_added",
	         GTK_SIGNAL_FUNC(object_add_data_functions_add_or_del_cb), qw,
		 GTK_OBJECT(qw));
    gtk_signal_connect_while_alive(GTK_OBJECT(qw->q->conf->srv), 
		 "data_function_removed",
		 GTK_SIGNAL_FUNC(object_add_data_functions_add_or_del_cb), qw,
		 GTK_OBJECT(qw));
    gtk_signal_connect_while_alive(GTK_OBJECT(qw->q->conf->srv), 
		 "data_aggregate_added",
	         GTK_SIGNAL_FUNC(object_add_data_aggregates_add_or_del_cb), qw,
		 GTK_OBJECT(qw));
    gtk_signal_connect_while_alive(GTK_OBJECT(qw->q->conf->srv), 
		 "data_aggregate_removed",
		 GTK_SIGNAL_FUNC(object_add_data_aggregates_add_or_del_cb), qw,
		 GTK_OBJECT(qw));
    gtk_widget_show(dlg);

    gtk_widget_set_sensitive(GTK_WIDGET(qw->add_button), FALSE);
  }
}

static void object_add_button_dlg_clicked_cb(GnomeDialog * dialog, 
					     gint button_number, 
					     SqlQueryWid *qw)
{
  gboolean doremove = TRUE; /* TRUE => object to be created */
  GtkWidget *wid;
  gchar *namestr, *pnamestr;
  GSList *list;

  switch (button_number) {
  case 0: /* Ok button, verifications before adding the object */
    /* checking the object name */
    wid = gtk_object_get_data(GTK_OBJECT(qw->obj_add_dlg), "obj_name_wid");
    namestr = gtk_entry_get_text(GTK_ENTRY(wid));
    list = qw->q->objects;
    while (list && doremove) {
      if (!strcmp(((QueryElement *)(list->data))->name, namestr)) {
	doremove = FALSE;
	gnome_app_error(GNOME_APP(qw->q->conf->app), 
			_("The object name is already used"));
      }
      list = g_slist_next(list);
    }
    if (doremove && (*namestr == '\0')) {
      gnome_app_error(GNOME_APP(qw->q->conf->app), 
		      _("The object name is not valid"));
      doremove = FALSE;
    }
    /* object print name */
    if (doremove) {
      wid = gtk_object_get_data(GTK_OBJECT(qw->obj_add_dlg), "obj_pname_wid");
      pnamestr = gtk_entry_get_text(GTK_ENTRY(wid));
      if (strpbrk(pnamestr, " .'\"\\-(){}[]|@")) {
	gnome_app_error(GNOME_APP(qw->q->conf->app), 
			_("The object print name is not valid"));
	doremove = FALSE;
      }
    }
    if (doremove) {
      /* Table Field Object */
      wid = gtk_object_get_data(GTK_OBJECT(qw->obj_add_dlg), "rb1");
      if (GTK_TOGGLE_BUTTON(wid)->active) {
	SqlMemField *field;
	
	field = gtk_object_get_data(GTK_OBJECT(qw->obj_add_dlg), 
				    "field_selected");
	/* NB: field cannot be null since OK was pressed */
	sql_query_new_elt_field(qw->q, namestr, pnamestr, field);
      }
      /* Function Object */
      wid = gtk_object_get_data(GTK_OBJECT(qw->obj_add_dlg), "rb2");
      if (GTK_TOGGLE_BUTTON(wid)->active) {
	SqlDataFunction *f;
	GSList *args;
	GtkWidget **args_tab;
	gint i, nbargs;

	wid = GTK_WIDGET(gtk_object_get_data(GTK_OBJECT(qw->obj_add_dlg), 
					     "func_choice_combo"));
	args_tab = (GtkWidget **) 
	  gtk_object_get_data(GTK_OBJECT(qw->obj_add_dlg), "fn_args_tab");
	f = SQL_DATA_FUNCTION(choice_combo_get_selection(CHOICE_COMBO(wid)));
	nbargs = g_slist_length(f->args);
	args = NULL;
	for (i=0; i<nbargs; i++) {
	  args = g_slist_append(args, 
		 choice_combo_get_selection(CHOICE_COMBO(args_tab[2*i+1])));
	}
	sql_query_new_elt_function(qw->q, namestr, pnamestr, f, args);
	g_slist_free(args);
      }
      /* Aggregate Object */
      wid = gtk_object_get_data(GTK_OBJECT(qw->obj_add_dlg), "rb3");
      if (GTK_TOGGLE_BUTTON(wid)->active) {
	GtkWidget *combo, *value;
	SqlDataAggregate *a, *agg = NULL;
	SqlDataType *v = NULL;
	QueryElement *elt;

	combo = gtk_object_get_data(GTK_OBJECT(qw->obj_add_dlg), 
				    "agg_choice_agg");
	value = gtk_object_get_data(GTK_OBJECT(qw->obj_add_dlg), 
				    "agg_choice_obj");

	a =SQL_DATA_AGGREGATE(choice_combo_get_selection(CHOICE_COMBO(combo)));
	elt =(QueryElement *)(choice_combo_get_selection(CHOICE_COMBO(value)));
	switch (elt->type) {
	case QUERY_ELT_FIELD:
	  v =SQL_MEM_FIELD(elt->main)->type;
	  break;
	case QUERY_ELT_FUNCTION:
	  v = SQL_DATA_FUNCTION(elt->main)->result_type;
	  break;
	case QUERY_ELT_AGGREGATE: /* cannot occur */
	  break;
	}

	list = qw->q->conf->srv->data_aggregates;
	while (list && !agg) {
	  if ((SQL_DATA_AGGREGATE(list->data)->arg_type == v) && 
	      (!strcmp(SQL_DATA_AGGREGATE(list->data)->sqlname, a->sqlname)))
	    agg = SQL_DATA_AGGREGATE(list->data);
	  list = g_slist_next(list);
	}
	if (agg) {
	  list = g_slist_append(NULL, elt);
	  sql_query_new_elt_aggregate(qw->q, namestr, pnamestr, agg, list);
	  g_slist_free(list);
	}
      }
    }
    break;
  case 1:
    /* Forgetting about adding an object */
    break;
  }
  if (doremove) {
    object_add_close_dlg(qw);
  }
}

static void obj_del_button(GtkWidget *widget, SqlQueryWid *qw) 
{
  gint sel_row = -1;
  GList *sel;
  QueryElement *elt;

  sel = GTK_CLIST(qw->clist)->selection;
  if (sel)
    sel_row = GPOINTER_TO_INT(sel->data);    

  elt = (QueryElement *) gtk_clist_get_row_data(GTK_CLIST(qw->clist), sel_row);
  if (elt)
    sql_query_del_elt(qw->q, elt);
}

static void object_add_field_sel_cb(GtkWidget *widget, 
				    SqlMemTable *t, SqlMemField *f,
				    GnomeDialog *dlg)
{
  gboolean *ok_state;
  GtkWidget *wid;
  gchar *str;

  gtk_object_set_data(GTK_OBJECT(dlg), "field_selected", f);
  gnome_dialog_set_sensitive(GNOME_DIALOG(dlg), 0, f ? TRUE : FALSE);

  ok_state = gtk_object_get_data(GTK_OBJECT(dlg), "ok_state");
  *ok_state = f ? TRUE : FALSE;

  /* name and print name if not already selected */
  if (f) {
    wid = gtk_object_get_data(GTK_OBJECT(dlg), "obj_name_wid");
    if (*gtk_entry_get_text(GTK_ENTRY(wid)) == '\0')
      gtk_entry_set_text(GTK_ENTRY(wid), f->name);
    wid = gtk_object_get_data(GTK_OBJECT(dlg), "obj_pname_wid");
    if (*gtk_entry_get_text(GTK_ENTRY(wid)) == '\0') {
      str = g_strdup_printf("%s_%s", t->name, f->name);
      gtk_entry_set_text(GTK_ENTRY(wid), str);
      g_free(str);
    }
  }
}

static void object_add_close_dlg(SqlQueryWid *qw)
{
  gboolean *ok_state;
  GtkWidget *args_tab;
  
  gtk_widget_set_sensitive(GTK_WIDGET(qw->add_button), TRUE);
  ok_state = gtk_object_get_data(GTK_OBJECT(qw->obj_add_dlg), "ok_state");
  g_free(ok_state);
  args_tab = gtk_object_get_data(GTK_OBJECT(qw->obj_add_dlg), "fn_args_tab");
  g_free(args_tab);
  gnome_dialog_close(GNOME_DIALOG(qw->obj_add_dlg));
  qw->obj_add_dlg = NULL;
}

static void object_add_type_choice_toggle_cb(GtkWidget *wid,
					     GnomeDialog *dlg)
{
  GtkWidget *wid1, *wid2;
  GtkWidget *rb1, *rb2, *rb3;
  gboolean *state;
  gint offset = 0;
  
  wid1 = gtk_object_get_data(GTK_OBJECT(wid), "sh_wid");
  wid2 = gtk_object_get_data(GTK_OBJECT(wid), "sh_wid2");
  if (GTK_TOGGLE_BUTTON(wid)->active) {
    gtk_widget_show(wid1);
    if (wid2)
      gtk_widget_show(wid2);
    rb1 = gtk_object_get_data(GTK_OBJECT(dlg), "rb1");
    rb2 = gtk_object_get_data(GTK_OBJECT(dlg), "rb2");
    rb3 = gtk_object_get_data(GTK_OBJECT(dlg), "rb3");
    state = gtk_object_get_data(GTK_OBJECT(dlg), "ok_state");
    if (wid == rb1)
      offset = 0;
    if (wid == rb2)
      offset = 1;
    if (wid == rb3)
      offset = 2;
    gnome_dialog_set_sensitive(GNOME_DIALOG(dlg), 0, *(state+offset));
  }
  else {
    gtk_widget_hide(wid1);
    if (wid2)
      gtk_widget_hide(wid2);
  }
}

static void object_edit_cb(GtkWidget *widget, SqlQueryWid *qw)
{
  GtkWidget *dlg, *wid, *table;
  gchar *str;
  gint sel_row = -1;
  GList *sel;
  QueryElement *elt;
  gboolean exists;
  GSList *list;

  /* Which element */
  sel = GTK_CLIST(qw->clist)->selection;
  if (sel)
    sel_row = GPOINTER_TO_INT(sel->data);    

  elt = (QueryElement *) gtk_clist_get_row_data(GTK_CLIST(qw->clist), sel_row);

  /* is there already a dlg opened for that element ? */
  exists = FALSE;
  list = qw->edited_objects;
  while (list && !exists) {
    if (elt == gtk_object_get_data(GTK_OBJECT(list->data), "QueryElement"))
      exists = TRUE;
    list = g_slist_next(list);
  }

  /* new dialog creation */
  if (elt && !exists) {
    str = g_strdup_printf(_("Edition of object %s for query %s"), elt->name, 
			  qw->q->name);
    dlg = gnome_dialog_new(str, GNOME_STOCK_BUTTON_OK, 
			   GNOME_STOCK_BUTTON_CANCEL , NULL);
    gtk_window_set_policy(GTK_WINDOW(dlg), TRUE, TRUE, FALSE);
    g_free(str);
    
    /* widgets at the top */
    table = gtk_table_new(3, 2, FALSE);
    gtk_box_pack_start(GTK_BOX(GNOME_DIALOG(dlg)->vbox), 
		       table, TRUE, TRUE, 0);    
    gtk_widget_show(table);

    wid = gtk_label_new(_("Object type:"));
    gtk_label_set_justify(GTK_LABEL(wid), GTK_JUSTIFY_LEFT);
    gtk_table_attach_defaults(GTK_TABLE(table), wid,
			      0, 1, 0, 1);
    gtk_widget_show(wid);

    switch (elt->type) {
    case QUERY_ELT_FIELD:
      wid = gtk_label_new(_("Table field"));
      break;
    case QUERY_ELT_AGGREGATE:
      wid = gtk_label_new(_("Aggregate"));
      break;
    case QUERY_ELT_FUNCTION:
      wid = gtk_label_new(_("Function"));
      break;
    }
    gtk_label_set_justify(GTK_LABEL(wid), GTK_JUSTIFY_LEFT);
    gtk_table_attach_defaults(GTK_TABLE(table), wid,
			      1, 2, 0, 1);
    gtk_widget_show(wid);

    wid = gtk_label_new(_("Object name:"));
    gtk_table_attach_defaults(GTK_TABLE(table), wid,
			      0, 1, 1, 2);
    gtk_widget_show(wid);

    wid = gtk_label_new(_("Print name:"));
    gtk_label_set_justify(GTK_LABEL(wid), GTK_JUSTIFY_LEFT);
    gtk_table_attach_defaults(GTK_TABLE(table), wid,
			      0, 1, 2, 3);
    gtk_widget_show(wid);

    wid = gtk_entry_new();
    gtk_table_attach_defaults(GTK_TABLE(table), wid,
			      1, 2, 1, 2);
    gtk_widget_show(wid);
    gtk_entry_set_text(GTK_ENTRY(wid), elt->name);
    gtk_object_set_data(GTK_OBJECT(dlg), "eltname", wid);

    wid = gtk_entry_new();
    gtk_table_attach_defaults(GTK_TABLE(table), wid,
			      1, 2, 2, 3);
    gtk_widget_show(wid);
    if (elt->print_name)
      gtk_entry_set_text(GTK_ENTRY(wid), elt->print_name);
    gtk_object_set_data(GTK_OBJECT(dlg), "eltprintname", wid);
    
    gtk_signal_connect(GTK_OBJECT(dlg), "clicked",
		       GTK_SIGNAL_FUNC(object_edit_dialog_clicked_cb), qw);
    gtk_object_set_data(GTK_OBJECT(dlg), "QueryElement", elt);
    qw->edited_objects = g_slist_append(qw->edited_objects, dlg);
    gtk_widget_show(dlg);
  }
}

static void object_edit_dialog_clicked_cb(GnomeDialog * dialog, 
					  gint button_number, 
					  SqlQueryWid *qw)
{
  QueryElement *elt;
  GtkWidget *name, *pname;
  gchar *namestr, *pnamestr;
  GSList *list;
  gboolean doremove;

  doremove = TRUE;
  elt = gtk_object_get_data(GTK_OBJECT(dialog), "QueryElement");
  switch (button_number) {
  case 0: /* OK button was pressed */
    name = gtk_object_get_data(GTK_OBJECT(dialog), "eltname");
    pname = gtk_object_get_data(GTK_OBJECT(dialog), "eltprintname");
    namestr = gtk_entry_get_text(GTK_ENTRY(name)), 
    pnamestr = gtk_entry_get_text(GTK_ENTRY(pname)), 
    
    /* testing the validity of the object name */
    list = qw->q->objects;
    while (list && doremove) {
      if ((list->data != elt) && 
	  (!strcmp(((QueryElement *)(list->data))->name, namestr))) {
	doremove = FALSE;
	gnome_app_error(GNOME_APP(qw->q->conf->app), 
			_("The object name is already used"));
      }
      list = g_slist_next(list);
    }
    if (doremove && (*namestr == '\0')) {
      gnome_app_error(GNOME_APP(qw->q->conf->app), 
		      _("The object name is not valid"));
      doremove = FALSE;
    }
    /* validity of the object print name */
    if (doremove) {
      if (strpbrk(pnamestr, " .'\"\\-(){}[]|@")) {
	gnome_app_error(GNOME_APP(qw->q->conf->app), 
			_("The object print name is not valid"));
	doremove = FALSE;
      }
    }

    if (doremove)
      sql_query_modif_elt(qw->q, elt, namestr, pnamestr);
    break;
  case 1: /* Cancel button was pressed */
    break;
  }

  /* final action */
  if (doremove) {
    qw->edited_objects = g_slist_remove(qw->edited_objects, dialog);
    gnome_dialog_close(dialog);
  }
}


/*************************************************************************/
/* Callbacks to centralize the addition and removal of the rows in the   */
/* CList widget.  Are being called when SqlQuery emits the "elt_created" */
/* and "elt_dropped" signals                                             */
/*************************************************************************/

static void set_clist_entry_text(SqlQueryWid *qw, QueryElement *elt, 
				 gchar *row[]) 
{
  /* WARNING: the allocated strings here will have to be freed */
  GSList *list;
  GString *gstr;
  gboolean first = TRUE;
  
  switch (elt->type) {
  case QUERY_ELT_FIELD:
    row[0] = g_strdup(_("Table field"));
    row[2] = g_strdup_printf("%s.%s", 
			     sql_db_find_table_from_field(qw->q->conf->db,
					      SQL_MEM_FIELD(elt->main))->name, 
			     SQL_MEM_FIELD(elt->main)->name);
    break;
  case QUERY_ELT_FUNCTION:
    row[0] = g_strdup(_("Function"));
    list = elt->args;
    gstr = g_string_new("");
    g_string_sprintf(gstr, "%s (", SQL_DATA_FUNCTION(elt->main)->sqlname);
    while (list) {
      if (first) {
	g_string_sprintfa(gstr, "%s", ((QueryElement *)(list->data))->name);
	first = FALSE;
      }
      else 
	g_string_sprintfa(gstr, ", %s", ((QueryElement *)(list->data))->name);
      list = g_slist_next(list);
    }
    g_string_sprintfa(gstr, ")");
    row[2] = g_strdup(gstr->str);
    g_string_free(gstr, TRUE);
    break;
  case QUERY_ELT_AGGREGATE:
    row[0] = g_strdup(_("Aggregate"));
    row[2] = g_strdup_printf("%s (%s)", SQL_DATA_AGGREGATE(elt->main)->sqlname,
			     ((QueryElement *)(elt->args->data))->name);
    break;
  }
  row[1] = g_strdup(elt->name);
  row[3] = g_strdup(elt->print_name);  
}

static void sql_query_wid_elt_created_cb(GtkObject *obj, 
					 QueryElement *elt,
					 SqlQueryWid *qw)
{
  gchar *row[4];
  gint i;

  set_clist_entry_text(qw, elt, row); 
  i = gtk_clist_append(GTK_CLIST(qw->clist), row);
  gtk_clist_set_row_data(GTK_CLIST(qw->clist), i, elt);
  gtk_clist_select_row(GTK_CLIST(qw->clist), i, 0);
  for (i=0; i<4; i++)
    g_free(row[i]);
}


static void sql_query_wid_elt_dropped_cb(GtkObject *obj, 
					 QueryElement *elt,
					 SqlQueryWid *qw)
{
  gint row, sel_row = -1;
  GList *sel;
  GSList *list;
  QueryElement *telt;

  /* in the CList */
  sel = GTK_CLIST(qw->clist)->selection;
  if (sel)
    sel_row = GPOINTER_TO_INT(sel->data);    

  row = gtk_clist_find_row_from_data(GTK_CLIST(qw->clist), elt);
  if (row >= 0) {
    gtk_clist_remove(GTK_CLIST(qw->clist), row);
    if (row == sel_row) {
      gtk_widget_set_sensitive(qw->del_button, FALSE);
      gtk_widget_set_sensitive(qw->edit_button, FALSE);
    }
  }

  /* ANY Edit dialog opened for that object is closed */
  list = qw->edited_objects;
  while (list) {
    telt = gtk_object_get_data(GTK_OBJECT(list->data), "QueryElement");
    if (telt == elt) {
      object_edit_dialog_clicked_cb(GNOME_DIALOG(list->data), 
				    1 /* CANCEL BUTTON */, qw);
      list = qw->edited_objects;
    }
    else
      list = g_slist_next(list);
  }
}

static void sql_query_wid_elt_modified_cb(GtkObject *obj, 
					  QueryElement *elt,
					  SqlQueryWid *qw)
{
  GSList *elts;
  gint row, i;
  gchar *rows[4];

  /* That element in the clist */
  row = gtk_clist_find_row_from_data(GTK_CLIST(qw->clist), elt);
  if (row >= 0) {
    gtk_clist_freeze(GTK_CLIST(qw->clist));
    set_clist_entry_text(qw, elt, rows); 
    for (i=0; i<4; i++) {
      gtk_clist_set_text(GTK_CLIST(qw->clist), row, i, rows[i]);
      g_free(rows[i]);
    }
    gtk_clist_thaw(GTK_CLIST(qw->clist));
  }
  
  /* other elements which might reference that elt.
     NO recursive parsing because of what is displayed is only one leved
     deep. */
  elts = qw->q->objects;
  while (elts) {
    if (g_slist_find(((QueryElement *)(elts->data))->args, elt)) {
      row = gtk_clist_find_row_from_data(GTK_CLIST(qw->clist), 
					 (QueryElement *)(elts->data));
      if (row) {
	gtk_clist_freeze(GTK_CLIST(qw->clist));
	set_clist_entry_text(qw, (QueryElement *)(elts->data), rows); 
	for (i=0; i<4; i++) {
	  gtk_clist_set_text(GTK_CLIST(qw->clist), row, i, rows[i]);
	  g_free(rows[i]);
	}
	gtk_clist_thaw(GTK_CLIST(qw->clist));
      }
    }
    elts = g_slist_next(elts);
  }
}

static void sql_query_wid_changed_cb(GtkObject *obj, 
				     SqlQueryWid *qw)
{
  gchar *str;

  gtk_widget_set_sensitive(qw->text_valid_btn, FALSE);
  if (! qw->q->text_only) {
    /* removing old content */
    gtk_text_freeze(GTK_TEXT(qw->text));
    gtk_text_set_point(GTK_TEXT(qw->text), 0);
    gtk_text_forward_delete(GTK_TEXT(qw->text), 
			    gtk_text_get_length(GTK_TEXT(qw->text)));

    /* putting new content */
    str = sql_query_get_select_query(qw->q);
    gtk_text_insert(GTK_TEXT(qw->text), NULL, NULL, NULL, str, -1);
    g_free(str);
    gtk_text_thaw(GTK_TEXT(qw->text));
  }
}

/****************************************************************************/
/*                                                                          */
/* Functions related to the Dialog in which is packed the SqlQueryWid obj.  */
/*                                                                          */
/****************************************************************************/
static void main_dialog_clicked_cb(GnomeDialog * dialog, 
				   gint button_number, 
				   SqlQueryWid *qw);

static void main_dialog_auto_close(GtkObject *obj, GnomeDialog *dlg)
{
  SqlQueryWid *qw = gtk_object_get_data(GTK_OBJECT(dlg), "SqlQueryWid");
  main_dialog_clicked_cb(dlg, 0, qw);
}

GtkWidget* sql_query_edit_new_dlg(SqlQuery *q)
{
  GtkWidget *dlg, *wid;
  gchar *str;

  str = g_strdup_printf(_("Edition of query %s"), q->name);
  dlg = gnome_dialog_new(str, GNOME_STOCK_BUTTON_OK, NULL);
  gtk_window_set_policy(GTK_WINDOW(dlg), TRUE, TRUE, FALSE);
  g_free(str);

  /* widgets at the top */
  wid = sql_query_wid_new(q);
  gtk_box_pack_start(GTK_BOX(GNOME_DIALOG(dlg)->vbox), 
		     wid, TRUE, TRUE, 0);
  gtk_widget_show(wid);
  
  gtk_signal_connect(GTK_OBJECT(dlg), "clicked",
		     GTK_SIGNAL_FUNC(main_dialog_clicked_cb),
		     wid);
  gtk_object_set_data(GTK_OBJECT(dlg), "SqlQueryWid", wid);
  gtk_signal_connect_while_alive(GTK_OBJECT(q->conf->srv), "conn_closed",
				 GTK_SIGNAL_FUNC(main_dialog_auto_close), dlg,
				 GTK_OBJECT(dlg));
  return dlg;
}

static void main_dialog_clicked_cb(GnomeDialog * dialog, 
				   gint button_number, 
				   SqlQueryWid *qw)
{
  if (qw->obj_add_dlg)
    object_add_close_dlg(qw);
  gnome_dialog_close(dialog);
}


/***********************************************************************/
/*                                                                     */
/* Notebook Text only Page CBs                                         */
/*                                                                     */
/***********************************************************************/
static void text_only_tbutton_cb(GtkWidget *widget, SqlQueryWid *qw)
{
  GtkWidget *wid;
  gchar *str;
  gint i;

  if (GTK_TOGGLE_BUTTON(widget)->active) {
    str = gtk_editable_get_chars(GTK_EDITABLE(qw->text), 0, -1);
    sql_query_set_text_sql(qw->q, str);
    g_free(str);
  }
  else
    sql_query_set_text_sql(qw->q, NULL);

  gtk_text_set_editable(GTK_TEXT(qw->text), qw->q->text_only);
  for (i=0; i<4; i++) {
    wid =  gtk_notebook_get_nth_page(GTK_NOTEBOOK(qw->notebook), i);
    gtk_widget_set_sensitive(wid, ! qw->q->text_only);
  }
}

static void text_only_validate_cb(GtkWidget *widget, SqlQueryWid *qw)
{
  gchar *str;

  str = gtk_editable_get_chars(GTK_EDITABLE(qw->text), 0, -1);
  sql_query_set_text_sql(qw->q, str);
  g_free(str);
}


/***********************************************************************/
/*                                                                     */
/* Automatic dependencies CBs                                          */
/*                                                                     */
/***********************************************************************/
static void sql_query_wid_auto_links_add_table(SqlQueryWid *qw, SqlMemTable *t)
{
  GSList *list;
  SqlMemField *f;
  SqlMemTable *tab;

  list = qw->q->conf->db->field_links;
  while (list) {
    f = SQL_MEM_FLINK(list->data)->from;
    tab = sql_db_find_table_from_field(qw->q->conf->db, f);
    if (tab == t)
      sql_query_auto_links_add_link(NULL, SQL_MEM_FLINK(list->data), qw->q);
    else {
      f = SQL_MEM_FLINK(list->data)->to;
      tab = sql_db_find_table_from_field(qw->q->conf->db, f);
      if (tab == t)
	sql_query_auto_links_add_link(NULL, SQL_MEM_FLINK(list->data), qw->q);
    }
    list = g_slist_next(list);
  }
}

static void sql_query_wid_auto_links_del_table(SqlQueryWid *qw, SqlMemTable *t)
{
  GSList *list;
  SqlMemField *f;
  SqlMemTable *tab;

  list = qw->q->conf->db->field_links;
  while (list) {
    f = SQL_MEM_FLINK(list->data)->from;
    tab = sql_db_find_table_from_field(qw->q->conf->db, f);
    if (tab == t)
      sql_query_auto_links_del_link(NULL, SQL_MEM_FLINK(list->data), qw->q);
    else {
      f = SQL_MEM_FLINK(list->data)->to;
      tab = sql_db_find_table_from_field(qw->q->conf->db, f);
      if (tab == t)
	sql_query_auto_links_del_link(NULL, SQL_MEM_FLINK(list->data), qw->q);
    }
    list = g_slist_next(list);
  }
}


static void depend_clist_select_row_cb(GtkWidget *wid, 
				       gint row, gint column, 
				       GdkEventButton *event,
				       SqlQueryWid *qw)
{
  QueryLink *ql;
  qw->depend_sel = row;
  ql = gtk_clist_get_row_data(GTK_CLIST(wid), row);
  if (ql) {
    gtk_widget_set_sensitive(qw->use_button, TRUE);
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(qw->use_button), ql->useit);
  }
  else
    gtk_widget_set_sensitive(qw->use_button, FALSE);
}

static void depend_update_one_entry_param(GtkWidget *button, SqlQueryWid *qw,
					  gint row)
{
  QueryLink *ql;
  gchar *str;

  ql = gtk_clist_get_row_data(GTK_CLIST(qw->depend_clist), row);
  if (ql) {
    if (ql->useit) {
      str = g_strdup(_("YES"));
    }
    else {
      str = g_strdup(_("NO"));
    }
    gtk_clist_set_text(GTK_CLIST(qw->depend_clist), row, 0, str);
    g_free(str);

    if (row == qw->depend_sel)
      gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(qw->use_button),
				   ql->useit); 
    sql_query_wid_changed_cb(GTK_OBJECT(qw->q), qw);   
  }
}

/* updates the entry status of the selected row */
static void depend_update_one_entry(GtkWidget *button, SqlQueryWid *qw)
{
  QueryLink *ql;

  if (qw->depend_sel > -1) {
    ql = gtk_clist_get_row_data(GTK_CLIST(qw->depend_clist), qw->depend_sel);
    ql->useit = gtk_toggle_button_get_active(
				   GTK_TOGGLE_BUTTON(qw->use_button)); 
    depend_update_one_entry_param(button, qw, qw->depend_sel);
#ifdef debug_signal
    g_print(">> 'CHANGED' from sqlquerywid->depend_update_one_entry\n");
#endif
    gtk_signal_emit_by_name(GTK_OBJECT(qw->q), "changed");
#ifdef debug_signal
    g_print("<< 'CHANGED' from sqlquerywid->depend_update_one_entry\n");
#endif
  }
  else
    gtk_widget_set_sensitive(qw->use_button, FALSE);
}

static void depend_add_to_list(GtkObject *obj, QueryLink *ql,
			       SqlQueryWid *qw)
{
  /* add an entry to the clist */
  gchar *text[3];
  SqlMemTable *t;
  gboolean ok, signalit;
  gint row;

  if (ql->useit) {
    text[0] = _("YES");
    signalit = TRUE;
  }
  else {
    text[0] = _("NO");
    signalit = FALSE;
  }
  ok = TRUE;

  t = sql_db_find_table_from_field(qw->q->conf->db, ql->link->from);
  if (t) {
    text[1] = g_strdup_printf("%s.%s", t->name, ql->link->from->name);
  }
  else {
    ok = FALSE;
    text[1] = NULL;
  }

  t = sql_db_find_table_from_field(qw->q->conf->db, ql->link->to);
  if (t && ok) {
    text[2] = g_strdup_printf("%s.%s", t->name, ql->link->to->name);
  }
  else
    ok = FALSE;

  if (ok) {
    row = gtk_clist_append(GTK_CLIST(qw->depend_clist), text);
    gtk_clist_set_row_data(GTK_CLIST(qw->depend_clist), row, ql);
    g_free(text[1]);
    g_free(text[2]);
  }
  else
    if (text[1])
      g_free(text[1]);

  if (signalit)
    sql_query_wid_changed_cb(GTK_OBJECT(qw->q), qw);   
  qw->q->conf->save_up_to_date = FALSE;
}

static void depend_remove_from_list(GtkObject *obj, QueryLink *ql,
				    SqlQueryWid *qw)
{
  gint row;

  row = gtk_clist_find_row_from_data(GTK_CLIST(qw->depend_clist), ql);
  if (row != -1) {
    if (qw->depend_sel == row) {
      qw->depend_sel = -1;
      gtk_widget_set_sensitive(qw->use_button, FALSE);
    }
    gtk_clist_remove(GTK_CLIST(qw->depend_clist), row);
  }

  if (ql->useit)
    sql_query_wid_changed_cb(GTK_OBJECT(qw->q), qw);   
  qw->q->conf->save_up_to_date = FALSE;
}

static void refresh_depend(GtkWidget *bttn, SqlQueryWid *qw) 
{
  gint i;
  QueryLink *ql;

  sql_query_auto_links_refresh(qw->q);

  for (i=0; i< GTK_CLIST(qw->depend_clist)->rows; i++) {
    ql = (QueryLink *) gtk_clist_get_row_data(GTK_CLIST(qw->depend_clist), i);
    depend_update_one_entry_param(NULL, qw, i);
  }
  qw->q->conf->save_up_to_date = FALSE;
}

static void use_all_depend(GtkWidget *bttn, SqlQueryWid *qw)
{
  gint i;
  QueryLink *ql;
 
  for (i=0; i< GTK_CLIST(qw->depend_clist)->rows; i++) {
    ql = (QueryLink *) gtk_clist_get_row_data(GTK_CLIST(qw->depend_clist), i);
    ql->useit = TRUE;
    depend_update_one_entry_param(NULL, qw, i);
  }
  qw->q->conf->save_up_to_date = FALSE;
}

static void use_none_depend(GtkWidget *bttn, SqlQueryWid *qw)
{
  QueryLink *ql;
  gint i;

  for (i=0; i< GTK_CLIST(qw->depend_clist)->rows; i++) {
    ql = (QueryLink *) gtk_clist_get_row_data(GTK_CLIST(qw->depend_clist), i);
    ql->useit = FALSE;
    depend_update_one_entry_param(NULL, qw, i);
  }
  qw->q->conf->save_up_to_date = FALSE;
}
